#
# $Id: bstream.rb,v 1.18 2002/06/18 07:21:01 abe Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   BlockStream Library
#
#      Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>
######################################################################


#
#   String utility
######################################################################

def str_to_hash ( string )
  return Ext::Block::str_to_hash( string )
end

def hash_to_str ( tbl )
  return Ext::Block::hash_to_str( tbl )
end

def hash_to_pstr ( tbl )
  nl = "\n"
  lines = [ ]
  maxlen = 0
  tbl.each_key do |key|
    key = String(key)
    maxlen = key.size if maxlen < key.size
  end
  tbl.each_pair do |key, cont|
    cont = String(cont)
    if cont.index(nl) then
      lines << "!#{key}\n#{cont}"
    else
      lines << "!#{key.to_s.ljust(maxlen)}  #{cont}"
    end
  end
  lines << ""
  return lines.join("\n")
end


#
#   String expansion
######################################################################

class String

  alias write concat

  def gets ( rs = $/ )
    return nil if self.empty?
   ($_ = self.index(rs))? self.slice!( 0, $_ + rs.length ): self.slice!( 0, self.length )
  end

  def display ( out )
    out << self
  end

  def to_hash ()
    str_to_hash( self )
  end

end


#
#   Array expansion
######################################################################

class Array

  alias write push

end


#
#   Extention
######################################################################

require File::join( File.dirname(__FILE__), "bstream.so" ) unless defined? Ext and defined? Ext::Block
Block = Ext::Block
DataBlock = Ext::DataBlock
PacketBlock = Ext::PacketBlock
HashBlock = Ext::HashBlock


#
#   Block Cache
######################################################################

class Block
  def self.mkstr_property ( prop )
    prop = prop.to_hash unless prop.is_a?(Hash)
    Array(prop).collect {|pair| "#{pair.first}=\"#{pair.last}\"" }.join(' ')
  end
end

class DataBlock
  class Cache
    def initialize ( block )
      @name = block.name
      @block_string = String.new("")
      block.out = @block_string
      block.close()
    end
    def display ( out = $defout )
      Block.display( @name, out ) do
        out << @block_string
      end
    end
    attr_reader :block_string
  end
end

class PacketBlock
  class Cache
    def initialize ( block )
      @name = block.name
      @block_list = [ ]
      block.each do |child|
        @block_list << BlockCache.new( child ) 
      end
    end
    def display ( out = $defout )
      Block.display( @name.out ) do
        @block_list.each do |child|
          next unless child
          child.display( out )
        end
      end
    end
    attr_reader :block_list
  end
end

class BlockCache
  # new (block) : block LbV
  #   nꂽ block ɉAK؂ȃNX̃CX^X𐶐B
  def self.new ( block )
    if DataBlock === block then
      DataBlock::Cache.new( block )
    elsif PacketBlock === block then
      PacketBlock::Cache.new( block )
    else
      raise "BlockCache.new() expects Ext::Block."
    end
  end
end


