/*                                                     project Gemini 
----------------------------------------------------------------------
   Graphical EM Core Module
      $Id: gem.c,v 1.7 2002/11/21 13:52:57 abe Exp $

       Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define PRINT_STORE_PARAM 0
#define PRINT_READ_PARAM 0
#define PRINT_SGRAPH_COMPILE 0
#define PRINT_INSIDE_OUTSIDE 0
#define PRINT_PARAM_UPDATE_LOG 0 /**< p[^XṼfobOx */
#include "gem.h"
#include <math.h>


static VALUE rb_mExt;
static VALUE rb_cGEM;
static VALUE rb_cSG;

static ID id_store;


/********************************************************************/


/*   ͕⏕
------------------------------------------------------------------- */
#define SKIP_WS(PTR)  for( ; *(PTR) == ' ' || *(PTR) == '\t'; ++(PTR) ) ;

static VALUE strparse_sgraph_path ( const char * p )
{
    VALUE volatile localpath = Qnil;
    const char *q, *r;
    int n;
    
    /* /\s+\+-+/ */
    if( *p != ' ' && *p != '\t' )
        return Qnil;
    ++p;
    SKIP_WS( p );
    
    if( p[0] != '+' && p[1] != '-' )
        return Qnil;
    p += 2;
    
    /* pXł邱Ƃm */
    while( *p == '-' )
        ++p;
    localpath = rb_ary_new();
    
    /* m[hEĂ */
    while( *p != '\r' && *p != '\n' && *p != '\0' ) {
        SKIP_WS( p );
        if( *p == '\r' || *p == '\n' || *p == '\0' )
            break;
        
        if( p[0] == '(' && p[1] == ':' ) { /* {m[h = (: ... :) */
            p += 2;
            SKIP_WS( p );
            q = p;
            while( p[0] != ':' || p[1] != ')' ) {
                if( *p == '\0' )
                    rb_raise( rb_eRuntimeError, "support graph syntax error (no ``:)'')" );
                ++p;
            }
            r = p;
            while( r > q && (r[-1] == ' ' || r[-1] == '\t') )
                --r;
            rb_ary_push( localpath, rb_str_new( q, r-q ) );
            p += 2;
        }
        else if( p[0] == '@' ) { /* Qƃm[h = @... */
            ++p;
            n = 0;
            if( !('0' <= p[0] && p[0] <= '9') )
                return Qnil;
            n += (*(p++) - '0');
            while( '0' <= p[0] && p[0] <= '9' )
                n = n*10 + (*(p++) - '0');
            rb_ary_push( localpath, INT2FIX(n) );
        }
        else {
            rb_raise( rb_eRuntimeError, "support graph syntax error (rest string=%s)", p );
        }
    }
    
    return localpath;
}


/********************************************************************/


/*   gEM IuWFNg̉
------------------------------------------------------------------- */
static void gem_free ( GraphicalEM* gem )
{
    if( gem ) {
        ruby_xfree( gem->swary );
        ruby_xfree( gem->probtbl );
        ruby_xfree( gem->xnumtbl );
        ruby_xfree( gem );
    }
}

static void gem_mark ( GraphicalEM* gem )
{
    if( gem ) {
        rb_gc_mark( gem->paramtbl );
        rb_gc_mark( gem->paramary );
        rb_gc_mark( gem->sgary );
    }
}


/*   gEM IuWFNg̐
----------------------------------------------------------------------
GraphicalEM.new( initparam )

   initparam ɂ́Ap[^f[^\wB
   initparam ̍\͎̒ʂB

initparam :- Array<Problist>
Problist  :- Array<[m, XCb`]>
*/
static VALUE gem_new ( VALUE klass, VALUE initparam )
{
    GraphicalEM* gem;
    VALUE volatile obj;
    
    gem = ALLOC( GraphicalEM );
    MEMZERO( gem, GraphicalEM, 1 );
    obj = Data_Wrap_Struct( klass, gem_mark, gem_free, gem );
    rb_obj_call_init( obj, 1, &initparam );
    
    return obj;
}

static VALUE gem_initialize ( VALUE self, VALUE initparam )
{
    VALUE volatile pair;
    VALUE volatile str;
    GraphicalEM* gem;
    RArray_p paramary, groupary;
    real_t sum;
    int i, j, n;
    
    /* IuWFNg̃ZbgAbv */
    Data_Get_Struct( self, GraphicalEM, gem );
    
    gem->paramtbl = rb_hash_new();
    gem->paramary = rb_ary_new();
    gem->sgary = rb_ary_new();
    
    /* ^ꂽf[^ǂݍ */
    initparam = rb_Array( initparam );
    paramary = RARRAY(initparam);
    for( i = 0; i < paramary->len; ++i ) { /* e Problist Ɋւ... */
        paramary->ptr[i] = rb_Array( paramary->ptr[i] ); /* a = prob Array */
        groupary = RARRAY(paramary->ptr[i]);
        for( j = 0; j < groupary->len; ++j ) { /* e Prob-Name Ɋւ... */
            pair = rb_Array( groupary->ptr[j] ); /* p = Pair */
            if( RARRAY(pair)->len != 2 )
                rb_raise( rb_eArgError, "wrong gEM parameter format." );
            str = RARRAY(pair)->ptr[1];
            groupary->ptr[j] = pair;
            
            /* ܂̓p[^Ƃ̃CfNXʒunbVɊi[ */
            n = RARRAY(gem->paramary)->len;
            rb_hash_aset( gem->paramtbl, str, INT2FIX(n) );
            #if PRINT_STORE_PARAM
            fprintf( stdout, "gem_initialize(): gem->paramtbl: %s = %d\n", STR2CSTR(str), n );
            #endif
            rb_ary_push( gem->paramary, str );
        }
    }
    gem->problen = RARRAY(gem->paramary)->len;
    gem->probtbl = ALLOC_N( real_t, gem->problen );
    gem->xnumtbl = ALLOC_N( real_t, gem->problen );
    
    /* mf[^蓖ĂĂ */
    n = 0;
    gem->swlen = paramary->len;
    gem->swary = ALLOC_N( int, gem->swlen ); /* XCb`̃Xg */
    for( i = 0; i < gem->swlen; ++i ) {
        groupary = RARRAY(paramary->ptr[i]);
        gem->swary[i] = groupary->len;
        for( j = 0; j < groupary->len; ++j ) {
            pair = groupary->ptr[j];
            gem->probtbl[n] = NUM2DBL( RARRAY(pair)->ptr[0] );
            gem->xnumtbl[n] = 0;
            ++n;
        }
    }
    
    /* ml̃`FbN (1.0 ɂȂȂꍇ͗O) */
    i = 0;
    for( n = 0; n < gem->swlen; ++n ) {
        sum = 0.0;
        for( j = 0; j < gem->swary[n]; ++j ) {
            sum += gem->probtbl[i];
            ++i;
        }
        if( sum < 0.999999 || sum > 1.000001 )
            rb_raise( rb_eRuntimeError, "unconsistent group #%d", n );
    }
    
    return self;
}


/*   gEM IuWFNgɎxOtǉ
----------------------------------------------------------------------
gem.store( sg )

   sg  GraphicalEM IuWFNg̏ΏۂɒǉB
   o^ĂȂ{m[h (p[^) ꍇ́A
ArgumentError OB
*/
static VALUE gem_store ( VALUE self, VALUE sgobj )
{
    SupportGraph* sg;
    GraphicalEM* gem;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    
    /* ^`FbN */
    if( CLASS_OF(sgobj) != rb_cSG )
        rb_raise( rb_eArgError, "expects SupportGraph instance." );
    
    /* ̂߂̃RpC */
    Data_Get_Struct( sgobj, SupportGraph, sg );
    sgraph_compile( sg, gem );
    
    /* o^ */
    rb_ary_push( gem->sgary, sgobj );
    
    return self;
}


char * node_string ( char * buffer, GraphicalEM const * gem, SupportGraph const * sg, SG_Node const * node )
{
    VALUE val;
    const char * srcstr;
    char * p = buffer;
    
    if( sgraph_compiled_p( sg ) ) {
        if( &gem->probtbl[0] <= node->d.dir.i_ptr && node->d.dir.i_ptr < &gem->probtbl[ gem->problen ] ) {
            /* i_ptr  gem->probtbl ȂΊ{m[h */
            val = RARRAY(gem->paramary)->ptr[ node->d.dir.i_ptr - &gem->probtbl[0] ];
        }
        else {
            /* ȊO̓Otobt@Qƃm[h */
            p += sprintf( p, "@%d", ((SG_Graph*)node->d.dir.i_ptr) - sg->buffer_graph );
            return p; /* Qƃm[hȂ̂łŕԂ */
        }
    }
    else {
        switch( node->d.ind.type ) {
        case 1:
            val = RARRAY(gem->paramary)->ptr[ node->d.ind.data.sw ];
            break;
        case 2:
            p += sprintf( p, "@%d", (node->d.ind.data.graph - sg->buffer_graph) );
            return p; /* Qƃm[h͂ŕԂ */
        case 3:
            val = node->d.ind.data.obj;
            break;
        default:
            p += sprintf( p, "[BROKEN NODE]" );
            return p;
        }
    }
    
    /* ȍ~͊{m[hp̏ (val [ĝݓB\) */
    if( TYPE(val) == T_STRING )
        srcstr = RSTRING(val)->ptr;
    else {
        p += sprintf( p, "[INVALID TYPE]" );
        return p;
    }
    p += sprintf( p, "(:%s:)", srcstr );
    
    return p;
}


/*   p[^XV̎
------------------------------------------------------------------- */

/** @brief m̌vZ
 *
 * w肳ꂽxOtɊւāAm̌vZs܂B
 * ʂ͎xOtIuWFNgɏ܂܂B
 *
 * @param sgraph [i/o] vZΏۂ̎xOt
 * @return xOt̐m
 */
static real_t inside_process_fast ( SupportGraph * sgraph )
{
    SG_Graph  *graph;
    SG_Path   *path;
    SG_Node   *node;
    real_t  w, sum, sentprob;
    int     i, j, k;
    
    /*  Inside */
    graph = &sgraph->buffer_graph[ sgraph->size_graph-1 ];
    for( i = sgraph->size_graph; i > 0; --i ) {
        sum = 0.0;
        path = graph->pathary;
        for( j = graph->pathlen; j > 0; --j ) {
            w = 1.0;
            node = path->nodeary;
            for( k = path->nodelen; k > 0; --k ) {
                w *= *(node->d.dir.i_ptr);
                ++node;
            }
            sum += w;
            path->inside = w;
            ++path;
        }
        graph->inside = sum;
        --graph;
    }
    sentprob = sgraph->buffer_graph[0].inside;
    
    return sentprob;
}

/** @brief Om̌vZ
 *
 * w肳ꂽxOtɊւāAOmƊғKp񐔂vZ܂B
 * Om͎xOtIuWFNgɏ܂܂AғKp񐔂́AxOtێĂ
 * gEM IuWFNgɏ܂܂B
 *
 * @param sgraph [i/o] vZΏۂ̎xOt
 */
static void outside_process_fast ( SupportGraph * sgraph )
{
    SG_Graph  *graph;
    SG_Path   *path;
    SG_Node   *node;
    real_t  sum;
    int     i, j, k;
    
    graph = sgraph->buffer_graph;
    for( i = 0; i < sgraph->size_graph; ++i )
        (graph++)->outside = 0;
    sgraph->buffer_graph[0].outside = 1.0;
    
    graph = sgraph->buffer_graph;
    for( i = sgraph->size_graph; i > 0; --i ) {
        path = graph->pathary;
        for( j = graph->pathlen; j > 0; --j ) {
            sum = graph->outside * path->inside;
            node = path->nodeary;
            for( k = path->nodelen; k > 0; --k ) {
                if( *node->d.dir.o_inside > 0 )
                    *(node->d.dir.o_ptr) += sum / *(node->d.dir.o_inside);
                ++node;
            }
            ++path;
        }
        ++graph;
    }
}

static real_t learn_param_fast ( GraphicalEM * gem, SupportGraph * sgraph, int num )
{
    real_t  sentprob;
    int  i;
    
    sentprob = inside_process_fast( sgraph );
    if( sentprob == 0.0 )
        rb_raise( rb_eRuntimeError, "Sentence[%d]'s probability is zero.", num );
    outside_process_fast( sgraph );
    
    #if PRINT_INSIDE_OUTSIDE
    printf( "INSIDE and OUTSIDE:\n" );
    for( i = 0; i < sgraph->size_graph; ++i ) {
        printf( "Graph[%d] inside=%g, outside=%g\n", i, sgraph->buffer_graph[i].inside, sgraph->buffer_graph[i].outside );
    }
    printf( "----\n" );
    #endif
    
    return log( sentprob );
}


/*   p[^s
------------------------------------------------------------------- */

static real_t gem_likelihood ( GraphicalEM * gem )
{
    SupportGraph * sgraph;
    real_t  likelihood, sentprob;
    VALUE * sgptr;
    int  i, sglen;
    
    sgptr = RARRAY(gem->sgary)->ptr;
    sglen = RARRAY(gem->sgary)->len;
    
    /* m̌vZ */
    likelihood = 0.0;
    for( i = 0; i < sglen; ++i ) {
        Data_Get_Struct( sgptr[i], SupportGraph, sgraph );
        if( sgraph_compiled_p(sgraph) ) {
            sentprob = inside_process_fast( sgraph );
            if( sentprob == 0.0 )
                rb_raise( rb_eRuntimeError, "inside probability of sentence #%d is zero", i+1 );
            likelihood += log(sentprob);
        }
        else
            rb_raise( rb_eNotImpError, "uncompiled supportgraph not supported." );
    }
    
    return likelihood;
}

static VALUE gem_likelihood_m ( VALUE self )
{
    GraphicalEM * gem;
    real_t  likelihood;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    
    likelihood = gem_likelihood( gem );
    
    return rb_float_new( likelihood );
}

static void gem_update ( GraphicalEM * gem )
{
    SupportGraph * sgraph;
    VALUE * sgptr;
    real_t  sum;
    int  sglen;
    int  i, j, k, swptr;
    
    sgptr = RARRAY(gem->sgary)->ptr;
    sglen = RARRAY(gem->sgary)->len;
    
    MEMZERO( gem->xnumtbl, real_t, gem->problen );
    
    /* Om̌vZ */
    for( i = 0; i < sglen; ++i ) {
        Data_Get_Struct( sgptr[i], SupportGraph, sgraph );
        if( sgraph_compiled_p(sgraph) )
            outside_process_fast( sgraph );
        else
            rb_raise( rb_eNotImpError, "uncompiled supportgraph not supported." );
    }
    
    /* p[^̍XV */
    swptr = 0;
    for( i = 0; i < gem->swlen; ++i ) {
        k = swptr + gem->swary[i];
        
        sum = 0.0;
        for( j = swptr; j < k; ++j )
            sum += gem->xnumtbl[ j ];
        
        if( sum > 0.0 ) {
            for( j = swptr; j < k; ++j ) {
                #if PRINT_PARAM_UPDATE_LOG > 0
                {
                    double prev = gem->probtbl[ j ];
                    double after;
                    char buf [ 4096 ];
                    char * p;
                #endif
                
                gem->probtbl[ j ] = gem->xnumtbl[ j ] / sum;
                
                #if PRINT_PARAM_UPDATE_LOG > 0
                    after = gem->probtbl[j];
                    if( prev != after && after != 0 ) {
                        p = buf;
                        sprintf( buf, "%s : %g --> %g\n", gem_param_at(gem, j), prev, after );
                        fprintf( stderr, buf );
                    }
                }
                #endif
            }
        }
        swptr = k;
    }
}

static VALUE gem_update_m ( VALUE self )
{
    GraphicalEM * gem;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    gem_update( gem );
    
    return self;
}


/*   o^ꂽxOt̃Xg𓾂
----------------------------------------------------------------------
gem.sglist

   o^ꂽxOt̃XgRs[ĕԂB
*/
static VALUE gem_sglist ( VALUE self )
{
    GraphicalEM* gem;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    
    return rb_obj_dup( gem->sgary );
}


/*   p[^̏o
----------------------------------------------------------------------
gem.display ( out )

   wKp[^ out ֏o͂B
*/
static VALUE gem_display ( VALUE self, VALUE out )
{
    GraphicalEM* gem;
    char buf [4096];
    int sw, i, n;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    
    rb_io_write( out, rb_str_new2("<pParameter>\n") );
    n = 0;
    for( sw = 0; sw < gem->swlen; ++sw ) {
        rb_io_write( out, rb_str_new2("<dGroup>\n") );
        for( i = 0; i < gem->swary[sw]; ++i ) {
            sprintf( buf, "%g %s\n", gem->probtbl[n], STR2CSTR(RARRAY(gem->paramary)->ptr[n]) );
            rb_io_write( out, rb_str_new2( buf ) );
            ++n;
        }
        rb_io_write( out, rb_str_new2("</dGroup>\n") );
    }
    rb_io_write( out, rb_str_new2("</pParameter>\n") );
    
    return self;
}


const char * gem_param_at ( GraphicalEM const * gem, int id )
{
    int  problen;
    
    problen = RARRAY(gem->paramary)->len;
    if( id < 0 || id >= problen )
        rb_raise( rb_eRangeError, "gEM parameter index out of range (given %d for 0...%d)", id, problen );
    
    if( problen != gem->problen )
        rb_raise( rb_eRuntimeError, "broken gEM object (problen confliction)" ); /* NȂƂ͎v */
    
    return RSTRING(RARRAY(gem->paramary)->ptr[id])->ptr;
}


/*   o^ꂽxOt𓾂
----------------------------------------------------------------------
gem.size

*/
static VALUE gem_size ( VALUE self )
{
    GraphicalEM* gem;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    
    return INT2NUM( RARRAY(gem->sgary)->len );
}


/*   o^ꂽxOt̓mo
------------------------------------------------------------------- */

static VALUE gem_display_inside ( VALUE self, VALUE out )
{
    GraphicalEM * gem;
    SupportGraph * sgraph;
    VALUE * sgptr;
    char  buffer [ 4096 ];
    int  i, sglen;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    sgptr = RARRAY(gem->sgary)->ptr;
    sglen = RARRAY(gem->sgary)->len;
    
    for( i = 0; i < sglen; ++i ) {
        Data_Get_Struct( sgptr[i], SupportGraph, sgraph );
        sprintf( buffer, "[#%d] %g\n", i, sgraph->buffer_graph[0].inside );
        rb_io_write( out, rb_str_new2(buffer) );
    }
    
    return self;
}


/*   o^ꂽxOto
------------------------------------------------------------------- */

static VALUE gem_display_sg ( VALUE self, VALUE out )
{
    GraphicalEM * gem;
    SupportGraph * sgraph;
    SG_Graph * graph;
    SG_Path * path;
    SG_Node * node;
    VALUE * sgptr;
    char  buffer [ 4096 ];
    char  strbuf [ 4096 ];
    char * p;
    int  i, j, k, sgnum, sglen;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    sgptr = RARRAY(gem->sgary)->ptr;
    sglen = RARRAY(gem->sgary)->len;
    
    for( sgnum = 0; sgnum < sglen; ++sgnum ) {
        Data_Get_Struct( sgptr[sgnum], SupportGraph, sgraph );
        sprintf( buffer, "[SupportGraph:#%p] index=%d\n", sgraph, sgnum );
        
        graph = sgraph->buffer_graph;
        for( i = 0; i < sgraph->size_graph; ++i, ++graph ) {
            sprintf( buffer, "[ %d ] (in=%g, out=%g)\n", i, graph->inside, graph->outside );
            rb_io_write( out, rb_str_new2(buffer) );
            
            path = graph->pathary;
            for( j = 0; j < graph->pathlen; ++j, ++path ) {
                p = buffer;
                p += sprintf( p, " +-" );
                
                node = path->nodeary;
                for( k = 0; k < path->nodelen; ++k, ++node ) {
                    p += sprintf( p, " " );
                    p = node_string( p, gem, sgraph, node );
                }
                p += sprintf( p, " (in=%g)\n", path->inside );
                rb_io_write( out, rb_str_new2(buffer) );
            }
        }
    }
    
    return self;
}


/*   p[^f[^\
------------------------------------------------------------------- */

static VALUE gem_display_parameter ( VALUE self, VALUE out )
{
    GraphicalEM * gem;
    SupportGraph * sgraph;
    real_t  sum;
    double  prob, xnum;
    char  buf [ 4096 ];
    char * p;
    int  i, j, k, swptr;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    
    /* p[^̍XV */
    swptr = 0;
    for( i = 0; i < gem->swlen; ++i ) {
        k = swptr + gem->swary[i];
        
        sum = 0.0;
        for( j = swptr; j < k; ++j )
            sum += gem->xnumtbl[ j ];
        
        if( sum > 0 ) {
            for( j = swptr; j < k; ++j ) {
                xnum = gem->xnumtbl[j];
                if( xnum > 0 ) {
                    prob = xnum / sum;
                    p = buf;
                    p += sprintf( p, "%s : xtbl=%g, param=%g\n", gem_param_at(gem, j), xnum, prob );
                    rb_io_write( out, rb_str_new2(buf) );
                }
            }
        }
        swptr = k;
    }
    
    return self;
}


/********************************************************************/


/*   xOtIuWFNg̉
------------------------------------------------------------------- */
static void sgraph_free ( SupportGraph* sg )
{
    if( sg ) {
        ruby_xfree( sg->buffer_graph );
        ruby_xfree( sg->buffer_path );
        ruby_xfree( sg->buffer_node );
        ruby_xfree( sg );
    }
}

static void sgraph_mark ( SupportGraph* sg )
{
    SG_Node* node;
    int i;
    
    if( sg && !sgraph_compiled_p(sg) ) {
        node = sg->buffer_node;
        for( i = 0; i < sg->size_node; ++i ) {
            if( node->d.ind.type == 3 )
                rb_gc_mark( node->d.ind.data.obj );
            ++node;
        }
    }
}


/*   xOt̐
----------------------------------------------------------------------
SupportGraph.new ( stream, table = nil )

   xOt stream ǂݍށB
   table w肳ꂽƂ́A{m[h table Ɋi[ (table[node] = node)B
   Atable  Hash łȂ΂ȂȂB
*/
static VALUE sgraph_new ( int argc, VALUE* argv, VALUE klass )
{
    SupportGraph* sg;
    VALUE volatile obj;
    
    sg = ALLOC( SupportGraph );
    MEMZERO( sg, SupportGraph, 1 );
    obj = Data_Wrap_Struct( klass, sgraph_mark, sgraph_free, sg );
    rb_obj_call_init( obj, argc, argv );
    
    return obj;
}

struct tagSGInitdata
{
    SupportGraph * sg;
    VALUE  graphbuf; /* Array */
    int    graph_index;
    int    node_size;
    int    path_size;
    int    graph_size;
};
typedef  struct tagSGInitdata  SGInitdata;

static VALUE sgraph_initialize_itr ( VALUE line, SGInitdata* inidat )
{
    VALUE volatile v;
    VALUE volatile subgrp;
    const char *p, *q;
    int  n, index;
    
    line = (TYPE(line) != T_STRING)? rb_String(line): line;
    
    p = RSTRING(line)->ptr;
    if( *p == '[' ) { /* Ot`̐錾sc */
        ++p;
        SKIP_WS( p );
        if( !('0' <= *p && *p <= '9') )
            rb_raise( rb_eRuntimeError, "syntax error (no graph index)" );
        n = *(p++) - '0';
        while( '0' <= *p && *p <= '9' )
            n = n*10 + (*(p++) - '0');
        SKIP_WS( p );
        
        if( *p != ']' )
            rb_raise( rb_eRuntimeError, "syntax error (no ``]'')" );
        ++p;
        SKIP_WS( p );
        if( *p != '#' && *p != '\r' && *p != '\n' && *p != '\0' )
            rb_raise( rb_eRuntimeError, "syntax error (fail to parse)" );
        
        inidat->graph_index = n;
        if( n > inidat->graph_size )
            inidat->graph_size = n;
    }
    else {
        v = strparse_sgraph_path( p ); /* v: [JpX */
        if( !NIL_P( v ) ) {
            /* pX̐ݒs */
            index = inidat->graph_index;
            if( index < 0 )
                rb_raise( rb_eRuntimeError, "no graph declaration" );
            subgrp = rb_ary_entry( inidat->graphbuf, index );
            if( NIL_P(subgrp) ) {
                subgrp = rb_ary_new();
                rb_ary_store( inidat->graphbuf, index, subgrp );
            }
            rb_ary_push( subgrp, v );
            inidat->path_size += 1;
            inidat->node_size += RARRAY(v)->len;
        }
        else {
            rb_raise( rb_eRuntimeError, "syntax error (fail to parse)" );
        }
    }
    
    return Qnil;
}

static void sgraph_check_buffer ( SupportGraph* sg, SGInitdata* inidat )
{
    RArray_p  gary, pary, nary;
    int i, j, k, n;
    
    gary = RARRAY(inidat->graphbuf);
    for( i = 0; i < gary->len; ++i ) {
        if( NIL_P(gary->ptr[i]) )
            rb_raise( rb_eRuntimeError, "found uninitialized subgraph #%d", i );
        pary = RARRAY(gary->ptr[i]);
        
        for( j = 0; j < pary->len; ++j ) {
            if( NIL_P(pary->ptr[j]) )
                rb_raise( rb_eRuntimeError, "found uninitialized path #%d. (in graph #%d)", j, i );
            nary = RARRAY(pary->ptr[j]);
            
            for( k = 0; k < nary->len; ++k ) {
                switch( TYPE( nary->ptr[k] ) ) {
                case T_FIXNUM:
                    n = FIX2INT(nary->ptr[k]);
                    if( n < 0 || n >= inidat->graph_size )
                        rb_raise( rb_eRangeError, "out of range subgraph ID (in graph #%d, path #%d, node #%d)", i, j, k );
                    if( n == i )
                        rb_raise( rb_eRuntimeError, "found self reference (in graph #%d, path #%d, node #%d)", i, j, k );
                    if( n < i )
                        rb_raise( rb_eRuntimeError, "found front reference (in graph #%d, path #%d, node #%d)", i, j, k );
                    break;
                case T_STRING:
                    break;
                default:
                    rb_raise( rb_eTypeError, "[BUG/gem.so] unexpected node class. (%s)", rb_class2name(CLASS_OF(nary->ptr[k])) );
                    break;
                }
            }
        }
    }
}

static void sgraph_init_buffer ( SupportGraph* sg, SGInitdata* inidat, VALUE table )
{
    RArray_p   gary, pary, nary;
    SG_Graph*  graph;
    SG_Path*   path;
    SG_Node*   node;
    VALUE nodeval;
    int i, j, k;
    int pindex, nindex;
    
    /* obt@̏ */
    sg->buffer_graph = ALLOC_N( SG_Graph, inidat->graph_size );
    sg->buffer_path  = ALLOC_N( SG_Path, inidat->path_size );
    sg->buffer_node  = ALLOC_N( SG_Node, inidat->node_size );
    
    pindex = nindex = 0;
    
    gary = RARRAY(inidat->graphbuf);
    for( i = 0; i < gary->len; ++i ) {
        graph = &sg->buffer_graph[i];
        pary = RARRAY( gary->ptr[i] );
        
        graph->pathary = &sg->buffer_path[ pindex ];
        graph->pathlen = pary->len;
        for( j = 0; j < pary->len; ++j ) {
            path = &graph->pathary[j];
            nary = RARRAY( pary->ptr[j] );
            
            path->nodeary = &sg->buffer_node[ nindex ];
            path->nodelen = nary->len;
            for( k = 0; k < nary->len; ++k ) {
                node = &path->nodeary[k];
                nodeval = nary->ptr[k];
                
                switch( TYPE( nodeval ) ) {
                case T_FIXNUM:
                    node->d.ind.type = 2;
                    node->d.ind.data.graph = &sg->buffer_graph[ FIX2INT(nodeval) ];
                    break;
                case T_STRING:
                    node->d.ind.type = 3;
                    if( !NIL_P(table) )
                        rb_hash_aset( table, nodeval, nodeval );
                    node->d.ind.data.obj = nodeval;
                    break;
                }
            }
            nindex += nary->len;
        }
        pindex += pary->len;
    }
    
    /* obt@g悤ɂȂĂTCYw肵Ȃƃ}[Nɗ */
    sg->size_graph = inidat->graph_size;
    sg->size_path  = inidat->path_size;
    sg->size_node  = inidat->node_size;
}

static VALUE sgraph_initialize ( int argc, VALUE* argv, VALUE self )
{
    VALUE volatile stream;
    VALUE volatile table;
    VALUE volatile gbuf;
    SGInitdata inidat;
    SupportGraph* sg;
    
    /* IuWFNg̃ZbgAbv */
    Data_Get_Struct( self, SupportGraph, sg );
    
    rb_scan_args( argc, argv, "11", &stream, &table );
    if( !NIL_P(table) && TYPE(table) != T_HASH )
        rb_raise( rb_eArgError, "requires Hash for table" );
    
    gbuf = rb_ary_new();
    
    inidat.sg = sg;
    inidat.graphbuf = gbuf;
    inidat.graph_index = -1;
    inidat.graph_size = -1; /* ̎_ł͌vZׂ炷߂ -1 Ă */
    inidat.node_size = 0;
    inidat.path_size = 0;
    
    rb_iterate( rb_each, stream, sgraph_initialize_itr, (VALUE)&inidat );
    inidat.graph_size += 1; /* O 1 炵Ă̂ */
    
    /* f[^`FbNĂ */
    sgraph_check_buffer( sg, &inidat );
    
    /* obt@̍\z */
    sgraph_init_buffer( sg, &inidat, table );
    
    return self;
}


/*   xOt̏o
----------------------------------------------------------------------
sg.display ( out )

   xOt out ֏o͂B
*/
static VALUE sgraph_display ( VALUE self, VALUE out )
{
    VALUE volatile lineary;
    VALUE volatile nlobj;
    SupportGraph* sgraph;
    SG_Graph* graph;
    SG_Path* path;
    SG_Node* node;
    char buf[1024];
    char* p;
    int i, j, k;
    
    Data_Get_Struct( self, SupportGraph, sgraph );
    if( sgraph_compiled_p(sgraph) ) {
        /* RpCĂꍇ͕\łȂ */
        sprintf( buf, "<SupportGraph#%p>", sgraph );
        rb_io_write( out, rb_str_new2(buf) );
    }
    else {
        /* RpCOȂ̂ŕ\\ */
        nlobj   = rb_str_new2("\n");
        lineary = rb_ary_new2( sgraph->size_path + sgraph->size_graph );
        for( i = 0; i < sgraph->size_graph; ++i ) {
            sprintf( buf, "[ %d ]", i );
            rb_ary_push( lineary, rb_str_new2(buf) );
            
            graph = &sgraph->buffer_graph[ i ];
            for( j = 0; j < graph->pathlen; ++j ) {
                p = buf;
                sprintf( p, " +-" ); p += strlen( p );
                
                path = &graph->pathary[ j ];
                for( k = 0; k < path->nodelen; ++k ) {
                    node = &path->nodeary[ k ];
                    
                    switch( node->d.ind.type ) {
                    case 1:
                        sprintf( p, " (:%d:)", node->d.ind.data.sw );
                        break;
                    case 2:
                        sprintf( p, " @%d", node->d.ind.data.graph - sgraph->buffer_graph );
                        break;
                    case 3:
                        sprintf( p, " (:%s:)", STR2CSTR(node->d.ind.data.obj) );
                        break;
                    default:
                        sprintf( p, " (:!BROKEN!:)" );
                        break;
                    }
                    p += strlen( p );
                }
                rb_ary_push( lineary, rb_str_new2(buf) );
            }
        }
        rb_io_write( out, rb_ary_join(lineary, nlobj) );
    }
    
    return self;
}


/*   gpʂ擾
------------------------------------------------------------------- */
static VALUE sgraph_memsize ( VALUE self )
{
    SupportGraph  *sg;
    int size;
    
    Data_Get_Struct( self, SupportGraph, sg );
    size = sizeof(SupportGraph);
    size += sg->size_graph * sizeof(SG_Graph);
    size += sg->size_path * sizeof(SG_Path);
    size += sg->size_node * sizeof(SG_Node);
    
    return INT2NUM(size);
}


/*   Ot𓾂
------------------------------------------------------------------- */
static VALUE sgraph_size_graph ( VALUE self )
{
    SupportGraph  *sg;
    
    Data_Get_Struct( self, SupportGraph, sg );
    
    return INT2NUM( sg->size_graph );
}


/*   m𓾂
------------------------------------------------------------------- */
static VALUE sgraph_inside ( VALUE self, VALUE n )
{
    SupportGraph  *sg;
    int  addr;
    
    Data_Get_Struct( self, SupportGraph, sg );
    addr = NUM2INT( rb_Integer(n) );
    if( addr < 0 || addr >= sg->size_graph )
        return Qnil;
    
    return rb_float_new( sg->buffer_graph[addr].inside );
}


/*   Om𓾂
------------------------------------------------------------------- */
static VALUE sgraph_outside ( VALUE self, VALUE n )
{
    SupportGraph  *sg;
    int  addr;
    
    Data_Get_Struct( self, SupportGraph, sg );
    addr = NUM2INT( rb_Integer(n) );
    if( addr < 0 || addr >= sg->size_graph )
        return Qnil;
    
    return rb_float_new( sg->buffer_graph[addr].outside );
}


/*   vZpɃOtRpC
------------------------------------------------------------------- */
SupportGraph* sgraph_compile ( SupportGraph* sg, GraphicalEM* gem )
{
    VALUE volatile v = Qnil;
    SG_Graph* g;
    SG_Node* node;
    int i, index, swsize;
    int type;
    
    /* owner 𒲂ׂėOo */
    if( sgraph_compiled_p(sg) )
        rb_raise( rb_eRuntimeError, "support graph already compiled" );
    
    /* CfbNX  ڃN */
    swsize = gem->problen;
    node = sg->buffer_node;
    for( i = 0; i < sg->size_node; ++i ) {
        switch( node->d.ind.type ) {
        case 1: /* {m[h */
            index = node->d.ind.data.sw;
            #if PRINT_SGRAPH_COMPILE
            printf( "Basic(INDEX): %d\n", index );
            #endif
            if( index < 0 || index >= swsize )
                rb_raise( rb_eRangeError, "out of range switch index (%d for 0...%d)", index, swsize );
            node->d.dir.i_ptr = &gem->probtbl[ index ];
            node->d.dir.o_ptr = &gem->xnumtbl[ index ];
            node->d.dir.o_inside = &sg->buffer_graph[0].inside;
            break;
        case 2: /* Qƃm[h */
            g = node->d.ind.data.graph;
            #if PRINT_SGRAPH_COMPILE
            printf( "Ref: %d\n", g - sg->buffer_graph );
            #endif
            if( !g )
                rb_raise( rb_eRuntimeError, "referenced subgraph is NULL" );
            node->d.dir.i_ptr = &g->inside;
            node->d.dir.o_ptr = &g->outside;
            node->d.dir.o_inside = &g->inside;
            break;
        case 3: /* {m[h () */
            v = rb_hash_aref( gem->paramtbl, node->d.ind.data.obj );
            if( NIL_P(v) )
                rb_raise( rb_eRuntimeError, "unknown switch string (%s)", STR2CSTR(node->d.ind.data.obj) );
            index = NUM2INT(v);
            #if PRINT_SGRAPH_COMPILE
            printf( "Basic(OBJECT): %d (%s)\n", index, STR2CSTR(node->d.ind.data.obj) );
            #endif
            #if PRINT_READ_PARAM
            fprintf( stdout, "sgraph_compile(): gem->paramtbl[%s] = %d\n", STR2CSTR(node->d.ind.data.obj), index );
            #endif
            if( index < 0 || index >= swsize )
                rb_raise( rb_eRangeError, "out of range Index of switch string (%d for 0...%d)", index, swsize );
            node->d.dir.i_ptr = &gem->probtbl[ index ];
            node->d.dir.o_ptr = &gem->xnumtbl[ index ];
            node->d.dir.o_inside = &sg->buffer_graph[0].inside;
            break;
        default:
            rb_raise( rb_eTypeError, "detected broken support graph" );
        }
        ++node;
    }
    sg->owner_gem = gem;
    
    return sg;
}


/********************************************************************/


/*   ̃W[̃Gg|Cg
------------------------------------------------------------------- */
void Init_gem ( void )
{
    /* Ext W[̒`: gW[Œ`IuWFNgׂ͂ Ext W[ɒ` */
    rb_mExt = rb_define_module( "Ext" );
    
    /* NX` */
    rb_cSG = rb_define_class_under( rb_mExt, "SupportGraph", rb_cData );
    rb_cGEM = rb_define_class_under( rb_mExt, "GraphicalEM", rb_cData );
    
    id_store = rb_intern( "store" );
    
    /* (GraphicalEM): NX\bh̒` */
    rb_define_singleton_method( rb_cGEM, "new", gem_new, 1 );
    
    /* (SupportGraph): NX\bh̒` */
    rb_define_singleton_method( rb_cSG, "new", sgraph_new, -1 );
    
    /* (GraphicalEM): \bh̒` */
    rb_define_method( rb_cGEM, "initialize", gem_initialize, 1 );
    rb_define_method( rb_cGEM, "store", gem_store, 1 );
    rb_define_method( rb_cGEM, "<<", gem_store, 1 );
    rb_define_method( rb_cGEM, "sglist", gem_sglist, 0 );
    rb_define_method( rb_cGEM, "to_a", gem_sglist, 0 );
    rb_define_method( rb_cGEM, "display", gem_display, 1 );
    rb_define_method( rb_cGEM, "size", gem_size, 0 );
    rb_define_method( rb_cGEM, "likelihood", gem_likelihood_m, 0 );
    rb_define_method( rb_cGEM, "update", gem_update_m, 0 );
    rb_define_method( rb_cGEM, "display_inside", gem_display_inside, 1 );
    rb_define_method( rb_cGEM, "display_sg", gem_display_sg, 1 );
    rb_define_method( rb_cGEM, "display_parameter", gem_display_parameter, 1 );
    
    /* (SupportGraph): \bh̒` */
    rb_define_method( rb_cSG, "initialize", sgraph_initialize, -1 );
    rb_define_method( rb_cSG, "display", sgraph_display, 1 );
    rb_define_method( rb_cSG, "memsize", sgraph_memsize, 0 );
    rb_define_method( rb_cSG, "size_graph", sgraph_size_graph, 0 );
    rb_define_method( rb_cSG, "length", sgraph_size_graph, 0 );
    rb_define_method( rb_cSG, "size", sgraph_size_graph, 0 );
    rb_define_method( rb_cSG, "inside", sgraph_inside, 1 );
    rb_define_method( rb_cSG, "outside", sgraph_outside, 1 );
}


