/*                                                     project Gemini 
----------------------------------------------------------------------
    `[g\
      $Id: chart.c,v 1.26 2002/11/22 03:00:03 abe Exp $

       Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define DEBUGSW 0
#define DEBUG_FREE  0  /**< IuWFNgɊւfobOx */
#define TRACK_CHARTFREE 0
#include "nlputil.h"


/*   O[oϐ
------------------------------------------------------------------- */
VALUE rb_cChart;
VALUE rb_cChart_Element;


/*   Ԋ֌W
------------------------------------------------------------------- */

State* state_new ( State* ptr, Rule* rule, int left )
{
    if( !ptr )
        ptr = ALLOC(State);
    
    MEMZERO( ptr, State, 1 );
    if( !rule )
        rb_raise( rb_eRuntimeError, "state requires rule" );
    ptr->rule = rule;
    ptr->left = left;
    ptr->right = left;
    ptr->dot = 0;
    ptr->top = rule->lhs;
    
    return ptr;
}

State* state_new2 ( State* ptr, int left, int right, int cat )
{
    if( !ptr )
        ptr = ALLOC(State);
    
    MEMZERO( ptr, State, 1 );
    ptr->left = left;
    ptr->right = right;
    ptr->top = cat;
    ptr->dot = 1;
    
    return ptr;
}

State* state_new3 ( State* ptr, State* source, State* part )
{
    if( !ptr )
        ptr = ALLOC(State);
    
    MEMZERO( ptr, State, 1 );
    ptr->rule = source->rule;
    ptr->left = source->left;
    ptr->right = part->right;
    ptr->dot = source->dot + 1;
    ptr->top = source->top;
    ptr->next_dot = NULL;
    ptr->next_comp = NULL;
    
    return ptr;
}

State* state_dup ( State* ptr, State const * source )
{
    if( !ptr )
        ptr = ALLOC(State);
    
    *ptr = *source;
    
    return ptr;
}

static int int_cmp ( int i1, int i2 )
{
    return (i1 == i2)? 0: ((i1 < i2)? -1: 1);
}

int state_cmp ( State const * s1, State const * s2 )
{
    int c;
    
    if( s1->rule != s2->rule ) {
        if( !s1->rule )
            return 1;
        if( !s2->rule )
            return -1;
        return int_cmp(s1->rule->id, s2->rule->id);
    }
    
    c = int_cmp( s1->top, s2->top );
    if( c != 0 )
        return c;
    
    c = int_cmp( s1->left, s2->left );
    if( c != 0 )
        return c;
    
    c = int_cmp( s1->right, s2->right );
    if( c != 0 )
        return c;
    
    c = int_cmp( s1->dot, s2->dot );
    if( c != 0 )
        return c;
    
    return 0;
}

char* state_inspect ( State const * state, Grammar const * grammar )
{
    static char buffer [ 4092 ];
    char* ptr = buffer;
    
    if( state->rule ) {
        ptr += sprintf( ptr, "(%d,%d)[", state->left, state->right );
        ptr = rule_typestring_internal( state->rule, ptr, state->dot );
        ptr += sprintf( ptr, "]" );
    }
    else {
        ptr += sprintf( ptr, "(%d,%d)", state->left, state->right );
        if( grammar ) {
            ptr += sprintf( ptr, "[%s]", cfg_getcatname(grammar, state->top) );
        }
        else {
            ptr += sprintf( ptr, "[#%d]", state->top );
        }
    }
    
    return buffer;
}

void state_free ( State* state )
{
    if( state ) {
        ruby_xfree( state );
    }
}


/*  ⏕֐
----------------------------------------------------------------------
   chart_clear_...()  Item ̃N܂ŏȂ̂ŒӁB
*/

static int chart_offset ( Chart* chart )
{
    /*  ID gƂɂʒu␳ (chart->hash_lwait, rwait ւ̃ANZXŎgpItZbgl) */
    return (chart->preterm - 1) * chart->len; /* -1: ID 0  nonterm ƋL邽 */
}

static Chart * chart_clear_hash ( Chart * chart )
{
    int offset = chart_offset( chart );
    
    /* nbV݂̂폜̂ hash_dot ɂ͎GȂ
       (ׂĂ̍f[^Ă̂ŏςȂƂɂȂI) */
    MEMZERO( chart->hash_comp, State_p, chart->len * chart->len );
    MEMZERO( chart->hash_lwait - offset, State_p, chart->len * chart->symbol );
    MEMZERO( chart->hash_rwait - offset, State_p, chart->len * chart->symbol );
    
    return chart;
}

static Chart * chart_clear_item ( Chart * chart )
{
    chart_clear_hash( chart );
    
    MEMZERO( chart->hash_dot, State_p, chart->len );
    
    return  chart;
}



/* `[g\̃RXgN^EfXgN^
------------------------------------------------------------------- */

void chart_free_internal ( Chart* chart )
{
    /* chart g͍폜Ȃo[W (GC ō폜邽) */
    State_p  p, q;
    int  i, offset;
    
    if( chart && chart->hash_dot ) {
        /* ܂sĂȂꍇ̂ݎs */
        #if DEBUG_FREE > 0
        fprintf( stderr, "chart_free_internal(): <%p>\n", chart );
        #endif
        for( i = chart->left; i <= chart->right; ++i ) {
            for( p = chart_dot( chart, i ); p; p = q ) {
                q = p->next_dot;
                state_free( p );
            }
        }
        
        offset = chart_offset( chart );
        ruby_xfree( chart->hash_dot );
        ruby_xfree( chart->hash_lwait - offset ); /* ItZbg߂Kv */
        ruby_xfree( chart->hash_rwait - offset );
        ruby_xfree( chart->hash_comp );
        
        MEMZERO( chart, Chart, 1 ); /* Sf[^폜 */
    }
}

void chart_free ( Chart* ptr )
{
    #if DEBUG_FREE > 0
    fprintf( stderr, "[FREE] chart_free(): begin.\n" );
    #endif
    if( ptr ) {
        chart_free_internal( ptr );
        ruby_xfree( ptr );
    }
    #if DEBUG_FREE > 0
    fprintf( stderr, "[FREE] chart_free(): end.\n" );
    #endif
}

void chart_mark ( Chart* chart )
{
    if( chart ) {
        if( chart->grammar )
            rb_gc_mark( chart->grammar->ruby_object );
        rb_gc_mark( chart->start );
    }
}

static Chart* chart_init ( Chart* chart, VALUE start, Grammar* grammar, int left, int right )
{
    int  offset;
    
    chart->grammar = grammar;
    chart->start = start;
    
    chart->left = left;
    chart->right = right;
    chart->len = right - left + 1;
    
    /* { grammar ̕ɎׂȂ̂c (bN@\΂) */
    chart->nonterm = grammar->size_nt; /* preterm  ID 0  nonterm ̕Ɏ */
    chart->preterm = grammar->size_pt;
    chart->symbol  = chart->nonterm + chart->preterm - 1; /* -1: ID 0  nonterm ƋL邽 */
    offset = chart_offset( chart );
    
    chart->hash_dot   = rb_alloc_zero( State_p, chart->len );
    chart->hash_comp  = rb_alloc_zero( State_p, chart->len * chart->len );
    chart->hash_rwait = rb_alloc_zero( State_p, chart->len * chart->symbol ) + offset;
    chart->hash_lwait = rb_alloc_zero( State_p, chart->len * chart->symbol ) + offset;
    
    return chart;
}

VALUE chart_new ( VALUE start, Grammar* grammar, int left, int right )
{
    VALUE volatile chartobj;
    Chart* chart;
    
    /* f[^𑵂 */
    chart = rb_alloc_zero(Chart, 1);
    chartobj = Data_Wrap_Struct( rb_cChart, chart_mark, chart_free, chart );
    chart->self_object = chartobj;
    chart_init( chart, start, grammar, left, right );
    
    return chartobj;
}


/* `[gɍǉ
------------------------------------------------------------------- */

static void chart_addhash ( Chart * chart, State_p state )
{
    Rule_p rule = state->rule;
    State_p *ptrp;
    int  wait;
    
    if( state_complete_p(state) ) { /* S */
        ptrp = chart_comp_p( chart, state->left, state->right );
        state->next_comp = *ptrp;
        *ptrp = state;
    }
    else { /*  */
        wait = (rule)? rule->rhs[ state->dot ]: state->top;
        
        ptrp = chart_rwait_p( chart, wait, state->right );
        state->next_rwait = *ptrp;
        *ptrp = state;
        
        ptrp = chart_lwait_p( chart, wait, state->left );
        state->next_lwait = *ptrp;
        *ptrp = state;
    }
}

State* chart_add_state ( Chart * chart, State* state )
{
    State_p *ptrp;
    
    /* Sԋ */
    ptrp = chart_dot_p( chart, state->right );
    state->next_dot = *ptrp;
    *ptrp = state;
    
    chart_addhash( chart, state );
    
    return state;
}

State* chart_add_state2 ( Chart* chart, State const * state )
{
    State* s;
    
    s = chart_find( chart, state );
    if( s )
        return s;
    
    s = state_dup( NULL, state );
    
    return chart_add_state( chart, s );
}

State_p chart_find_state ( Chart * chart, Rule const * rule, int left, int right, int dot )
{
    State * p;
    int  waiting;
    
    if( rule->rhslen == dot ) {
        for( p = chart_comp( chart, left, right ); p; p = p->next_comp ) {
            /* rule Ȃ dot rKv͂Ȃ */
            if( rule == p->rule )
                return p;
        }
    }
    else {
        waiting = rule->rhs[dot];
        for( p = chart_lwait( chart, waiting, left ); p; p = p->next_lwait ) {
            /* lwait ŌĂ̂ left ͕K */
            if( rule == p->rule && right == p->right && dot == p->dot )
                return p;
        }
    }
    
    return NULL;
}

State_p chart_find_terminal ( Chart * chart, int left, int right, int top )
{
    State * p;
    
    for( p = chart_comp( chart, left, right ); p; p = p->next_comp ) {
        if( p->rule == NULL && top == p->top )
            return p;
    }
    return NULL;
}

State_p chart_find ( Chart const * chart, State const * state )
{
    Rule_p rule  = state->rule;
    int    left  = state->left;
    int    right = state->right;
    int    dot   = state->dot;
    int    top   = state->top;
    int waiting;
    State_p p;

    /* ʏ Earley  */
    if( state_complete_p(state) ) { /*  */
        for( p = chart_comp( chart, left, right ); p; p = p->next_comp )
        {
            /* rule Ȃ dot rKv͂Ȃ */
            if( rule == p->rule && top == p->top )
                return p;
        }
    }
    else { /*  */
        waiting = (rule)? rule->rhs[dot]: top;
        for( p = chart_lwait( chart, waiting, left ); p; p = p->next_lwait )
        {
            /* lwait ŌĂ̂ left ͕K */
            if( rule == p->rule && right == p->right && dot == p->dot )
                return p;
        }
    }
    
    return NULL;
}


/*   \؂Ă邩𒲂ׂ
------------------------------------------------------------------- */

VALUE chart_exist_p ( VALUE self )
{
    Chart*  chart;
    State_p  p;
    int  start;
    int  itemcount = 0;
    
    Data_Get_Struct( self, Chart, chart );
    start = category_int(chart->start);
    if( start == 0 )
        rb_raise( rb_eTypeError, "invalid start symbol in chart" );
    
    /* JnLgbvɎSJo[鍀΂ OK */
    for( p = chart_comp( chart, chart->left, chart->right ); p; p = p->next_comp ) {
        if( p->top != start )
            continue;
        ++itemcount;
    }
    
    return (itemcount == 0)? Qfalse: Qtrue;
}


/*  `[g\̍𓾂
------------------------------------------------------------------- */

int chart_count_item ( Chart * chart )
{
    State_p  item;
    int  i, count;
    
    if( chart->count_item == 0 ) {
        count = 0;
        for( i = chart->left; i <= chart->right; ++i ) {
            for( item = chart_dot( chart, i ); item; item = item->next_dot )
                ++count;
        }
        chart->count_item = count;
    }
    
    return chart->count_item;
}

static VALUE chart_count_item_m ( VALUE self )
{
    Chart*  chart;
    
    Data_Get_Struct( self, Chart, chart );
    
    return rb_int_new(chart_count_item(chart));
}


/*  OmvZɍ폜ꂽ̐𓾂
------------------------------------------------------------------- */
static  VALUE  chart_count_deleted  ( VALUE self )
{
    Chart* chart;
    
    Data_Get_Struct( self, Chart, chart );
    
    return rb_int_new(chart->count_deleted);
}


/*  `[gɓo^悤ƂĎꂽ̐𓾂
------------------------------------------------------------------- */
static  VALUE  chart_count_cancelled  ( VALUE self )
{
    Chart* chart;
    
    Data_Get_Struct( self, Chart, chart );
    
    return INT2NUM(chart->count_cancelled);
}


/* `[gmۂĂ郁̏𓾂
------------------------------------------------------------------- */
int chart_memsize ( Chart * chart )
{
    int count = chart_count_item( chart );
    int len, sym, total;
    
    len = chart->len;
    sym = chart->symbol;
    
    total = sizeof(Chart);
    total += sizeof(State) * count;
    total += sizeof(State_p) * len;       /* hash_dot */
    total += sizeof(State_p) * len * len; /* hash_comp */
    total += sizeof(State_p) * len * sym; /* hash_rwait */
    total += sizeof(State_p) * len * sym; /* hash_lwait */
    
    return total;
}

static VALUE chart_memsize_m ( VALUE self )
{
    Chart* chart;
    
    Data_Get_Struct( self, Chart, chart );
    
    return INT2NUM(chart_memsize(chart));
}


/*   `[gf[^Ƃďo͂
------------------------------------------------------------------- */

static VALUE chart_display ( VALUE self, VALUE out )
{
    Chart* chart;
    State_p p;
    int i;
    
    Data_Get_Struct( self, Chart, chart );
    
    /* ɏo͂ (₷邽) */
    for( i = 0; i < chart->len; ++i ) {
        for( p = chart_dot( chart, i ); p; p = p->next_dot ) {
            if( p->dot >= p->rule->rhslen ) {
                rb_io_write( out, item_inspect( p ) );
                rb_io_write( out, rb_default_rs );
            }
        }
    }
    for( i = 0; i < chart->len; ++i ) {
        for( p = chart_dot( chart, i ); p; p = p->next_dot ) {
            /* hbgʒu 0 ̂̂͏o͂Ȃ悤ɂ */
            if( p->dot > 0 && p->dot < p->rule->rhslen ) {
                rb_io_write( out, item_inspect( p ) );
                rb_io_write( out, rb_default_rs );
            }
        }
    }
    
    return self;
}


/*   fobOpɃ`[g\
------------------------------------------------------------------- */

void chart_debug_display ( FILE* out, Chart const * chart )
{
    Grammar* grammar = chart->grammar;
    State_p state;
    int i;
    
    fprintf( out, "[Chart] >>>>>>>> <%p>\n", chart );
    for( i = chart->left; i <= chart->right; ++i ) {
        for( state = chart_dot( chart, i ); state; state = state->next_dot ) {
            fprintf( out, "%s\n", state_inspect(state, grammar) );
        }
    }
    fprintf( out, "[Chart] <<<<<<<< <%p>\n", chart );
}


/*   \͌ʂ̕\Ɏgpf[^\
------------------------------------------------------------------- */

/** @brief \͌ʍ\zpf[^
 *
 * \͌ʂ\zƂ̂߂Ɏgpf[^\łB
 */
struct  tagExtractData
{
    /** @brief @
     *
     * gpꂽ@IuWFNg܂B
     */
    Grammar*  grammar;
    
    /** @brief `[g
     *
     * 񂷂`[gIuWFNg܂B
     */
    Chart*  chart;
    
    /** @brief ԃX^bN
     *
     * \͌ʂ܂܂B
     * \ɊւĂ extract_visit_item() QƂĂB
     */
    State_p  *ptr;
    
    /** @brief X^bN
     *
     * X^bÑf[^łB
     */
    int  len;
    
    /** @brief X^bNe
     *
     * X^bNmۂĂ郁̃TCYłB
     * [U[̕ϐύXĂ͂Ȃ܂B
     */
    int  cap;
    
    /** @brief K ID 环ʃCfNXɕϊ邽߂̃}bv
     *
     * K ID A\͌ʂŎgpKʗp̃CfNXɕϊ邽߂̃e[ułB
     * 傫 cfg_size_rule(grammar) ƓȂ܂B
     */
    int  *rulemap;
    
    /** @brief ʃCfNXKIuWFNgւ̃}bv
     *
     * \͌ʂŎgpKʗp̃CfNXKIuWFNgɕϊ邽߂̃e[ułB
     */
    Rule_p  *rulelist;
    
    /** @brief ʃCfNX
     *
     * rulelist ̃TCY܂B
     */
    int  rulelist_size;
};
typedef  struct tagExtractData  ExtractData;

struct tagExtractList
{
    struct tagExtractList* next;
    State* state;
};
typedef struct tagExtractList ExtractList;

static void extract_visit_item_terminal ( ExtractData* data, State* state );
static void extract_visit_item_finished ( ExtractData* data, State* state, ExtractList* prev );
static void extract_visit_item_underway ( ExtractData* data, State* state, ExtractList* prev );


static void extract_data_free ( ExtractData * dat )
{
    if( dat ) {
        ruby_xfree( dat->ptr );
        ruby_xfree( dat->rulemap );
        ruby_xfree( dat->rulelist );
        ruby_xfree( dat );
    }
}


/*   `[g̏
------------------------------------------------------------------- */

static void resize_itemstack ( ExtractData* data, int newlen )
{
    if( data->cap <= newlen ) {
        while( data->cap <= newlen )
            data->cap = (data->cap < 1024)? 1024: data->cap * 2;
        REALLOC_N( data->ptr, State_p, data->cap );
    }
}

static void expand_itemstack ( ExtractData* data, int expandlen )
{
    resize_itemstack( data, data->len + expandlen );
}

/** @brief `[g̏
 *
 * state Ŏw肳ꂽԈȉ񂵁A\͌ʂ data ɒ~ς܂B
 *
 * ̊֐ł state ̖ɂL܂B
 *
 * data->ptr ɂ́Â悤ɂďԂL^Ă܂B
 *  - hbgʒu 0 ̏ (X --> * A B C)
 *    - hbgʒu 1 ₵ĂE[܂ (X --> A B C *)
 *  - [ (hbgʒu͕K 1)
 *
 * @param data [i/o] L^
 * @param state [in] 
 * @param prev [in] Ăяoɓnꂽf[^
 */
static void extract_visit_item ( ExtractData * data, State * state, ExtractList * prev )
{
    if( state->tag & TAG_Visited )
        return; /* łɖKς݂ȂΖⓚp */
    
    /* I[L͓̍ʈ */
    if( !state->rule ) {
        extract_visit_item_terminal( data, state );
    }
    else if( state->dot == 0 ) {
        extract_visit_item_finished( data, state, prev );
    }
    else {
        extract_visit_item_underway( data, state, prev );
    }
}

static void extract_visit_item_terminal ( ExtractData* data, State* state )
{
    /* state->rule == NULL ͕ۏႳĂ */
    
    /* ݂Ă͂ȂȂ̂ǁc݂ĂT͈ӂȂ̂ōsӖȂ */
    if( state->dot == 0 )
        return;
    
    state->tag |= TAG_Visited; /* ł */
    expand_itemstack( data, 1 );
    
    /* ނ̂ --> a *  */
    data->ptr[ data->len++ ] = state;
}

static void extract_visit_item_finished ( ExtractData* data, State* state, ExtractList* prev )
{
    /* state->rule != NULL ͕ۏႳĂ */
    State_p  ptr;
    Rule_p  rule = state->rule;
    int  dot = 0;
    
    /* ׂ̉͂ďÎŃX^bNɏԂ̃Xg */
    expand_itemstack( data, 1 + state->rule->rhslen );
    
    /* X --> * A B C  X --> A * B C, X --> A B * C, ... Əł */
    data->ptr[ data->len++ ] = state;
    for( ; prev; prev = prev->next ) {
        ++dot;
        
        ptr = prev->state;
        if( ptr->rule != rule )
            rb_raise( rb_eRuntimeError, "broken ExtractList chain (rule) probed" );
        if( ptr->dot != dot )
            rb_raise( rb_eRuntimeError, "broken ExtractList chain (dot) probed" );
        
        data->ptr[ data->len++ ] = ptr;
    }
    if( dot != rule->rhslen )
        rb_raise( rb_eRuntimeError, "too short ExtractList chain probed" );
}

static void extract_visit_item_underway ( ExtractData* data, State* state, ExtractList* prev )
{
    /* state->rule != NULL ͕ۏႳĂ */
    
    ExtractList  unit;
    
    Chart* const chart = data->chart;
    Rule const * rule = state->rule;
    
    State_p  i, j;
    int  right, wantdot, wantcat;
    int  processed;
    
    if( state->dot >= rule->rhslen )
        state->tag |= TAG_Visited;  /* ɂ͏dď񂵂Ȃ悤ɔԕu */
    
    unit.next = prev;
    unit.state = state;
    
    wantdot = state->dot - 1;
    wantcat = rule->rhs[ wantdot ];
    right = state->right;
    for( i = chart_lwait( chart, wantcat, state->left ); i; i = i->next_lwait )
    {
        if( i->right >= right || i->rule != rule || i->dot != wantdot )
            continue;
        /* ̎_ i  state ̖hbgЂƂԂƂȂ */
        
        if( wantcat >= 0 ) {
            /* I[L҂Ăꍇ̂ݖLɊւďs */
            processed = 0;
            for( j = chart_comp( chart, i->right, right ); j; j = j->next_comp ) {
                if( j->top != wantcat )
                    continue;
                extract_visit_item( data, j, NULL );  /* LɊւď񂷂 */
                ++processed;
            }
            if( processed == 0 )
                continue; /* \zɎgȂ̂̏ꍇ̓LZ */
        }
        else {
            /* I[L҂Ăꍇ͐Ă邩𒲂ׂ */
            if( i->right+1 != right )
                continue;
        }
        
        extract_visit_item( data, i, &unit ); /* c񂷂 */
    }
}


/*  \͌ʂ\
----------------------------------------------------------------------
    X^bN̓X^bNq[vɂƂB(݂͕ύXĂH)
    ŐV̕擪ɂȂB
    (l,r) X --> A B * C D
    ƂAl n܂ B ҂B
    AKƒĂ͈͂ŁAhbgʒu 1 Ȃ̂B
    
    xOt쐬Ƀ\[ĝō̏͊֌WȂB
    
    obt@̍\
     [0] : X --> * A B ...
     [1] : X --> A * B ...
     ...
*/

/** @brief `[g̏ƃX^bN̍\z
 *
 * chart œnꂽ`[g񂵁A\͌ʂ̂߂̃f[^\z܂B
 *
 * @param chart [in] 񂷂`[g
 * @param data [out] \͌ʂރf[^
 */
static void chart_display_itemset_init ( ExtractData* data, Chart* chart )
{
    /* `[g̏ */
    State_p state;
    int i, start;
    
    start = category_int( chart->start );
    if( start == 0 )
        rb_raise(rb_eTypeError, "chart has invalid start symbol");
    
    data->chart = chart;
    data->grammar = chart->grammar;
    for( state = chart_comp( chart, chart->left, chart->right ); state; state = state->next_comp ) {
        if( state->top != start )
            continue;
        extract_visit_item( data, state, NULL );
    }
}

static void extract_build_rule_table ( ExtractData* data )
{
    const int rulesize = cfg_size_rule(data->grammar);
    State_p  state, *stateptr;
    Rule_p  rule, *rulelist;
    int*  rulemap;
    int  id, len;
    
    /* eKɔԍUāc */
    data->rulemap = ALLOC_N( int, rulesize ); /* rulemap: rule-id --> no. (>= 1) */
    rulemap = data->rulemap;
    
    MEMZERO( rulemap, int, rulesize );
    len = 1;
    stateptr = data->ptr;
    while( (state = *stateptr) ) {
        rule = state->rule;
        if( rule ) {
            id = rule->id;
            if( rulemap[id] == 0 )
                rulemap[id] = len++;
            stateptr += 1 + rule->rhslen;
        }
        else
            ++stateptr;
    }
    
    /* KIuWFNgւ̃Nɕϊāc */
    data->rulelist_size = len;
    data->rulelist = ALLOC_N( Rule_p, data->rulelist_size );
    rulelist = data->rulelist;
    
    stateptr = data->ptr;
    while( (state = *stateptr) ) {
        rule = state->rule;
        if( rule ) {
            rulelist[ rulemap[rule->id] ] = rule;
            stateptr += 1 + state->rule->rhslen;
        }
        else
            ++stateptr;
    }
}

static void chart_display_itemset_build_output_rules ( ExtractData* data, VALUE out )
{
    Rule_p  *rulelist = data->rulelist;
    char  buf [ 4096 ];
    char *p;
    int  i;
    
    for( i = 1; i < data->rulelist_size; ++i ) {
        p = buf;
        p += sprintf( p, "#%-5d ", i );
        p = rule_typestring_internal( rulelist[i], p, -1 );
        p += sprintf( p, "\n" );
        rb_io_write( out, rb_str_new2( buf ) );
    }
}

static void chart_display_itemset_build_output_result ( ExtractData* data, VALUE out )
{
    State_p  state, *stateptr;
    Rule_p  rule;
    char  buffer [ 4096 ];
    char *p;
    int* rulemap = data->rulemap;
    int i, rhslen;
    
    stateptr = data->ptr;
    while( (state = *stateptr) ) {
        rule = state->rule;
        if( rule ) {
            rhslen = rule->rhslen;
            p = buffer;
            p += sprintf( p, "[%d", state->left );
            state = *(++stateptr);
            for( i = 0; i < rhslen; ++i ) {
                p += sprintf( p, ",%d", state->right );
                state = *(++stateptr);
            }
            p += sprintf( p, "] %d\n", rulemap[rule->id] );
            rb_io_write( out, rb_str_new2( buffer ) );
        }
        else {
            /* $$$ ͏o͂悤 */
            state = *(++stateptr);
        }
    }
}

static void chart_display_itemset_build ( ExtractData* data, VALUE out )
{
    /* obt@ċK}bv쐬 */
    extract_build_rule_table( data );
    
    /* o͂镶쐬 */
    chart_display_itemset_build_output_rules( data, out );
    
    /* \͌ʂ̏o */
    chart_display_itemset_build_output_result( data, out );
}

VALUE chart_display_itemset ( VALUE self, VALUE io )
{
    VALUE volatile data_obj;
    ExtractData*  data;
    Chart*  chart;
    
    /* f[^Ă */
    data = ALLOC_ZERO_N( ExtractData, 1 );
    data_obj = Data_Wrap_Struct( rb_cData, NULL, extract_data_free, data );
    Data_Get_Struct( self, Chart, chart );
    
    /* `[g̏ƃX^bN̍\z (data->ptr \z) sāc */
    chart_display_itemset_init( data, chart );
    expand_itemstack( data, 1 );
    data->ptr[ data->len ] = NULL;
    
    /* o͂f[^쐬āc */
    chart_display_itemset_build( data, io );
    
    return io;
}


/*   ؂̃gbvׂČ
----------------------------------------------------------------------
 - Item::data j󂳂̂ŒӁB
   Item::data ͕؂ɎgĂ邱Ƃ߂̃}[NB( NULL ̂Ƃ)
*/
static void chart_mark_children ( Chart* chart, State_p item )
{
    /* ڂ̎q (Completion Ŋւ̍) }[NȂ */
    Rule_p rule = item->rule;
    State_p head, tail;
    int wantdot, wantcat;
    int used, found;
    
    if( !rule || item->dot == 0 ) {
        item->data = item;
        item->tag &= ~TAG_TopItem;
        return; /* [ȂΕԂ */
    }
    wantdot = item->dot - 1;
    wantcat = rule->rhs[ wantdot ];
    
    used = 0;
    for( head = chart_lwait( chart, wantcat, item->left ); head; head = head->next_lwait ) {
        if( head->right >= item->right || head->rule != rule || head->dot != wantdot )
            continue;
        /* ł悤₭ head  item ̖hbgЂƂ߂ */
        ++used;
        
        if( wantcat > 0 ) {
            found = 0;
            for( tail = chart_comp( chart, head->right, item->right ); tail; tail = tail->next_comp ) {
                if( tail->top != wantcat )
                    continue;
                tail->data = item;
                tail->tag &= ~TAG_TopItem;
                ++found;
            }
            if( found == 0 )
                continue; /*  head  item \z邽߂ head łȂ */
        }
        else {
            /* I[L҂Ă邩H */
            if( head->right + 1 != item->right )
                continue;
        }
        head->data = item;
        head->tag &= ~TAG_TopItem;
    }
    
    /* item ؂`ĂƂ */
    if( used > 0 )
        item->data = item;
}

Chart* chart_mark_top ( Chart* chart )
{
    State_p item;
    int dot;
    
    /* ܂ׂ͂Ă̍𖢏ɐݒ肷 */
    for( dot = chart->left; dot <= chart->right; ++dot ) {
        for( item = chart_dot( chart, dot ); item; item = item->next_dot ) {
            item->data = NULL;
            if( item->dot > 0 )
                item->tag |= TAG_TopItem;
        }
    }
    
    /* ׂĂ̍ɂĎq}[N (ʂƂăgbvc) */
    for( dot = chart->left; dot <= chart->right; ++dot ) {
        for( item = chart_dot( chart, dot ); item; item = item->next_dot ) {
            chart_mark_children( chart, item );
        }
    }
    
    return chart;
}

Chart * chart_mark_start_state ( Chart * chart )
{
    State_p state;
    int start;
    
    start = category_int( chart->start );
    for( state = chart_comp( chart, chart->left, chart->right ); state; state = state->next_comp ) {
        if( state->top == start )
            state->tag |= TAG_TopItem;
    }
    
    return chart;
}


/*   ؃Xg̗vf
----------------------------------------------------------------------
elem.rule
   vfĂ镶@K𓾂 (ꍇ nil Ԃ)

elem.top
   vfɑΉ郉x̋L𓾂

elem.left, elem.right
   vfɑΉ郉xĂ͈͂𓾂

elem.dot
   qƂȂ镔؂̂ǂ܂ŉ͂ił邩

elem.complete?
   ؂ׂĂ܂Ă邩𒲂ׂ

elem.children
   qƂȂ镔؂̃XgW߂zB
   children[n] ... n ʂڂ̃Xg
*/

static void chart_element_mark ( ChartElement* element )
{
    if( element ) {
        if( element->rule ) {
            rb_gc_mark( element->rule->gramobj );
        }
        rb_gc_mark( element->children );
    }
}

static void chart_element_free ( ChartElement* element )
{
    if( element ) {
        ruby_xfree( element );
    }
}

static VALUE chart_element_init ( ChartElement* elem, Chart* chart, State_p item )
{
    VALUE volatile ary = Qnil;
    
    if( !elem )
        elem = ALLOC(ChartElement);
    
    elem->chart = chart;
    elem->rule = item->rule;
    elem->top = item->top;
    elem->dot = item->dot;
    elem->left = item->left;
    elem->right = item->right;
    if( item->rule ) {
        ary = rb_ary_new();
        elem->children = ary;
    }
    else {
        elem->children = Qnil;
    }
    
    return Data_Wrap_Struct( rb_cChart_Element, chart_element_mark, chart_element_free, elem );
}

static VALUE chart_element_init2 ( ChartElement* elem, Chart* chart, int top, int left, int right )
{
    if( !elem )
        elem = ALLOC(ChartElement);
    
    elem->chart = chart;
    elem->rule = NULL;
    elem->top = top;
    elem->dot = 0;
    elem->left = left;
    elem->right = right;
    elem->children = Qnil;
    
    return Data_Wrap_Struct( rb_cChart_Element, chart_element_mark, chart_element_free, elem );
}

static VALUE chart_element_get_rule ( VALUE self )
{
    ChartElement* elem;
    
    Data_Get_Struct( self, ChartElement, elem );
    
    if( !elem->rule )
        return Qnil;
    
    return elem->rule->ruleobj;
}

static VALUE chart_element_get_left ( VALUE self )
{
    ChartElement* elem;
    
    Data_Get_Struct( self, ChartElement, elem );
    
    return INT2FIX(elem->left);
}

static VALUE chart_element_get_right ( VALUE self )
{
    ChartElement* elem;
    
    Data_Get_Struct( self, ChartElement, elem );
    
    return INT2FIX(elem->right);
}

static VALUE chart_element_get_dot ( VALUE self )
{
    ChartElement* elem;
    
    Data_Get_Struct( self, ChartElement, elem );
    
    return INT2FIX(elem->dot);
}

static VALUE chart_element_get_top ( VALUE self )
{
    ChartElement* elem;
    
    Data_Get_Struct( self, ChartElement, elem );
    
    return cfg_get_category( elem->chart->grammar, elem->top );
}

static VALUE chart_element_complete_p ( VALUE self )
{
    ChartElement* elem;
    
    Data_Get_Struct( self, ChartElement, elem );
    
    if( !elem->rule )
        return Qtrue; /* [͊ */
    
    return (elem->dot >= elem->rule->rhslen)? Qtrue: Qfalse;
}

static VALUE chart_element_children ( VALUE self )
{
    ChartElement* elem;
    
    Data_Get_Struct( self, ChartElement, elem );
    
    return elem->children;
}

static char* chart_element_to_s_internal ( ChartElement* elem, char* buffer )
{
    char *p = buffer;
    VALUE catobj;
    
    catobj = rb_String( cfg_get_category( elem->chart->grammar, elem->top ) );
    p += sprintf( p, "%s (%d, %d)", RSTRING(catobj)->ptr, elem->left, elem->right );
    if( elem->rule ) {
        p += sprintf( p, " : " );
        rule_typestring_internal( elem->rule, p, elem->dot );
    }
    
    return buffer;
}

static VALUE chart_element_to_s ( VALUE self )
{
    ChartElement* elem;
    char buffer [4096];
    
    Data_Get_Struct( self, ChartElement, elem );
    
    return rb_str_new2( chart_element_to_s_internal(elem, buffer) );
}

static VALUE chart_element_inspect ( VALUE self )
{
    ChartElement* elem;
    char buffer [4096];
    char *p;
    
    Data_Get_Struct( self, ChartElement, elem );
    p = buffer;
    
    p += sprintf( p, "#<%s:", rb_class2name(CLASS_OF(self)) );
    
    chart_element_to_s_internal( elem, p );
    p += strlen( p );
    
    p += sprintf( p, ">" );
    
    return rb_str_new2(buffer);
}


/*   ؃Xg̎擾
----------------------------------------------------------------------
 - eGgvf
   * K
   * hbgʒu
   * ͈
   * qGgp^[̃Xg
*/

struct tagBuildSubtreeList
{
    struct tagBuildSubtreeList * prev;
    VALUE itemobj;
};
typedef struct tagBuildSubtreeList BuildSubtreeList;

struct tagBuildSubtreeData
{
    /* ȉׂ͂Ă̕؂ŋLf[^ (PɒlRs[) */
    Chart* chart;
    VALUE* cache; /* [ (left + right * chart->len) * rulesize + ruleid ] ŃANZX */
    int rulesize, catsize, catoffset;
    
    /* cache: LbV
    ͈ (left .. right) 𒣂KL[ƂāA؂̃LbVsB
    ͈̔͂ŁA̋Kł΁AIuWFNgɂȂ邽߁B
    AK蓖ĂĂȂ̂ɊւẮAKp邱ƂłȂB
    ̂悤Ȃ̂͑SėBɒ߂̂ŁALbV̈ӖȂB
    */
    
    /* ȉ͕؂ƂɈقȂf[^ */
    VALUE children;  /* \zꂽzi[邽߂̔z (Array) : ̃o Qnil Ȃ΃gbvf[^ */
};
typedef struct tagBuildSubtreeData BuildSubtreeData;

static VALUE chart_create_elements ( BuildSubtreeData const * parent, State_p item );

static void build_subtree_data_free ( BuildSubtreeData* data )
{
    if( data && NIL_P(data->children) ) {
        ruby_xfree( data->cache );
        ruby_xfree( data );
    }
}

/* BuildSubtreeData.cache ɃANZX邽߂̃AhX߂֐ (₽疼Oǁc) */
static int get_build_subtree_data_cacheaddr ( BuildSubtreeData const * data, int left, int right, int rule )
{
    return (left + right * data->chart->len) * data->rulesize + rule;
}

static void chart_build_subtree_write_children ( BuildSubtreeData* data, BuildSubtreeList* prev )
{
    VALUE volatile ary = rb_ary_new();
    
    while( prev ) {
        rb_ary_push( ary, prev->itemobj );
        prev = prev->prev;
    }
    
    /* Element  children ɂ͊ Element ߂ */
    rb_ary_push( data->children, ary );
}

static void chart_build_subtree ( BuildSubtreeData* data, State_p item, BuildSubtreeList* prev )
{
    /* item gbvƂ镔؂\z */
    BuildSubtreeList list;
    VALUE volatile itemobj = Qnil;
    Chart* chart = data->chart;
    State_p head, tail;
    int wantdot, wantcat;
    int found;
    
    list.prev = prev;
    if( item->dot == 0 ) {
        /* ЂƂ̉͌ʂꂽ */
        chart_build_subtree_write_children( data, prev );
        return;
    }
    wantdot = item->dot - 1;
    wantcat = item->rule->rhs[ wantdot ];
    
    /* item hbg 1 ߂񂷂 */
    for( head = chart_lwait( chart, wantcat, item->left ); head; head = head->next_lwait ) {
        if( head->right >= item->right || head->rule != item->rule || head->dot != wantdot )
            continue;
        /* ł悤₭ head  item ̖hbgЂƂ߂ */
        
        found = 0;
        for( tail = chart_comp( chart, head->right, item->right ); tail; tail = tail->next_comp ) {
            if( tail->top != wantcat )
                continue;
            /* item hbgO̔I[L؂𓾂 (tail gbvƂ镔) */
            itemobj = chart_create_elements( data, tail );
            list.itemobj = itemobj;
            chart_build_subtree( data, head, &list );
            ++found;
        }
        if( found == 0 ) {
            /* I[L҂Ăꍇ̓`[gɑΉ鍀݂Ȃꍇ */
            if( (wantcat < 0) && (head->right == item->right - 1) ) {
                itemobj = chart_element_init2( NULL, chart, wantcat, head->right, item->right );
                list.itemobj = itemobj;
                chart_build_subtree( data, head, &list );
            }
            continue; /*  head  item \z邽߂ head łȂ */
        }
    }
}

/* ؂̃gbvɑΉ鍀𐶐 */
static VALUE chart_create_elements ( BuildSubtreeData const * parent, State_p item )
{
    /* Ȃ񂩃w */
    VALUE volatile obj;
    ChartElement* elem;
    BuildSubtreeData data;
    int addr;
    
    /* I[LɊւĂ̓LbVq̂Ȃ */
    if( !item->rule )
        return chart_element_init( ALLOC(ChartElement), parent->chart, item );
    
    if( item->rule->rhslen == 1 ) {
        /* ɋAĂĂ܂K͎gȂ */
        if( item->rule->lhs == item->rule->rhs[0] ) {
            obj = rule_typestring( 0, NULL, item->rule->ruleobj );
            rb_raise( rb_eNLPException, "Grammar has loop! (%s)", RSTRING(obj)->ptr );
        }
    }
    
    /* LbV̒ɂ΂Ԃ */
    addr = get_build_subtree_data_cacheaddr( parent, item->left, item->right, item->rule->id );
    obj = parent->cache[addr];
    if( !NIL_P(obj) )
        return obj;
    
    /* LbV̒ɖ̂ŐVɃIuWFNgēo^ */
    elem = ALLOC(ChartElement);
    obj = chart_element_init( elem, parent->chart, item );
    
    data = *parent;
    data.children = elem->children;
    chart_build_subtree( &data, item, NULL );
    
    parent->cache[addr] = obj;
    return obj;
}

static VALUE chart_list_subtrees ( VALUE self )
{
    VALUE volatile ret = rb_ary_new(); /* ret ɂ́u؂̃gbvł鍀̂݁vi[ */
    VALUE volatile datobj;
    VALUE volatile elemobj;
    BuildSubtreeData* data;
    Grammar* grammar;
    Chart* chart;
    State_p item;
    int size, dot;
    
    Data_Get_Struct( self, Chart, chart );
    data = ALLOC(BuildSubtreeData);
    data->chart = chart;
    data->children = Qnil;
    
    /* ؂̃gbvKv */
    chart_mark_top( chart );
    
    /* LbVp̃IuWFNgz_𐶐 */
    grammar = chart->grammar;
    data->catsize = grammar->size_nt + grammar->size_pt - 1; /* ID(0) d̂ -1 Ă */
    data->catoffset = grammar->size_pt - 1; /* Category.id ̍ŏl 0 ɂ悤ȃItZbgl */
    data->rulesize = cfg_count_rules( grammar );
    size = data->rulesize * chart->len * chart->len;
    data->cache = ALLOC_N(VALUE, size);
    for( --size; size >= 0; --size )
        data->cache[size] = Qnil;
    
    datobj = Data_Wrap_Struct( rb_cData, NULL, build_subtree_data_free, data );
    
    /* ؂̃gbvGgW߂ */
    for( dot = chart->left; dot <= chart->right; ++dot ) {
        for( item = chart_dot( chart, dot ); item; item = item->next_dot ) {
            /* ؂̈ꕔ, ؃gbvł鍀HĂ */
            if( item->data && (item->tag & TAG_TopItem) ) {
                elemobj = chart_create_elements(data, item);
                rb_ary_push( ret, elemobj );
            }
        }
    }
    
    return ret;
}



/*   `[g̏擾
------------------------------------------------------------------- */

static VALUE chart_get_left ( VALUE self )
{
    Chart* chart;
    
    Data_Get_Struct( self, Chart, chart );
    
    return NUM2INT(chart->left);
}

static VALUE chart_get_right ( VALUE self )
{
    Chart* chart;
    
    Data_Get_Struct( self, Chart, chart );
    
    return NUM2INT(chart->right);
}

static VALUE chart_get_length ( VALUE self )
{
    Chart* chart;
    
    Data_Get_Struct( self, Chart, chart );
    
    return INT2NUM(chart->len);
}

static VALUE chart_get_grammar ( VALUE self )
{
    Chart* chart;
    
    Data_Get_Struct( self, Chart, chart );
    
    return (chart->grammar)? chart->grammar->ruby_object: Qnil;
}


/*   ԃL[
------------------------------------------------------------------- */

StateQueue * queue_init ( StateQueue * queue )
{
    if( !queue )
        queue = ALLOC(StateQueue);
    
    queue->capacity = 64; /* ̒l͏\ɑ傫ĂȂƌオ */
    queue->buffer = ALLOC_N(State*, queue->capacity);
    queue->prepared_capacity = 0;
    queue->prepared_buffer = NULL;
    queue->length = 0;
    queue->pos_head = -1;
    queue->pos_tail = -1;
    
    return queue;
}

StateQueue * queue_init_prepared ( StateQueue * queue, State * * prebuf, int precap )
{
    queue = queue_init( queue );
    if( prebuf && precap > 0 ) {
        queue->prepared_capacity = precap;
        queue->capacity = precap;
        queue->prepared_buffer = prebuf;
        queue->buffer = prebuf;
    }
    
    return queue;
}

void queue_order ( StateQueue * queue, int headspace, int tailspace )
{
    State * * newbuf;
    int  total;
    int  line_p;
    int  taillen;
    int  capacity;
    
    /* L[̏ꍇ͕ёւȂ */
    if( queue->length == 0 )
        return;
    
    total = queue->length + headspace + tailspace;
    line_p = (queue->pos_head < queue->pos_tail);
    capacity = queue->capacity;
    
    if( headspace == 0 && tailspace == 0 ) {
        /* Ĕzu̗v */
        if( line_p )
            return; /* 荞݂NĂȂ̂ŉȂ */
        headspace = (queue->capacity - queue->length) / 2;
        tailspace = queue->capacity - headspace - queue->length;
    }
    
    capacity = (capacity <= 0)? 1: capacity;
    while( capacity < total )
        capacity *= 2;
    
    newbuf = ALLOC_N( State*, capacity );
    if( line_p ) {
        /* 荞łȂ̂ŒPɃRs[邾 */
        MEMCPY( &newbuf[headspace], &queue->buffer[queue->pos_head], State*, queue->length );
    }
    else {
        /* 荞ł̂ōĔzuKv */
        taillen = queue->capacity - queue->pos_head;
        MEMCPY( &newbuf[headspace], &queue->buffer[queue->pos_head], State*, taillen );
        MEMCPY( &newbuf[headspace + taillen], &queue->buffer[0], State*, queue->pos_tail );
    }
    queue_free_internal( queue );
    
    queue->buffer = newbuf;
    queue->pos_head = headspace;
    queue->pos_tail = headspace + queue->length;
    queue->capacity = capacity;
}

void queue_arrange ( StateQueue * queue )
{
    queue_order( queue, 0, 0 );
}

static void queue_expand_buffer ( StateQueue* queue )
{
    if( queue->pos_tail == queue->pos_head )
        queue_order( queue, queue->length / 2, queue->length / 2 );
}

void queue_free_internal ( StateQueue * queue )
{
    if( queue ) {
        if( queue->buffer != queue->prepared_buffer )
            ruby_xfree( queue->buffer ); /* OėpӂĂobt@܂gĂꍇ͉Ȃ */
    }
}

void queue_free ( StateQueue * queue )
{
    if( queue ) {
        queue_free_internal( queue );
        ruby_xfree( queue );
    }
}

static State * queue_add_first_state ( StateQueue * queue, State * state )
{
    /* obt@̒ɏԂzu */
    queue->pos_head = queue->capacity / 2;
    
    queue->pos_tail = queue->pos_head + 1;
    queue->buffer[queue->pos_head] = state;
    queue->length = 1;
    
    return state;
}

State * queue_push ( StateQueue * queue, State * state )
{
    int  wpos;
    
    if( queue->length == 0 ) {
        queue_add_first_state( queue, state );
    }
    else {
        queue_expand_buffer( queue );
        
        wpos = queue->pos_tail;
        queue->buffer[ wpos ] = state;
        ++wpos;
        if( wpos >= queue->capacity )
            wpos = 0;
        queue->pos_tail = wpos;
        queue->length += 1;
    }
    
    return state;
}

State * queue_unshift ( StateQueue * queue, State * state )
{
    int  wpos;
    
    if( queue->length == 0 ) {
        queue_add_first_state( queue, state );
    }
    else {
        queue_expand_buffer( queue );
        
        wpos = queue->pos_head;
        --wpos;
        if( wpos < 0 )
            wpos = queue->capacity - 1;
        queue->buffer[ wpos ] = state;
        queue->pos_head = wpos;
        queue->length += 1;
    }
    
    return state;
}

State * queue_pop ( StateQueue * queue )
{
    State * ret;
    int  rpos;
    
    if( queue->length == 0 )
        return NULL;
    
    rpos = queue->pos_tail;
    --rpos;
    if( rpos < 0 )
        rpos = queue->capacity - 1;
    ret = queue->buffer[ rpos ];
    queue->pos_tail = rpos;
    queue->length -= 1;
    
    if( queue->length == 0 )
        queue->pos_head = queue->pos_tail = -1;
    
    return ret;
}

State * queue_shift ( StateQueue * queue )
{
    State * ret;
    int  rpos;
    
    if( queue->length == 0 )
        return NULL;
    
    rpos = queue->pos_head;
    ret = queue->buffer[ rpos ];
    ++rpos;
    if( rpos >= queue->capacity )
        rpos = 0;
    queue->pos_head = rpos;
    queue->length -= 1;
    
    if( queue->length == 0 )
        queue->pos_head = queue->pos_tail = -1;
    
    return ret;
}

State * queue_insert ( StateQueue * queue, State * state, int pos )
{
    State * * buffer;
    int  head, tail;
    int  delta;
    
    if( pos < 0 || pos > queue->length )
        return NULL;
    
    if( queue->length == 0 )
        return queue_add_first_state( queue, state );
    
    if( queue->length == queue->capacity )
        queue_order( queue, queue->length / 2, queue->length / 2 );
    
    pos += queue->pos_head;
    if( pos >= queue->capacity )
        pos -= queue->capacity;
    
    buffer = queue->buffer;
    head = queue->pos_head;
    tail = queue->pos_tail;
    if( head < tail ) {
        /* ɕłꍇ */
        delta = (head < queue->capacity - tail)? -1: 1;
    }
    else {
        /* 荞łꍇ */
        delta = (pos < head)? -1: 1;
    }
    
    if( delta > 0 ) {
        /* Oɂ炷 */
        MEMMOVE( &buffer[head - 1], &buffer[head], State*, pos - head );
        queue->pos_head -= 1;
        pos -= 1;
    }
    else {
        /* ɂ炷 */
        MEMMOVE( &buffer[pos+1], &buffer[pos], State*, tail - pos );
        queue->pos_tail += 1;
    }
    queue->buffer[pos] = state;
    queue->length += 1;
    
    return state;
}

void queue_display ( FILE * out, StateQueue const * queue )
{
    State const * state;
    char  buf [ 4096 ];
    char * p;
    int  i;
    
    fprintf( out, "[StateQueue:%p]\n", queue );
    
    fprintf( out, " - cap:%d, len:%d, head:%d, tail:%d.\n", queue->capacity, queue->length, queue->pos_head, queue->pos_tail );
    
    for( i = queue->pos_head; i != queue->pos_tail; ++i ) {
        if( i == queue->capacity )
            i = 0;
        state = queue_ref(queue, i);
        p = buf;
        p += sprintf( p, "  %4d: (%d, %d) ", i, state->left, state->right );
        p = rule_typestring_internal( state->rule, p, state->dot );
        p += sprintf( p, "\n" );
        fprintf( out, buf );
    }
    fprintf( out, "\n" );
}

State * queue_find_equal ( StateQueue * queue, State * state )
{
    State * * ptr;
    int  i;
    
    ptr = &queue->buffer[ queue->pos_head ];
    for( i = queue->pos_head; i != queue->pos_tail; ++i, ++ptr ) {
        if( i == queue->capacity ) {
            i = 0;
            ptr = &queue->buffer[ 0 ];
        }
        
        if( state == *ptr )
            return *ptr;
    }
    
    return NULL;
}


/*  Chart NX̍\z
----------------------------------------------------------------------
*/
void init_chart ( void )
{
    /* lCeBu `[gNX̒` */
    rb_cChart = rb_define_class_under( rb_mExt, "Chart", rb_cData );
    rb_cChart_Element = rb_define_class_under( rb_cChart, "Element", rb_cData );
    
    /* \bh */
    rb_define_method( rb_cChart, "exist?", chart_exist_p, 0 );
    rb_define_method( rb_cChart, "count_item", chart_count_item_m, 0 );
    rb_define_method( rb_cChart, "count_deleted", chart_count_deleted, 0 );
    rb_define_method( rb_cChart, "count_cancelled", chart_count_cancelled, 0 );
    rb_define_method( rb_cChart, "memsize", chart_memsize_m, 0 );
    rb_define_method( rb_cChart, "display", chart_display, 1 );
    rb_define_method( rb_cChart, "display_itemset", chart_display_itemset, 1 );
    rb_define_method( rb_cChart, "list_subtrees", chart_list_subtrees, 0 );
    rb_define_method( rb_cChart, "left", chart_get_left, 0 );
    rb_define_method( rb_cChart, "right", chart_get_right, 0 );
    rb_define_method( rb_cChart, "length", chart_get_length, 0 );
    rb_define_method( rb_cChart, "size", chart_get_length, 0 );
    rb_define_method( rb_cChart, "grammar", chart_get_grammar, 0 );
    
    /* \bh */
    rb_define_method( rb_cChart_Element, "rule", chart_element_get_rule, 0 );
    rb_define_method( rb_cChart_Element, "top", chart_element_get_top, 0 );
    rb_define_method( rb_cChart_Element, "left", chart_element_get_left, 0 );
    rb_define_method( rb_cChart_Element, "right", chart_element_get_right, 0 );
    rb_define_method( rb_cChart_Element, "dot", chart_element_get_dot, 0 );
    rb_define_method( rb_cChart_Element, "children", chart_element_children, 0 );
    rb_define_method( rb_cChart_Element, "complete?", chart_element_complete_p, 0 );
    rb_define_method( rb_cChart_Element, "to_s", chart_element_to_s, 0 );
    rb_define_method( rb_cChart_Element, "inspect", chart_element_inspect, 0 );
}


