/*                                                     project Gemini 
----------------------------------------------------------------------
    @NX
      $Id: grammar.c,v 1.22 2002/11/21 13:46:43 abe Exp $

       Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
TODO:
 - freeze() ̃T|[g
*/
#define DEBUGSW 0
#define DEBUG_FREE  0  /**< IuWFNg̃fobOx */
#define DEBUG_MARK  0  /**< IuWFNg}[ÑfobOx */
#include "nlputil.h"
#include <time.h>
#include <string.h>


/* Ruby ̃NXIuWFNg */
VALUE rb_cGrammar;
VALUE rb_cGrammar_Rule;
VALUE rb_cGrammar_Category;


/* Ruby ̃\bhID */
static ID ID_strip_bang;


/* ̃t@Ĉ݂̃O[oϐ */
static VALUE local_object_cache;
static VALUE rbstr_prop_pr, rbstr_prop_head;


/* vg^Cv */
static void cfg_initialize_rule ( Grammar * gram, VALUE block );
static void cfg_initialize_FIRST ( Grammar * gram, VALUE block );
static void cfg_initialize_array ( Grammar* gram, VALUE ruleary, VALUE firstary );


/*   Ext::Grammar::Category
------------------------------------------------------------------- */

Category_p nlp_Category ( VALUE catobj )
{
    Category_p cat;
    
    if( CLASS_OF(catobj) != rb_cGrammar_Category )
        rb_raise( rb_eTypeError, "Cannot convert %s into Ext::Grammar::Category.", rb_class2name(CLASS_OF(catobj)) );
    
    Data_Get_Struct( catobj, Category, cat );
    
    return cat;
}

static void category_free ( Category * cat )
{
    if( cat ) {
        ruby_xfree( cat );
    }
}

static VALUE category_new ( Grammar * gram, const char * name, int id )
{
    VALUE volatile catstr;
    VALUE volatile ary;
    VALUE volatile str;
    VALUE volatile obj;
    Category * cat;
    int  n;
    
    /* [U[ ID [0] ̋L邱Ƃ͂łȂ (ߍ܂ꂽvÔ݂\) */
    if( id == 0 )
        rb_raise( rb_eNLPException, "Cannot use zero (0) for category-ID." );
    
    catstr = rb_str_new2( name );
    rb_funcall( catstr, ID_strip_bang, 0 );
    
    cat = ALLOC( Category );
    cat->id     = id;
    cat->string = catstr;
    
    obj = Data_Wrap_Struct( rb_cGrammar_Category, NULL, category_free, cat );
    
    /* @IuWFNgɓo^ ( id ɑꏊ󂯂Ă) */
    if( id < 0 ) {
        ary = gram->preterm_ary; /* ary: IuWFNgz*/
        str = gram->preterm_str; /* str: z */
        n = -id; /* n: 1 ` */
    }
    else {
        ary = gram->nonterm_ary;
        str = gram->nonterm_str;
        n = id; /* n: 1 ` */
    }
    if( n < RARRAY(ary)->len && !NIL_P(RARRAY(ary)->ptr[n]) )
        rb_raise( rb_eNLPException, "[BUG/nlputil.so] Already allocated category ID #%d.", n );
    rb_ary_store( str, n, catstr );
    rb_ary_store( ary, n, obj );
    rb_hash_aset( gram->category_map, catstr, obj );
    
    return obj;
}

static VALUE category_id ( VALUE self )
{
    Category* cat;
    
    Data_Get_Struct( self, Category, cat );
    
    return rb_fix_new(cat->id);
}

static VALUE category_nonterm_p ( VALUE self )
{
    Category* cat;
    
    Data_Get_Struct( self, Category, cat );
    
    return (cat->id > 0);
}

static VALUE category_preterm_p ( VALUE self )
{
    Category* cat;
    
    Data_Get_Struct( self, Category, cat );
    
    return (cat->id < 0);
}

static VALUE category_inspect ( VALUE self )
{
    Category* cat;
    char buf [256];
    int  n;
    
    Data_Get_Struct( self, Category, cat );
    #if SAFE_sprintf
    n = _snprintf( buf, ARYSIZE(buf)-1, "%s (%d)", RSTRING(cat->string)->ptr, cat->id );
    buf[ ARYSIZE(buf)-1 ] = '\0';
    #else
    n = sprintf( buf, "%s (%d)", RSTRING(cat->string)->ptr, cat->id );
    #endif
    
    return rb_str_new( buf, n );
}

static VALUE category_eql_p ( VALUE self, VALUE obj )
{
    Category* cat;
    Category* target;
    
    /* IuWFNgȂΈv͓̂R */
    if( self == obj )
        return Qtrue;
    
    Data_Get_Struct( self, Category, cat );
    
    /* lƔrꍇ͂ɂ킩 */
    if( FIXNUM_P(obj) )
        return (cat->id == rb_fix2int(obj))? Qtrue: Qfalse;
    
    /* Category IuWFNgnꂽꍇ */
    if( CLASS_OF(obj) == rb_cGrammar_Category ) {
        Data_Get_Struct( obj, Category, target );
        return (cat->id == target->id)? Qtrue: Qfalse;
    }
    
    /* obj  String łrł悤Ɂc */
    return rb_equal( cat->string, obj );
}

static VALUE category_hash ( VALUE self )
{
    Category* cat;
    
    Data_Get_Struct( self, Category, cat );
    
    /* LnbṼL[ɂ */
    return rb_fix_new( rb_str_hash( cat->string ) );
}

static VALUE category_name ( VALUE self )
{
    Category* cat;
    
    Data_Get_Struct( self, Category, cat );
    
    return cat->string;
}

static VALUE category_compare ( VALUE self, VALUE obj )
{
    Category* cat;
    Category* arg;
    int argid = 0;
    
    Data_Get_Struct( self, Category, cat );
    
    if( CLASS_OF(obj) == rb_cGrammar_Category ) {
        Data_Get_Struct( obj, Category, arg );
        argid = arg->id;
    }
    else if( FIXNUM_P(obj) ) {
        argid = rb_fix2int( obj );
    }
    
    if( argid != 0 )
        return rb_fix_new( (cat->id == argid)? 0: ((cat->id > argid)? 1: -1) );
    
    return rb_fix_new( rb_str_cmp( cat->string, rb_String(obj) ) );
}

int category_int ( VALUE obj )
{
    Category* cat;
    
    if( CLASS_OF(obj) != rb_cGrammar_Category )
        return 0;
    Data_Get_Struct( obj, Category, cat );
    
    return cat->id;
}


/*   Ext::Grammar::Rule
------------------------------------------------------------------- */

Rule_p nlp_Rule ( VALUE ruleobj )
{
    Rule_p rule;
    
    if( CLASS_OF(ruleobj) != rb_cGrammar_Rule )
        rb_raise( rb_eTypeError, "Cannot convert %s into Ext::Grammar::Rule.", rb_class2name(CLASS_OF(ruleobj)) );
    
    Data_Get_Struct( ruleobj, Rule, rule );
    
    return rule;
}

static void rule_free ( Rule * rule )
{
    if( rule ) {
        if( rule->rhs != rule->rhsbuf )
            ruby_xfree( rule->rhs );
        ruby_xfree( rule );
    }
}

/** @brief @KIuWFNg̐
 *
 * @IuWFNg𐶐A Ruby IuWFNgԂ܂B
 *
 * @param gram [i/o] I[i[ƂȂ镶@IuWFNg
 * @param lhs [in] ӋL
 * @param rhs [in] EӋL (zɕϊ܂)
 * @return ꂽ@KIuWFNg
 */
static VALUE rule_new ( Grammar * gram, VALUE lhs, VALUE rhs )
{
    VALUE volatile obj;
    VALUE volatile rhsary;
    VALUE volatile v;
    Category *cat;
    Rule *rule;
    int n, len, hash;
    
    if( CLASS_OF(lhs) != rb_cGrammar_Category )
        rb_raise( rb_eTypeError, "lhs requires Ext::Grammar::Category (given %s)", rb_class2name(CLASS_OF(lhs)) );
    
    rule = ALLOC_ZERO_N( Rule, 1 );
    obj = Data_Wrap_Struct( rb_cGrammar_Rule, NULL, rule_free, rule );
    
    rule->gramobj = gram->ruby_object;
    rule->ruleobj = obj;
    
    /* ӂ̐ݒ */
    Data_Get_Struct( lhs, Category, cat );
    rule->lhs = cat->id;
    
    /* Eӂ̐ݒ */
    rhsary = rb_Array( rhs );
    len = RARRAY(rhsary)->len;
    if( len > RULE_BUILTIN_BUFFER_SIZE )
        rule->rhs = ALLOC_N( int, len );
    else
        rule->rhs = rule->rhsbuf;
    
    for( n = 0; n < len; ++n ) {
        v = RARRAY(rhsary)->ptr[n];
        if( CLASS_OF(v) != rb_cGrammar_Category )
            rb_raise( rb_eTypeError, "rhs element requires Ext::Grammar::Category (given %s)", rb_class2name(CLASS_OF(v)) );
        Data_Get_Struct( v, Category, cat );
        rule->rhs[n] = cat->id;
    }
    rule->rhslen = len;
    
    /* nbV̐ (rb_ary_hash() Ql) */
    hash = rule->rhslen;
    hash = rule->lhs ^ ((hash << 1) | ((hash < 0)? 1: 0));
    for( n = len - 1; n >= 0; --n )
        hash = rule->rhs[n] ^ ((hash << 1) | ((hash < 0)? 1: 0));
    rule->hash = rb_fix_new(hash);
    
    /* f[^̐ݒ */
    rule->id           = -1; /* ID ͂܂s */
    rule->prop         = Qnil;
    rule->object_cache = Qnil;
    rule->next_lc      = NULL;
    rule->next_lhs     = NULL;
    
    return obj;
}

static VALUE rule_eql_p ( VALUE self, VALUE obj )
{
    Rule *r1, *r2;
    int n;
    
    if( CLASS_OF(obj) != rb_cGrammar_Rule )
        return Qfalse;
    
    Data_Get_Struct( self, Rule, r1 );
    Data_Get_Struct( obj, Rule, r2 );
    
    /* v͎̒ʂB
     *  - ӂv
     *  - Eӂ̓ev
     * ȀB
     *  - nbVlv (L݂̂nbVlvZĂ̂)
     */
    if( r1->hash != r2->hash )
        return Qfalse;
    
    /* nbVlvĂIuWFNgv킯ł͂Ȃ */
    if( r1->rhslen != r2->rhslen )
        return Qfalse;
    
    if( r1->lhs != r2->lhs )
        return Qfalse;
    
    for( n = r1->rhslen - 1; n >= 0; --n ) {
        if( r1->rhs[n] != r2->rhs[n] )
            return Qfalse;
    }
    
    return Qtrue;
}

static VALUE rule_compare ( VALUE self, VALUE obj )
{
    Rule *r1, *r2;
    
    /* Rule > (other object) */
    if( CLASS_OF(obj) != rb_cGrammar_Rule )
        return rb_fix_new(1);
    
    /* Rule ̏ꍇ ID Ŕr */
    Data_Get_Struct( self, Rule, r1 );
    Data_Get_Struct( obj, Rule, r2 );
    
    return rb_fix_new( (r1->id == r2->id)? 0: ((r1->id > r2->id)? 1: -1) );
}

static VALUE rule_hash ( VALUE self )
{
    Rule* r;
    
    Data_Get_Struct( self, Rule, r );
    
    return r->hash;
}

VALUE rule_get_property ( Rule_p rule, VALUE key )
{
    VALUE volatile keystr;
    VALUE* ptr;
    int n, len;
    
    if( NIL_P(rule->prop) )
        return Qnil;
    
    rb_ary_getptr( rule->prop, &ptr, &len );
    if( len % 2 != 0 )
        rb_raise( rb_eRuntimeError, "broken rule's property found (len=%d)", len );
    
    keystr = rb_String(key);
    for( n = 0; n < len; n += 2 ) { /* prop: ɃL[, ɒlz */
        if( rb_equal( keystr, ptr[n] ) )
            return ptr[n+1];
    }
    return Qnil;
}

static VALUE rule_get_prop ( VALUE self, VALUE key )
{
    Rule* r;
    
    Data_Get_Struct( self, Rule, r );
    
    return rule_get_property( r, key );
}

static void rule_setobj_rhs ( Rule * rule, VALUE object )
{
    if( !NIL_P(rule->object_cache) )
        rb_ary_store( rule->object_cache, 1, object );
}

static VALUE rule_getobj_rhs ( Rule const * rule )
{
    if( NIL_P(rule->object_cache) )
        return Qnil;
    return RARRAY(rule->object_cache)->ptr[ 1 ];
}

static VALUE rule_getobj_typestring ( Rule const * rule )
{
    if( NIL_P(rule->object_cache) )
        return Qnil;
    return RARRAY(rule->object_cache)->ptr[ 2 ];
}

static void rule_alloc_prop ( Rule* rule )
{
    Grammar* gram;
    VALUE volatile v;
    
    if( NIL_P(rule->object_cache) ) /* (|| r->id < 0) ID 莞 prop ݒłȂ (Grammar  GC Ǘ邽) */
        rb_raise( rb_eRuntimeError, "property table unallocatable to no-ID rule" );
    
    if( NIL_P(rule->prop) ) {
        Data_Get_Struct( rule->gramobj, Grammar, gram );
        
        /* ʒuŒ肳Ă镪͗̈ƃvpeB蓖ĂĂ */
        v = rb_ary_new3( 4, rbstr_prop_pr, Qnil, rbstr_prop_head, Qnil );
        
        /* K̃vpeBݒ肷 */
        rule->prop = v;
        rb_ary_store( rule->object_cache, 0, v ); /* cache[0]: vpeB */
    }
}

static VALUE rule_set_prop ( VALUE self, VALUE key, VALUE value )
{
    /* ̕ҏW freeze() ĂĂs\ */
    VALUE volatile v;
    VALUE volatile keystr;
    VALUE* ptr;
    Rule* rule;
    int n, len;
    
    Data_Get_Struct( self, Rule, rule );
    
    /* vpeB̈悪mۂĂȂ΍쐬Ă */
    rule_alloc_prop( rule );
    
    rb_ary_getptr( rule->prop, &ptr, &len );
    if( len % 2 != 0 )
        rb_raise( rb_eRuntimeError, "broken rule property" );
    
    /* o^悤ƂĂL[邩Ȃ */
    keystr = rb_String( key );
    for( n = 0; len; n += 2 ) {
        if( rb_equal( keystr, ptr[n] ) ) {
            ptr[n+1] = value;
            return value;
        }
    }
    
    /* ΐVɓo^ */
    rb_ary_push( rule->prop, keystr );
    rb_ary_push( rule->prop, value );
    return value;
}

char * rule_typestring_internal ( Rule const * rule, char* buf, int dot )
{
    VALUE obj;
    Grammar* gram;
    char* p = buf;
    int i;
    
    if( dot < 0 ) {
        /* LbVꂽ typestring g */
        obj = rule_getobj_typestring( rule );
        if( !NIL_P(obj) ) {
            strcpy( buf, RSTRING(obj)->ptr );
            return buf + strlen(buf);
        }
    }
    Data_Get_Struct( rule->gramobj, Grammar, gram );
    
    /* ӂ̏ */
    p += sprintf( p, "%s -->", cfg_getcatname(gram, rule->lhs) );
    
    /* Eӂ̏ */
    for( i = 0; i < rule->rhslen; ++i ) {
        if( dot == i )
            p += sprintf( p, " *" );
        p += sprintf( p, " %s", cfg_getcatname(gram, rule->rhs[i]) );
    }
    if( dot == i )
        p += sprintf( p, " *" );
    *p = '\0';
    
    return p;
}

VALUE rule_typestring ( int argc, VALUE * argv, VALUE self )
{
    VALUE volatile dotobj;
    Rule_p  r;
    char  buf[2048];
    char*  tail;
    int  dot;
    
    Data_Get_Struct( self, Rule, r );
    
    rb_scan_args( argc, argv, "01", &dotobj );
    dot = (NIL_P(dotobj))? -1: rb_num2int(dotobj);
    
    /* LbVꂽ typestring gꍇ (obt@̃Rs[͂Ȃ) */
    if( dot < 0 ) {
        dotobj = rule_getobj_typestring(r);
        if( !NIL_P(dotobj) )
            return dotobj;
    }
    
    /* \z */
    tail = rule_typestring_internal( r, buf, dot );
    
    return rb_str_new( buf, tail - buf );
}

char * rule_to_s_internal ( Rule const * r, char * p )
{
    VALUE volatile str;
    VALUE *ptr;
    int n, len, count;
    
    /* vpeBĂȂꍇ typestring  */
    if( NIL_P(r->prop) ) {
        rule_typestring_internal( r, p, -1 );
        return;
    }
    
    /* vpeB */
    count = 0;
    rb_ary_getptr( r->prop, &ptr, &len );
    *(p++) = '(';
    for( n = 0; n < len; n += 2 ) {
        if( !NIL_P(ptr[1]) ) {
            if( ptr[1] == Qtrue )
                p += sprintf( p, "%s,", RSTRING(ptr[0])->ptr );
            else {
                str = rb_String(ptr[1]);
                p += sprintf( p, "%s=%s,", RSTRING(ptr[0])->ptr, RSTRING(str)->ptr );
            }
            ++count;
        }
        ptr += 2;
    }
    --p; /* |C^ʒu ,  ( ɖ߂ */
    if( count == 0 ) {
        *p = '\0'; /* ЂƂf[^܂Ȃ̂ ( 폜 */
    }
    else {
        *(p++) = ')'; /* f[^܂ꂽ̂Ŋʂ */
        *(p++) = ' ';
    }
    
    /* K */
    p = rule_typestring_internal( r, p, -1 );
    
    return p;
}

static VALUE rule_to_s ( VALUE self )
{
    char buf[4096]; /* łΕKvʂɉĊmۂc */
    Rule* r;
    
    Data_Get_Struct( self, Rule, r );
    rule_to_s_internal( r, buf );
    
    return rb_str_new2( buf );
}

char* rule_inspect_internal ( Rule const * rule, char* buf )
{
    char* p = buf;
    
    p += sprintf( p, "[#%d: ", rule->id );
    p = rule_to_s_internal( rule, p );
    p += sprintf( p, "]" );
    
    return p;
}

static VALUE rule_inspect ( VALUE self )
{
    char buf [4096]; /* łΕKvʂɉĊmۂc */
    char* tail;
    Rule_p rule;
    
    Data_Get_Struct( self, Rule, rule );
    tail = rule_inspect_internal( rule, buf );
    
    return rb_str_new(buf, tail - buf);
}

static VALUE rule_get_builtin_prop ( Rule_p r, int index )
{
    Grammar* gram;
    
    if( NIL_P(r->prop) )
        return Qnil;
    
    index *= 2;
    if( index+1 >= RARRAY(r->prop)->len )
        return Qnil;
    
    return RARRAY(r->prop)->ptr[ index+1 ];
}

static VALUE rule_set_builtin_prop ( Rule * r, int index, VALUE value )
{
    Grammar* gram;
    
    rule_alloc_prop( r );
    
    /* VXe`͈̔͊Ow肷Ɗmɑ̃f[^j󂷂 */
    rb_ary_store( r->prop, index*2 + 1, value );
    
    return value;
}

static VALUE rule_get_head ( VALUE self )
{
    Rule* r;
    
    Data_Get_Struct( self, Rule, r );
    
    return rule_get_builtin_prop( r, 1 );
}

static VALUE rule_get_lhs ( VALUE self )
{
    Grammar* grammar;
    Rule* r;
    
    Data_Get_Struct( self, Rule, r );
    Data_Get_Struct( r->gramobj, Grammar, grammar );
    
    if( r->lhs <= 0 ) {
        if( r->lhs == 0 )
            rb_raise( rb_eRuntimeError, "rule #%d has null category LHS", r->id );
        else
            rb_raise( rb_eRuntimeError, "rule #%d has preterm LHS", r->id );
    }
    
    return cfg_get_category(grammar, r->lhs);
}

static VALUE rule_get_rhs ( VALUE self )
{
    VALUE volatile v;
    Grammar* gram;
    VALUE* ary;
    Rule* r;
    int n, id;
    
    Data_Get_Struct( self, Rule, r );
    
    /* LbV΂g */
    v = rule_getobj_rhs( r );
    if( !NIL_P(v) )
        return v;
    
    /* f[^\z */
    Data_Get_Struct( r->gramobj, Grammar, gram );
    ary = ALLOCA_N( VALUE, r->rhslen );
    for( n = r->rhslen - 1; n >= 0; --n ) {
        id = r->rhs[n];
        if( id == 0 )
            rb_raise( rb_eNLPException, "RHS[%d] in rule #%d has null category", n, r->id );
        ary[n] = cfg_get_category(gram, id);
    }
    
    v = rb_ary_new4( r->rhslen, ary );
    rule_setobj_rhs( r, v );
    return v;
}

static VALUE rule_get_id ( VALUE self )
{
    Rule* r;
    
    Data_Get_Struct( self, Rule, r );
    
    return rb_fix_new(r->id);
}

/** @brief @K ID ݒ肷
 *
 * ^ꂽ@K ID ݒ肵܂B
 * ȂAݒ肳KɁA܂ ID ݒ肳ĂȂƂɂȂ܂B
 *
 * @param ruleobj [out] ݒ肷K
 * @param id [in] ݒ肷 ID
 * @return KIuWFNg
 */
static VALUE rule_set_id ( VALUE ruleobj, int id )
{
    VALUE volatile typestr;
    VALUE volatile v;
    Grammar* gram;
    Rule* r;
    int n;
    
    Data_Get_Struct( ruleobj, Rule, r );
    
    if( r->id >= 0 )
        rb_raise( rb_eRuntimeError, "rule ID has alread setted" );
    if( id < 0 )
        rb_raise( rb_eArgError, "cannot set negative rule ID" );
    
    Data_Get_Struct( r->gramobj, Grammar, gram );
    r->id = id;
    
    /* ID ݒ肳ꂽ̂ŕ@IuWFNgɓo^ */
    typestr = rule_typestring(0,NULL,ruleobj);
    v = rb_ary_new3( 3, Qnil, Qnil, typestr );
    r->object_cache = v;
    rb_ary_store( gram->rule_ary, id, ruleobj );
    rb_ary_store( gram->rule_buf, id, v );
    rb_hash_aset( gram->rule_str, typestr, ruleobj ); /* v[2]  typestring */
    
    /* nbVւ̓o^ (nbVp̃obt@ cfg_add_symbol() ō쐬EXV) */
    r->next_lhs = gram->hash_lhs[ r->lhs ];
    gram->hash_lhs[ r->lhs ] = r;
    
    if( r->rhs[0] < 0 ) {
        n = -r->rhs[0];
        r->next_lc = gram->hash_lc_pt[n];
        gram->hash_lc_pt[n] = r;
    }
    else {
        n = r->rhs[0];
        r->next_lc = gram->hash_lc[n];
        gram->hash_lc[n] = r;
    }
    
    return ruleobj;
}


/*  L擾
-------------------------------------------------------------------------
   size_pt  size_nt ̓CfNX 0 ̃_~[L܂񂾒Ȃ̂ŁA
ꂼ (ۂɎgƂ̂łL) + 1 ɂȂB
*/
VALUE cfg_get_category ( Grammar const * gram, int id )
{
    if( id < 0 ) {
        if( -id >= gram->size_pt )
            rb_raise( rb_eIndexError, "Preterm-ID out of range" );
        
        return RARRAY(gram->preterm_ary)->ptr[-id];
    }
    
    if( id >= gram->size_nt )
        rb_raise( rb_eIndexError, "Nonterm-ID out of range" );
    
    return RARRAY(gram->nonterm_ary)->ptr[id];
}

VALUE cfg_getcat_internal ( Grammar const * gram, VALUE key )
{
    VALUE volatile keystr;
    
    /* Category NXłΎgԂ */
    if( CLASS_OF(key) == rb_cGrammar_Category )
        return key;
    
    /* FIXNUM łΑΉ ID ̋LԂ */
    if( FIXNUM_P(key) )
        return cfg_get_category( gram, rb_fix2int(key) );
    
    /* ňvLԂ */
    keystr = rb_String(key);
    return rb_hash_aref( gram->category_map, keystr );
}

static VALUE cfg_getcat ( VALUE self, VALUE key )
{
    Grammar * gram;
    
    Data_Get_Struct( self, Grammar, gram );
    
    return cfg_getcat_internal( gram, key );
}

const char* cfg_getcatname ( Grammar const * gram, int id )
{
    VALUE catobj = cfg_get_category( gram, id );
    Category* cat;
    
    Data_Get_Struct( catobj, Category, cat );
    
    return RSTRING(cat->string)->ptr;
}


/*   p[^𒲂ׂ
------------------------------------------------------------------- */

int cfg_count_rules ( Grammar const * gram )
{
    return RARRAY(gram->rule_ary)->len;
}

static VALUE cfg_size_rule_m ( VALUE self )
{
    Grammar  *g;
    
    Data_Get_Struct( self, Grammar, g );
    
    return rb_fix_new( RARRAY(g->rule_ary)->len );
}

static VALUE cfg_size_nonterm_m ( VALUE self )
{
    Grammar  *g;
    
    Data_Get_Struct( self, Grammar, g );
    
    return rb_fix_new(cfg_size_nonterm(g));
}

static VALUE cfg_size_preterm_m ( VALUE self )
{
    Grammar  *g;
    
    Data_Get_Struct( self, Grammar, g );
    
    return rb_fix_new(cfg_size_preterm(g));
}

static VALUE cfg_size_symbol_m ( VALUE self )
{
    Grammar  *g;
    
    Data_Get_Struct( self, Grammar, g );
    
    return rb_fix_new(cfg_size_symbol(g));
}

static VALUE cfg_FIRST ( VALUE self )
{
    Grammar  *g;
    
    Data_Get_Struct( self, Grammar, g );
    
    if( NIL_P(g->first_map) )
        return Qnil; /* nil Ȃǂ͕łȂ */
    
    return rb_obj_dup( g->first_map );
}

static VALUE cfg_has_FIRST_p ( VALUE self )
{
    Grammar* g;
    
    Data_Get_Struct( self, Grammar, g );
    
    return (NIL_P(g->first_map))? Qfalse: Qtrue;
}


/*   FIRST W̑
------------------------------------------------------------------- */

void firstset_delete ( FIRSTSet* fset )
{
    if( fset ) {
        ruby_xfree( fset->buf );
        ruby_xfree( fset->ptr );
        ruby_xfree( fset );
    }
}

static int firstset_check_firstary ( Grammar* gram, VALUE firstary )
{
    VALUE *ptr;
    VALUE *datptr;
    int len, datlen;
    
    VALUE volatile aryobj;
    VALUE pair, catobj;
    Category* cat;
    int i, j;
    int keycat, preterm, nonterm;
    
    nonterm = cfg_size_nonterm(gram);
    preterm = -cfg_size_preterm(gram);
    
    len = RARRAY(firstary)->len;
    ptr = RARRAY(firstary)->ptr;
    for( i = 0; i < len; ++i ) {
        pair = ptr[i];
        Check_Type( pair, T_ARRAY );
        
        catobj = RARRAY(pair)->ptr[0];
        if( CLASS_OF(catobj) != rb_cGrammar_Category )
            rb_raise( rb_eNLPException, "FIRST key is not Ext::Grammar::Category." );
        Data_Get_Struct( catobj, Category, cat );
        keycat = cat->id;
        if( keycat == 0 )
            rb_raise( rb_eNLPException, "given invalid category for FIRST key" );
        if( keycat < 0 )
            continue; /* I[L FIRST W͂ĂӖȂ */
        if( keycat >= nonterm )
            rb_raise( rb_eNLPException, "FIRST key ID out of range" );
        
        aryobj = rb_Array( RARRAY(pair)->ptr[1] );
        datlen = RARRAY(aryobj)->len;
        datptr = RARRAY(aryobj)->ptr;
        for( j = 0; j < datlen; ++j ) {
            if( CLASS_OF(datptr[j]) != rb_cGrammar_Category )
                rb_raise( rb_eNLPException, "FIRST value is not Ext::Grammar::Category." );
            Data_Get_Struct( datptr[j], Category, cat );
            if( cat->id == 0 )
                rb_raise( rb_eNLPException, "invalid category for FIRST value" );
            if( cat->id <= preterm || cat->id >= nonterm )
                rb_raise( rb_eNLPException, "FIRST value ID out of range" );
        }
    }
}

static FIRSTSet* firstset_initialize ( FIRSTSet* fset, VALUE firstary )
{
    Grammar* grammar = fset->grammar;
    VALUE *ptr, *datptr;
    int len, datlen;
    int keycat, preterm, nonterm, symbols;
    
    nonterm = cfg_size_nonterm(grammar);
    preterm = cfg_size_preterm(grammar);
    symbols = cfg_size_symbol(grammar);
    
    len = RARRAY(firstary)->len;
    ptr = RARRAY(firstary)->ptr;
    for( ; len > 0; --len, ++ptr ) {
        keycat = category_int( RARRAY(*ptr)->ptr[0] );
        
        /* I[L FIRST ͋Ɍ܂Ă̂ň͔̂I[L̂ */
        if( keycat >= 0 ) {
            fset->ptr[ keycat ] = &fset->buf[ symbols * keycat + (preterm - 1) ];
            rb_ary_getptr( RARRAY(*ptr)->ptr[1], &datptr, &datlen );
            for( ; datlen > 0; --datlen, ++datptr )
                fset->ptr[ keycat ][ category_int(*datptr) ] = 1;
        }
    }
    
    return fset;
}

FIRSTSet* firstset_new ( Grammar* gram )
{
    VALUE volatile firstary = rb_Array( gram->first_map );
    FIRSTSet* fset;
    int nonterm, preterm, symbols;
    
    if( !firstset_check_firstary(gram, firstary) )
        return NULL;
    
    /* e[ȕ */
    fset = ALLOC(FIRSTSet);
    fset->grammar = gram;
    nonterm = cfg_size_nonterm(gram);
    preterm = cfg_size_preterm(gram);
    symbols = cfg_size_symbol(gram);
    
    fset->buf = rb_alloc_zero( int, nonterm * symbols );
    fset->ptr = rb_alloc_zero( int*, nonterm );
    
    return firstset_initialize( fset, firstary );
}



/*   w肳ꂽIuWFNg Ext::Grammar ł邩H
----------------------------------------------------------------------
*/
static VALUE cfg_proper_p ( VALUE klass, VALUE obj )
{
    VALUE volatile strobj;
    VALUE v;
    char *p;
    
    /* Ext::PacketBlock 󂯕tȂ */
    v = rb_const_get( rb_cObject, rb_intern("Ext") );
    if( NIL_P(v) )
        return Qfalse;
    
    v = rb_const_get( v, rb_intern("PacketBlock") );
    if( !RTEST(rb_obj_is_kind_of(obj, v)) )
        return Qfalse;
    
    /* ubN𒲂ׂ */
    v = rb_funcall( obj, rb_intern("name"), 0 );
    strobj = rb_String(v);
    p = RSTRING(strobj)->ptr;
    if( *p != 'p' && *p != 'P' )
        return Qfalse;
    ++p;
    
    /* /^p[gG]rammar$/ 󂯕t */
    if( strcmp( p, "grammar" ) != 0 && strcmp( p, "Grammar" ) != 0 )
        return Qfalse;
    
    return Qtrue;
}


/*   Ruby Ō݊`p̃RXgN^EfXgN^
------------------------------------------------------------------- */

static void cfg_free ( Grammar* gram )
{
    if( gram ) {
        #if DEBUG_FREE > 0
        fprintf( stderr, "[FREE] cfg_free(): begin.\n" );
        #endif
        ruby_xfree( gram->hash_lhs );
        ruby_xfree( gram->hash_lc );
        ruby_xfree( gram->hash_lc_pt );
        ruby_xfree( gram );
        #if DEBUG_FREE > 0
        fprintf( stderr, "[FREE] cfg_free(): end.\n" );
        #endif
    }
}

static void cfg_mark ( Grammar* gram )
{
    if( gram ) {
        #if DEBUG_MARK > 0
        fprintf( stderr, "[MARK] cfg_mark(): begin.\n" );
        #endif
        rb_gc_mark( gram->grammar_property );
        rb_gc_mark( gram->block_property );
        rb_gc_mark( gram->category_map );
        rb_gc_mark( gram->nonterm_ary );
        rb_gc_mark( gram->nonterm_str );
        rb_gc_mark( gram->preterm_ary );
        rb_gc_mark( gram->preterm_str );
        rb_gc_mark( gram->rule_ary );
        rb_gc_mark( gram->rule_buf );
        rb_gc_mark( gram->rule_str );
        rb_gc_mark( gram->first_map );
        #if DEBUG_MARK > 0
        fprintf( stderr, "[MARK] cfg_mark(): end.\n" );
        #endif
    }
}

static VALUE cfg_initialize_block ( Grammar* gram, VALUE root )
{
    VALUE volatile childname;
    VALUE child;
    char *p;
    
    gram->block_property = rb_obj_dup( rb_funcall(root, rb_intern("property"), 0) );
    
    while( !NIL_P( child = rb_funcall(root, rb_intern("getb"), 0) ) )
    {
        childname = rb_String( rb_funcall(child, rb_intern("name"), 0) );
        p = RSTRING(childname)->ptr;
        if( p[0] == 'd' || p[0] == 'D' ) {
            ++p;
            if( strcmp( p, "rule" ) == 0 || strcmp( p, "Rule" ) == 0 || strcmp( p, "rules" ) == 0 || strcmp( p, "Rules" ) == 0 ) {
                /* @K`ubN */
                cfg_initialize_rule( gram, child );
            }
            else if( strcmp( p, "first" ) == 0 || strcmp( p, "First" ) == 0 || strcmp( p, "FIRST" ) == 0 ) {
                /* FIRST W`ubN */
                cfg_initialize_FIRST( gram, child );
            }
        }
        rb_funcall( child, rb_intern("close"), 0 );
    }
}

static void cfg_initialize_category ( Grammar* grammar )
{
    VALUE volatile v;
    Category* cat;
    
    /* LID 0 ɂĂ͓ʈ */
    v = rb_str_new2( "(null)" );
    rb_ary_store( grammar->nonterm_str, 0, v );
    rb_ary_store( grammar->preterm_str, 0, v );
    
    /* IuWFNg𐶐Ă (gȂǂƈS) */
    cat = ALLOC( Category );
    cat->id = 0;
    cat->string = v; /* v  gram->nonterm_str ȂǂŃ}[N̂ňS */
    v = Data_Wrap_Struct( rb_cGrammar_Category, NULL, category_free, cat );
    
    /* nbVe[uɖIuWFNgo^ */
    grammar->cap_nt = 1024;
    grammar->size_nt = 1;
    grammar->hash_lhs = ALLOC_N( Rule_p, grammar->cap_nt );
    grammar->hash_lc = ALLOC_N( Rule_p, grammar->cap_nt );
    grammar->hash_lhs[0] = NULL;
    grammar->hash_lc[0] = NULL;
    rb_ary_store( grammar->nonterm_ary, 0, v );
    
    grammar->cap_pt = 1024;
    grammar->size_pt = 1;
    grammar->hash_lc_pt = ALLOC_N( Rule_p, grammar->cap_pt );
    grammar->hash_lc_pt[0] = NULL;
    rb_ary_store( grammar->preterm_ary, 0, v );
}

static VALUE cfg_initialize ( int argc, VALUE * argv, VALUE self )
{
    VALUE volatile root;
    VALUE volatile firstset;
    Grammar* gram;
    
    Data_Get_Struct( self, Grammar, gram );
    rb_scan_args( argc, argv, "02", &root, &firstset );
    
    gram->ruby_object = self; /* ruby_object ͎g (Ruby IuWFNg) 悤ɂ */
    
    /* f[^̏ */
    gram->grammar_property = rb_hash_new();
    gram->block_property = Qnil;
    gram->category_map = rb_hash_new();
    gram->nonterm_ary = rb_ary_new();
    gram->nonterm_str = rb_ary_new();
    gram->preterm_ary = rb_ary_new();
    gram->preterm_str = rb_ary_new();
    gram->rule_ary = rb_ary_new();
    gram->rule_buf = rb_ary_new();
    gram->rule_str = rb_hash_new();
    gram->first_map = Qnil;
    
    /* ʈ̋LɊւ鏉 */
    cfg_initialize_category( gram );
    
    /* ^ĂȂꍇ͋̂܂܂ŏI */
    if( NIL_P(root) )
        return self;
    
    /* z񂪓nꂽꍇ̓f[^łɓǂݍł̂Ƃď */
    if( TYPE(root) == T_ARRAY ) {
        cfg_initialize_array( gram, root, firstset );
    }
    else {
        /* ubNf[^ǂݍł */
        if( !RTEST( cfg_proper_p( Qnil, root ) ) )
            rb_raise( rb_eArgError, "Ext::Grammar requires <pGrammar>" );
        cfg_initialize_block( gram, root );
    }
    
    return self;
}

static VALUE cfg_new ( int argc, VALUE * argv, VALUE klass )
{
    VALUE volatile obj;
    Grammar* gram;
    
    gram = rb_alloc_zero( Grammar, 1 );
    obj = Data_Wrap_Struct( klass, cfg_mark, cfg_free, gram );
    rb_obj_call_init( obj, argc, argv );
    
    return obj;
}


/*   Lo^
----------------------------------------------------------------------
 - L݂ꍇ 0 ԂB݂ȂƂ 1 ԂB
 - *objptr ͋LIuWFNgŏB
   *objptr ɂ͋LIuWFNgi[ĂB
*/
int cfg_add_symbol ( Grammar * gram, VALUE * objptr )
{
    VALUE volatile str;
    VALUE volatile cat;
    int  newid;
    
    /* freeze() ĂsłȂ */
    if( OBJ_FROZEN(gram->ruby_object) )
        rb_error_frozen( rb_class2name(rb_cGrammar) );
    
    /* L݂邩𒲂ׂ */
    str = rb_String( *objptr );
    cat = rb_hash_aref( gram->category_map, str );
    if( !NIL_P(cat) ) {
        *objptr = cat;
        return 0;
    }
    
    /* L𐶐
     *  - Xgւ̓o^ category_new ōsB
     *  - @nbVp̃obt@͂ō쐬B
     */
    if( strparse_is_nonterm( RSTRING(str)->ptr, RSTRING(str)->len ) ) {
        newid = gram->size_nt; /* I[LID 1  */
        
        if( gram->size_nt >= gram->cap_nt ) {
            gram->cap_nt = (gram->cap_nt < 1024)? 1024: gram->cap_nt * 2;
            REALLOC_N( gram->hash_lc, Rule_p, gram->cap_nt );
            REALLOC_N( gram->hash_lhs, Rule_p, gram->cap_nt );
        }
        gram->hash_lc[ gram->size_nt ] = NULL;
        gram->hash_lhs[ gram->size_nt ] = NULL;
        gram->size_nt += 1;
    }
    else {
        newid = -(gram->size_pt); /* I[LID -1  */
        
        if( gram->size_pt >= gram->cap_pt ) {
            gram->cap_pt = (gram->cap_pt < 1024)? 1024: gram->cap_pt * 2;
            REALLOC_N( gram->hash_lc_pt, Rule_p, gram->cap_pt );
        }
        gram->hash_lc_pt[ gram->size_pt ] = NULL;
        gram->size_pt += 1;
    }
    
    /* @IuWFNgւ̓o^ */
    *objptr = category_new( gram, RSTRING(str)->ptr, newid );
    return 1;
}


/*   ̋K
----------------------------------------------------------------------
 - symary  [ lhs, rhs[0], rhs[1], ... ]
*/
static Rule * cfg_find_rule ( Grammar const * gram, VALUE symary )
{
    Category* cat;
    VALUE* ptr;
    Rule* rule;
    int* rhs;
    int i, len, rhslen, lhs;
    
    rb_ary_getptr( symary, &ptr, &len );
    
    /* symary ̍ӕ擾 */
    Data_Get_Struct( ptr[0], Category, cat );
    lhs = cat->id;
    
    /* symary ̉Eӕ rhs ɃRs[ */
    rhslen = len - 1;
    rhs = ALLOCA_N( int, rhslen );
    for( i = 0; i < rhslen; ++i ) {
        Data_Get_Struct( ptr[i+1], Category, cat );
        rhs[i] = cat->id;
    }
    
    /* gram ̃f[^猟 */
    for( rule = gram->hash_lhs[lhs]; rule; rule = rule->next_lhs ) {
        if( rule->rhslen != rhslen )
            continue;
        for( i = 0; i >= 0 && i < rhslen; ++i ) {
            if( rule->rhs[i] != rhs[i] ) {
                i = -1;
                break;
            }
        }
        if( i > 0 )
            return rule;
    }
    
    return NULL;
}


/*   Kǉ
----------------------------------------------------------------------
   ̋K݂ȂꍇA͋󕶎񂾂ꍇɂ̂
NULL ԂB

   allowprop ^̏ꍇ ptr ɃvpeB񂪌̂B
Ȕꍇ͋ȂB

   ruleptr  NULL ȊȌꍇAɑΉ鐶Kւ̃|C^
܂B
*/
static Rule* cfg_add_rule ( Grammar * gram, const char * ptr, int allowprop, Rule_p* ruleptr )
{
    VALUE volatile obj;
    VALUE volatile rhs;
    VALUE volatile lhs;
    VALUE volatile prop;
    Rule_p rule, dummy;
    int i, newid, newcount;
    
    /* freeze() ĂsłȂ */
    if( OBJ_FROZEN(gram->ruby_object) )
        rb_error_frozen( rb_class2name(rb_cGrammar) );
    
    if( !ruleptr )
        ruleptr = &dummy;
    
    if( allowprop ) {
        /* vpeB̉͂s */
        prop = rb_ary_new();
        ptr = strparse_parse_rule_property( ptr, prop );
    }
    else {
        /* ͕̂@K̂ */
        prop = Qnil;
    }
    rhs = strparse_parse_rule( ptr ); /* rhs ƌ擪ɂ lhs  */
    
    if( RARRAY(rhs)->len <= 0 ) {
        *ruleptr = NULL;
        return NULL; /* ̕񂾂ꍇ */
    }
    
    newcount = 0;
    for( i = 0; i < RARRAY(rhs)->len; ++i )
        newcount += cfg_add_symbol( gram, &(RARRAY(rhs)->ptr[i]) );
    
    if( newcount == 0 ) {
        /* VLĂȂ̂œ̋K邩Ȃ */
        rule = cfg_find_rule( gram, rhs );
        if( rule ) {
            *ruleptr = rule;
            return rule;
        }
    }
    
    /* KIuWFNg𐶐 (o^ rule_set_id s) */
    lhs = rb_ary_shift( rhs );
    obj = rule_new( gram, lhs, rhs );
    newid = RARRAY(gram->rule_ary)->len;
    rule_set_id( obj, newid );
    if( allowprop ) {
        for( i = 0; i < RARRAY(prop)->len; i += 2 )
            rule_set_prop( obj, RARRAY(prop)->ptr[i*2], RARRAY(prop)->ptr[i*2+1] );
    }
    
    Data_Get_Struct( obj, Rule, *ruleptr );
    return NULL;
}

Rule* cfg_add_rule_with_prop ( Grammar* gram, const char * ptr, Rule_p* ruleptr )
{
    return cfg_add_rule( gram, ptr, 1, ruleptr );
}

Rule* cfg_add_rule_without_prop ( Grammar* gram, const char * ptr, Rule_p* ruleptr )
{
    return cfg_add_rule( gram, ptr, 0, ruleptr );
}


/*   @KubN (pGrammar/dRules) ɂ鏉
------------------------------------------------------------------- */

static VALUE cfg_initialize_rule_itr ( VALUE line, Grammar * arg )
{
    Rule_p rule;
    VALUE volatile str;
    char *p;
    
    /* # ȍ~菜 */
    line = rb_String(line);
    p = strchr( RSTRING(line)->ptr, '#' );
    if( p )
        *p = '\0';
    
    rule = cfg_add_rule_with_prop( arg, RSTRING(line)->ptr, &rule ); /* vpeB\͂ */
    if( rule ) {
        str = rule_typestring(0, NULL, rule->ruleobj);
        rb_raise( rb_eNLPException, "Found same rule. (%s)", RSTRING(str)->ptr );
    }
    
    return Qnil;
}

static void cfg_initialize_rule ( Grammar * gram, VALUE block )
{
    /* block ̊esǂݍł */
    rb_iterate( rb_each, block, cfg_initialize_rule_itr, (VALUE)gram );
}


/*   FIRST WubN (pGrammar/dFIRST) ɂ鏉
------------------------------------------------------------------- */

static void cfg_initialize_FIRST ( Grammar * gram, VALUE block )
{
    const ID id_gets = rb_intern("gets");
    VALUE volatile top;
    VALUE volatile ary;
    VALUE volatile tbl;
    VALUE volatile line;
    VALUE list;
    int i;
    
    tbl = rb_hash_new();
    gram->first_map = tbl;
    
    while( !NIL_P(line = rb_funcall(block, id_gets, 0)) ) {
        ary = strparse_parse_firstset( RSTRING(line)->ptr );
        if( NIL_P(ary) )
            continue;
        
        for( i = 0; i < RARRAY(ary)->len; ++i )
            cfg_add_symbol( gram, &RARRAY(ary)->ptr[i] );
        
        top = rb_ary_shift(ary);
        list = rb_hash_aref( tbl, top );
        
        if( NIL_P(list) ) {
            /* ܂Xg݂ȂƂ */
            rb_hash_aset( tbl, top, ary );
        }
        else {
            /* łɃXg݂Ƃ͊̂̂ɘA */
            rb_ary_concat( list, ary );
        }
    }
}


/*   Ruby f[^ɂ鏉
----------------------------------------------------------------------
 - gram->first_map()  Ruby IuWFNgƂč\zB
*/

static void cfg_initialize_array_rule ( Grammar* gram, VALUE ruleary )
{
    int    len = RARRAY(ruleary)->len;
    VALUE* ptr = RARRAY(ruleary)->ptr;
    
    VALUE volatile line;
    Rule* rule;
    int i;
    
    for( i = 0; i < len; ++i ) {
        line = rb_String( ptr[i] );
        rule = cfg_add_rule_with_prop( gram, RSTRING(line)->ptr, NULL );
        if( rule )
            rb_raise( rb_eNLPException, "Found same rule, ``%s''.", STR2CSTR(rule_typestring(0,NULL,rule->ruleobj)) );
    }
}

static VALUE cfg_initialize_array_first ( Grammar* gram, VALUE firstary )
{
    int    len = RARRAY(firstary)->len;
    VALUE* ptr = RARRAY(firstary)->ptr;
    
    VALUE volatile table, value;
    VALUE key, pair;
    int i, j;
    
    if( len <= 0 )
        return Qnil;
    table = rb_hash_new();
    
    for( i = 0; i < len; ++i ) {
        pair = ptr[i];
        
        Check_Type( pair, T_ARRAY );
        if( RARRAY(pair)->len != 2 )
            rb_raise( rb_eNLPException, "Broken FIRST array." );
        
        key = RARRAY(pair)->ptr[0];
        value = rb_Array( RARRAY(pair)->ptr[1] );
        
        /* LIuWFNgɕϊĂ */
        for( j = 0; j < RARRAY(value)->len; ++j )
            cfg_add_symbol( gram, &RARRAY(value)->ptr[j] );
        cfg_add_symbol( gram, &key );
        
        rb_hash_aset( table, key, value );
    }
    
    return table;
}

static void cfg_initialize_array ( Grammar* gram, VALUE ruleary, VALUE firstary )
{
    /* ruleary ̏ */
    cfg_initialize_array_rule( gram, rb_Array(ruleary) );
    
    /* firstary ̏ */
    gram->first_map = cfg_initialize_array_first( gram, rb_Array(firstary) );
}


/*   zւ̕ϊ
------------------------------------------------------------------- */

static VALUE cfg_rules ( VALUE self )
{
    Grammar* gram;
    
    Data_Get_Struct( self, Grammar, gram );
    
    return rb_obj_dup( gram->rule_ary );
}

static VALUE cfg_categories ( VALUE self )
{
    VALUE volatile ary;
    Grammar* gram;
    VALUE *ptr;
    int i, len;
    
    Data_Get_Struct( self, Grammar, gram );
    ary = rb_ary_new2( cfg_size_symbol(gram) - 1 );
    
    rb_ary_getptr( gram->nonterm_ary, &ptr, &len );
    for( i = 1; i < len; ++i )
        rb_ary_push( ary, ptr[i] );
    
    rb_ary_getptr( gram->preterm_ary, &ptr, &len );
    for( i = 1; i < len; ++i )
        rb_ary_push( ary, ptr[i] );
    
    return ary;
}


/*   @̏o
------------------------------------------------------------------- */

static void cfg_display_output_open ( VALUE out, VALUE blockprop )
{
    VALUE volatile ary;
    VALUE volatile str;
    VALUE*  ptr;
    VALUE  pair;
    int  i, len;
    
    str = rb_str_new(NULL, 0);
    ary = rb_Array( blockprop );
    
    rb_str_cat2( str, "<pGrammar" );
    rb_ary_getptr( ary, &ptr, &len );
    for( i = 0; i < len; ++i, ++ptr ) {
        pair = *ptr;
        if( TYPE(pair) == T_ARRAY && RARRAY(pair)->len == 2 ) {
            rb_str_cat( str, " ", 1 );
            rb_str_append( str, RARRAY(pair)->ptr[0] );
            rb_str_cat( str, "=\"", 2 );
            rb_str_append( str, RARRAY(pair)->ptr[1] );
            rb_str_cat( str, "\"", 1 );
        }
    }
    rb_str_cat( str, ">\n", 2 );
    
    rb_io_write( out, str );
}

static void cfg_display_output_rule ( VALUE out, Grammar* grammar )
{
    Rule_p  rule;
    VALUE*  ptr;
    char    buf [ 4096 ];
    char*   p;
    int     i, len;
    
    rb_io_write( out, rb_str_new2("<dRules>\n") );
    
    rb_ary_getptr( grammar->rule_ary, &ptr, &len );
    for( i = 0; i < len; ++i ) {
        Data_Get_Struct( ptr[i], Rule, rule );
        p = rule_to_s_internal( rule, buf );
        *(p++) = '\n';
        *(p++) = '\0';
        rb_io_write( out, rb_str_new( buf, p - buf ) );
    }
    
    rb_io_write( out, rb_str_new2("</dRules>\n") );
}

static void cfg_display_output_first ( VALUE out, Grammar* grammar )
{
    VALUE volatile str;
    VALUE volatile ary;
    VALUE volatile key;
    VALUE  pair, *ptr, *vptr;
    char   buf [ 2048 ];
    int    i, j, len, vlen;
    
    ary = rb_Array( grammar->first_map ); /* Hash --> Array ϊ */
    rb_ary_getptr( ary, &ptr, &len );
    if( len > 0 ) {
        rb_io_write( out, rb_str_new2("<dFIRST>\n") );
        
        for( i = 0; i < len; ++i ) {
            pair = ptr[i];
            if( TYPE(pair) == T_ARRAY && RARRAY(pair)->len == 2 ) {
                rb_ary_getptr( RARRAY(pair)->ptr[1], &vptr, &vlen );
                if( vlen > 0 ) {
                    str = rb_str_new(NULL, 0);
                    
                    key = rb_String( RARRAY(pair)->ptr[0] );
                    j = sprintf( buf, "FIRST(%s) =", RSTRING(key)->ptr );
                    rb_str_cat( str, buf, j );
                    
                    for( j = 0; j < vlen; ++j ) {
                        key = rb_String( vptr[j] );
                        rb_str_cat( str, " ", 1 );
                        rb_str_append( str, key );
                    }
                    rb_str_cat( str, "\n", 1 );
                    
                    rb_io_write( out, str );
                }
            }
        }
        
        rb_io_write( out, rb_str_new2("</dFIRST>\n") );
    }
}

/** @brief <pGrammar> ubN̏o
 *
 * @ubNo͂܂B
 * Ext::Grammar#display() ɑΉĂ܂B
 *
 * ɂ͏o͐̃IuWFNgw肵܂Bȗ nil w肳ꂽꍇ $defout g܂B
 */
static VALUE cfg_display ( int argc, VALUE * argv, VALUE self )
{
    VALUE volatile blockprop;
    VALUE volatile out;
    Grammar*  gram;
    
    rb_scan_args( argc, argv, "01", &out );
    if( NIL_P(out) )
        out = rb_gv_get("$defout");
    
    Data_Get_Struct( self, Grammar, gram );
    cfg_display_output_open( out, gram->block_property );
    cfg_display_output_rule( out, gram );
    cfg_display_output_first( out, gram );
    
    rb_io_write( out, rb_str_new2("</pGrammar>\n") );
}


/*   @KXg̎擾
----------------------------------------------------------------------
*/

Rule* cfg_get_rulelist_lhs ( Grammar * gram, int key )
{
    if( key < 0 || key >= gram->size_nt )
        rb_raise( rb_eNLPException, "Out of range. (inputted LHS-key=%d, LHS MAX-ID=%d)", key, gram->size_nt );
    
    return gram->hash_lhs[ key ];
}

Rule* cfg_get_rulelist_lc ( Grammar * gram, int key )
{
    Rule* ret;
    
    if( key < 0 ) {
        key = -key;
        if( key >= gram->size_pt )
            rb_raise( rb_eNLPException, "Out of range. (inputted LC-key=%d, preterm MIN-ID=%d)", -key, -(gram->size_pt - 1) );
        ret = gram->hash_lc_pt[ key ];
    }
    else {
        if( key >= gram->size_nt )
            rb_raise( rb_eNLPException, "Out of range. (inputted LC-key=%d, nonterm MAX-ID=%d)", key, gram->size_nt - 1 );
        ret = gram->hash_lc[ key ];
    }
    
    return ret;
}


/*   w肳ꂽŕ\镶@K𓾂
------------------------------------------------------------------- */

VALUE cfg_get_rule_internal ( Grammar* gram, VALUE keystr )
{
    Rule_p  rule;
    VALUE   ruleobj;
    
    keystr = rb_String(keystr);
    ruleobj = rb_hash_aref( gram->rule_str, keystr );
    if( NIL_P(ruleobj) ) /* 񂩂ł͌Ȃ̂ŕ@Kɒǉ */
    {
        /* vpeB̍\͍͂sȂ */
        cfg_add_rule_without_prop( gram, RSTRING(keystr)->ptr, &rule );
        
        if( rule )
            ruleobj = rule->ruleobj;
        else
            ruleobj = Qnil; /* KȂǉȂ --> ̋K */
    }
    
    return ruleobj;
}

static VALUE cfg_get_rule_m ( VALUE self, VALUE key )
{
    Grammar * gram;
    
    Data_Get_Struct( self, Grammar, gram );
    
    return cfg_get_rule_internal( gram, key );
}

Rule * cfg_get_rule ( Grammar const * grammar, int id )
{
    VALUE  ruleobj;
    Rule * rule;
    
    if( id < 0 || id >= cfg_size_rule(grammar) )
        return NULL;
    ruleobj = cfg_rule_at(grammar, id);
    Data_Get_Struct( ruleobj, Rule, rule );
    
    return rule;
}


/*  K̒ǉ
------------------------------------------------------------------- */

static VALUE cfg_store ( VALUE self, VALUE string )
{
    Grammar * gram;
    Rule_p rule;
    Rule_p exist;
    VALUE ruleobj;
    
    Data_Get_Struct( self, Grammar, gram );
    string = rb_String(string);
    ruleobj = rb_hash_aref( gram->rule_str, string );
    if( !NIL_P(ruleobj) )
        return ruleobj; /* łɋKIuWFNg݂Ă (LbVŔ) */
    
    exist = cfg_add_rule_with_prop( gram, RSTRING(string)->ptr, &rule );
    if( exist )
        return exist->ruleobj; /* łɓKIuWFNg݂Ăꍇ */
    
    return Qnil; /* VɋKǉꂽꍇ */
}


/*  @IuWFNg̊gpvpeBANZX\bh
------------------------------------------------------------------- */

static VALUE cfg_set_property ( VALUE self, VALUE key, VALUE value )
{
    Grammar* gram;
    
    Data_Get_Struct( self, Grammar, gram );
    
    return rb_hash_aset( gram->grammar_property, rb_String(key), value );
}

VALUE cfg_get_property ( VALUE self, VALUE key )
{
    Grammar* gram;
    
    Data_Get_Struct( self, Grammar, gram );
    key = rb_String(key);
    
    return rb_hash_aref( gram->grammar_property, key );
}

static VALUE cfg_get_property_object ( VALUE self )
{
    Grammar* gram;
    
    Data_Get_Struct( self, Grammar, gram );
    
    return gram->grammar_property;
}

static VALUE cfg_get_block_property ( VALUE self )
{
    Grammar* gram;
    
    Data_Get_Struct( self, Grammar, gram );
    
    return rb_obj_dup( gram->block_property );
}

static VALUE cfg_set_block_property ( VALUE self, VALUE value )
{
    Grammar* gram;
    
    Data_Get_Struct( self, Grammar, gram );
    if( TYPE(value) != T_HASH )
        rb_raise( rb_eArgError, "Ext::Grammar: set_block_property() requires Hash for value." );
    gram->block_property = value;
    
    return gram->block_property;
}


/*   top WJĂ leaf [Ɍ邩𒲂ׂ
------------------------------------------------------------------- */
static VALUE cfg_reachable_p ( VALUE self, VALUE top, VALUE leaf )
{
    Grammar* gram;
    Category* cat;
    VALUE* ptr;
    VALUE list;
    int i, id, len;
    
    Data_Get_Struct( self, Grammar, gram );
    
    if( NIL_P(gram->first_map) )
        rb_raise( rb_eNLPException, "No FIRST-set found." );
    
    top = cfg_getcat_internal( gram, top );
    leaf = cfg_getcat_internal( gram, leaf );
    Data_Get_Struct( top, Category, cat );
    id = cat->id;
    
    list = rb_hash_aref( gram->first_map, top );
    if( NIL_P(list) ) {
        rb_raise( rb_eNLPException, "FIRST(%s) not defined.", RSTRING(cat->string)->ptr );
    }
    
    /* list  leaf ΂ */
    len = RARRAY(list)->len;
    ptr = RARRAY(list)->ptr;
    for( ; len > 0; --len ) {
        Data_Get_Struct( *ptr, Category, cat );
        if( cat->id == id )
            return Qtrue;
        ++ptr;
    }
    
    return Qfalse;
}

VALUE rule_to_a ( VALUE self )
{
    VALUE volatile ary;
    Grammar* gram;
    Rule* rule;
    int i;
    
    Data_Get_Struct( self, Rule, rule );
    Data_Get_Struct( rule->gramobj, Grammar, gram );
    ary = rb_ary_new2( 1 + rule->rhslen );
    
    rb_ary_push( ary, cfg_get_category(gram, rule->lhs) );
    for( i = 0; i < rule->rhslen; ++i )
        rb_ary_push( ary, cfg_get_category(gram, rule->rhs[i]) );
    
    return ary;
}

int cfg_symbol_p ( Grammar const * grammar, int symbol )
{
    return (symbol >= grammar->size_nt || symbol <= -grammar->size_pt)? 0: 1;
}


/*  ̊gW[̏R[h
------------------------------------------------------------------- */
void init_local_gvar ( void )
{
    VALUE volatile v;
    
    /* K̃vpeBIuWFNg */
    local_object_cache = rb_ary_new();
    rb_global_variable( &local_object_cache );
    
    v = rb_str_new( "p", 1 );
    rbstr_prop_pr = v;
    rb_ary_push( local_object_cache, v );
    
    v = rb_str_new( "head", 4 );
    rbstr_prop_pr = v;
    rb_ary_push( local_object_cache, v );
    
    /* \bhID */
    ID_strip_bang = rb_intern("strip!");
}

void init_grammar ( void )
{
    init_local_gvar(); /* ̃W[ŎgpO[o Ruby ϐ̏ */
    
    /* lCeBu Rule NX̒`
    ---------------------------------------- */
    rb_cGrammar = rb_define_class_under( rb_mExt, "Grammar", rb_cData );
    rb_cGrammar_Rule = rb_define_class_under( rb_cGrammar, "Rule", rb_cData );
    rb_cGrammar_Category = rb_define_class_under( rb_cGrammar, "Category", rb_cData );
    
    /* Ext::Grammar - NX\bh
    ---------------------------------------- */
    rb_define_singleton_method( rb_cGrammar, "new", cfg_new, -1 );
    rb_define_singleton_method( rb_cGrammar, "proper?", cfg_proper_p, 1 );
    rb_define_singleton_method( rb_cGrammar, "===", cfg_proper_p, 1 );
    
    /* Ext::Grammar::Category - \bh (L)
    ---------------------------------------- */
    rb_define_method( rb_cGrammar_Category, "id", category_id, 0 );
    rb_define_method( rb_cGrammar_Category, "to_i", category_id, 0 );
    rb_define_method( rb_cGrammar_Category, "to_int", category_id, 0 );
    rb_define_method( rb_cGrammar_Category, "nonterm?", category_nonterm_p, 0 );
    rb_define_method( rb_cGrammar_Category, "preterm?", category_preterm_p, 0 );
    rb_define_method( rb_cGrammar_Category, "inspect", category_inspect, 0 );
    rb_define_method( rb_cGrammar_Category, "eql?", category_eql_p, 1 );
    rb_define_method( rb_cGrammar_Category, "==", category_eql_p, 1 );
    rb_define_method( rb_cGrammar_Category, "hash", category_hash, 0 );
    rb_define_method( rb_cGrammar_Category, "name", category_name, 0 );
    rb_define_method( rb_cGrammar_Category, "to_s", category_name, 0 );
    rb_define_method( rb_cGrammar_Category, "to_str", category_name, 0 );
    rb_define_method( rb_cGrammar_Category, "<=>", category_compare, 1 );
    
    /* Ext::Grammar::Rule - \bh (@K)
    ---------------------------------------- */
    rb_define_method( rb_cGrammar_Rule, "eql?", rule_eql_p, 1 );
    rb_define_method( rb_cGrammar_Rule, "==", rule_eql_p, 1 );
    rb_define_method( rb_cGrammar_Rule, "<=>", rule_compare, 1 );
    rb_define_method( rb_cGrammar_Rule, "hash", rule_hash, 0 );
    rb_define_method( rb_cGrammar_Rule, "get_prop", rule_get_prop, 1 );
    rb_define_method( rb_cGrammar_Rule, "[]", rule_get_prop, 1 );
    rb_define_method( rb_cGrammar_Rule, "set_prop", rule_set_prop, 2 );
    rb_define_method( rb_cGrammar_Rule, "[]=", rule_set_prop, 2 );
    rb_define_method( rb_cGrammar_Rule, "inspect", rule_inspect, 0 );
    rb_define_method( rb_cGrammar_Rule, "to_s", rule_to_s, 0 );
    rb_define_method( rb_cGrammar_Rule, "to_str", rule_to_s, 0 );
    rb_define_method( rb_cGrammar_Rule, "typestring", rule_typestring, -1 );
    rb_define_method( rb_cGrammar_Rule, "lhs", rule_get_lhs, 0 );
    rb_define_method( rb_cGrammar_Rule, "rhs", rule_get_rhs, 0 );
    rb_define_method( rb_cGrammar_Rule, "id", rule_get_id, 0 );
    rb_define_method( rb_cGrammar_Rule, "to_i", rule_get_id, 0 );
    rb_define_method( rb_cGrammar_Rule, "to_int", rule_get_id, 0 );
    rb_define_method( rb_cGrammar_Rule, "to_a", rule_to_a, 0 );
    rb_define_method( rb_cGrammar_Rule, "to_ary", rule_to_a, 0 );
    
    /* Ext::Grammar - \bh (@)
    ---------------------------------------- */
    rb_define_method( rb_cGrammar, "initialize", cfg_initialize, -1 );
    rb_define_method( rb_cGrammar, "size_rule", cfg_size_rule_m, 0 );
    rb_define_method( rb_cGrammar, "size_nonterm", cfg_size_nonterm_m, 0 );
    rb_define_method( rb_cGrammar, "size_preterm", cfg_size_preterm_m, 0 );
    rb_define_method( rb_cGrammar, "size_symbol", cfg_size_symbol_m, 0 );
    rb_define_method( rb_cGrammar, "FIRST", cfg_FIRST, 0 );
    rb_define_method( rb_cGrammar, "has_FIRST?"	, cfg_has_FIRST_p, 0 );
    rb_define_method( rb_cGrammar, "rules", cfg_rules, 0 );
    rb_define_method( rb_cGrammar, "to_a", cfg_rules, 0 );
    rb_define_method( rb_cGrammar, "to_ary", cfg_rules, 0 );
    rb_define_method( rb_cGrammar, "categories"	, cfg_categories, 0 );
    rb_define_method( rb_cGrammar, "getcat", cfg_getcat, 1 );
    rb_define_method( rb_cGrammar, "display", cfg_display, -1 );
    rb_define_method( rb_cGrammar, "getrule", cfg_get_rule_m, 1 );
    rb_define_method( rb_cGrammar, "[]", cfg_get_rule_m, 1 );
    rb_define_method( rb_cGrammar, "store", cfg_store, 1 );
    rb_define_method( rb_cGrammar, "<<", cfg_store, 1 );
    rb_define_method( rb_cGrammar, "set_property", cfg_set_property, 2 );
    rb_define_method( rb_cGrammar, "get_property", cfg_get_property, 1 );
    rb_define_method( rb_cGrammar, "property", cfg_get_property_object, 0 );
    rb_define_method( rb_cGrammar, "block_property", cfg_get_block_property, 0 );
    rb_define_method( rb_cGrammar, "block_property=", cfg_set_block_property, 1 );
    rb_define_method( rb_cGrammar, "reachable?", cfg_reachable_p, 2 );
}


