/*                                                     project Gemini 
----------------------------------------------------------------------
    \͌ʑp[eBeB
      $Id: itemset.c,v 1.22 2002/11/21 13:46:48 abe Exp $

       Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define DEBUGSW 0
#include "nlputil.h"
#include <time.h>
#include <stdarg.h>

/* ł֐̓}NŒ`邩ǂ */
#define USE_MACRO 1


/* O[oϐ */
VALUE rb_cItemSet;


/* vg^Cv */
static void itemset_initialize_construct ( ItemSet * itemset );


/* ItemSet ̓e\
------------------------------------------------------------------- */

void itemset_debug_display ( FILE * out, ItemSet const * itemset )
{
    VALUE volatile strobj;
    ParsedItem const * item;
    Rule const * rule;
    int  i, j, right;
    
    fprintf( out, "[ItemSet] >>>>>>>>>>>>>>>>>>>>\n" );
    for( i = 0; i < itemset->itembuf_size; ++i ) {
        item = &itemset->itembuf[i];
        rule = item->rule;
        
        if( rule ) {
            fprintf( out, "[%d", item->left );
            right = item->left;
            for( j = 0; j < rule->rhslen; ++j ) {
                if( item->child[j] )
                    right = item->child[j]->right;
                else
                    ++right;
                fprintf( out, ",%d", right );
            }
            fprintf( out, "] " );
            strobj = rule_typestring( 0, NULL, rule->ruleobj );
            fprintf( out, "%s\n", RSTRING(strobj)->ptr );
        }
        else {
            fprintf( out, "[%d,%d] %s\n", item->left, item->right, cfg_getcatname(itemset->grammar, item->top) );
        }
    }
    fprintf( out, "[ItemSet] <<<<<<<<<<<<<<<<<<<<\n" );
}

static VALUE itemset_display ( VALUE self, VALUE out )
{
    ItemSet*  itemset;
    Grammar*  gram;
    ParsedItem_p  item, child;
    Rule_p  rule;
    
    VALUE volatile v;
    char   buf[4096];
    char  *p;
    int    i, j;
    
    Data_Get_Struct( self, ItemSet, itemset );
    gram = itemset->grammar;
    for( i = 0; i < itemset->itembuf_size; ++i ) {
        item = &itemset->itembuf[i];
        
        p = buf;
        v = rb_String( cfg_get_category( gram, item->top ) );
        p += sprintf( p, "<%p> %s(%d,%d)", item, RSTRING(v)->ptr, item->left, item->right );
        rule = item->rule;
        if( rule ) {
            p += sprintf( p, " : " );
            p = rule_typestring_internal( rule, p, -1 );
        }
        p += sprintf( p, "\n" );
        rb_io_write( out, rb_str_new2(buf) );
        
        if( item->child ) {
            if( !rule ) {
                sprintf( buf, " ... item->child is NULL, but rule is also NULL.\n" );
                rb_io_write( out, rb_str_new2(buf) );
            }
            else {
                for( j = 0; j < rule->rhslen; ++j ) {
                    child = item->child[j];
                    if( !child )
                        continue;
                    v = rb_String( cfg_get_category(gram, child->top) );
                    sprintf( buf, " [%d:%p] %s(%d,%d)\n", j+1, child, RSTRING(v)->ptr, child->left, child->right );
                    rb_io_write( out, rb_str_new2(buf) );
                }
            }
        }
    }
    
    return self;
}


/* Ext::ItemSet ̃RXgN^EfXgN^
------------------------------------------------------------------- */

static void itemset_free ( ItemSet* itemset )
{
    if( itemset ) {
        ruby_xfree( itemset->hashbuf );
        ruby_xfree( itemset->headbuf );
        ruby_xfree( itemset->childbuf );
        ruby_xfree( itemset->itembuf );
        ruby_xfree( itemset );
    }
}

static void itemset_mark ( ItemSet* itemset )
{
    if( itemset ) {
        rb_gc_mark( itemset->rulemap );
        rb_gc_mark( itemset->start_name );
        rb_gc_mark( itemset->grammar->ruby_object );
        
        rb_gc_mark( itemset->rule_storage );
        rb_gc_mark( itemset->pos_storage );
    }
}

static VALUE itemset_new ( int argc, VALUE * argv, VALUE klass )
{
    ItemSet* itemset;
    VALUE volatile obj;
    
    itemset = rb_alloc_zero( ItemSet, 1 );
    
    obj = Data_Wrap_Struct( klass, itemset_mark, itemset_free, itemset );
    rb_obj_call_init( obj, argc, argv );
    
    return obj;
}

static int itemset_initialize_itr_command ( ItemSet * itemset, const char * p )
{
    /* ItemSet : wߍs̏ꍇ */
    VALUE volatile v;
    const char * q;
    int n;
    
    if( *p != '#' )
        return 0;
    ++p;
    
    if( p[0] == '@' && p[1] == 'R' && p[2] == 'O' && p[3] == 'O' && p[4] == 'T' ) {
        /* #@ROOT = ... */
        p += 5;
        STRPARSE_SKIP_WS( p );
        if( *p == '=' ) { /* = ͖Ă */
            ++p;
            STRPARSE_SKIP_WS( p );
        }
        q = p;
        while( *p != '\r' && *p != '\n' && *p != '\0' )
            ++p;
        
        /* Ƃ肠ƂĊi[Ă (ŕϊ) */
        itemset->start_name = rb_str_new( q, p-q );
    }
    else {
        /* #ddd ... */
        q = p;
        n = 0;
        while( '0' <= *p && *p <= '9' )
            n = n*10 + (*(p++) - '0');
        
        if( p != q ) {
            STRPARSE_SKIP_WS( p );
            q = p;
            while( *p != '\r' && *p != '\n' && *p != '\0' )
                ++p;
            v = cfg_get_rule_internal( itemset->grammar, rb_str_new( q, p-q ) );
            if( NIL_P(v) )
                rb_raise( rb_eNLPException, "Empty rule string." );
            rb_ary_store( itemset->rulemap, n, v );
        }
    }
    
    return 1;
}

static int itemset_initialize_itr_item ( ItemSet * itemset, const char * p )
{
    /* ItemSet : 錾s̏ꍇ */
    VALUE volatile ruleobj = Qnil;
    VALUE volatile ary = rb_ary_new();
    VALUE volatile v;
    VALUE catobj = Qnil;
    VALUE regobj;
    VALUE arg;
    Rule_p rule;
    Category_p cat;
    
    v = strparse_parse_item_poslist( p, ary ); /* v ɂ͈ʒuXǧ̕ (K) i[ */
    if( NIL_P(v) )
        return 0; /* s */
    
    if( RARRAY(ary)->len == 0 )
        rb_raise( rb_eNLPException, "Empty position list." );
    
    if( TYPE(v) == T_FIXNUM ) {
        /* f[^l̎͋Kԍ\Ă */
        ruleobj = rb_ary_entry( itemset->rulemap, FIX2INT(v) );
        if( NIL_P(ruleobj) )
            rb_raise( rb_eNLPException, "Rule number %d out of range.", FIX2INT(v) );
    }
    else {
        /* f[^̎ (lȊO͕ɂĂ܂) Kg[\Ă */
        v = rb_String(v);
        if( RSTRING(v)->ptr[0] == '@' ) {
            v = rb_str_new2( &RSTRING(v)->ptr[1] );
            catobj = v;
            cfg_add_symbol( itemset->grammar, &catobj );
            if( NIL_P(catobj) )
                rb_raise( rb_eNLPException, "Failed to register new category, %s.", RSTRING(v)->ptr );
        }
        else {
            ruleobj = cfg_get_rule_internal( itemset->grammar, v );
            if( NIL_P(ruleobj) )
                rb_raise( rb_eNLPException, "No such rule, %s.", RSTRING(v)->ptr );
        }
    }
    
    if( !NIL_P(ruleobj) ) {
        /* Rule s̃`FbN */
        rule = nlp_Rule( ruleobj );
        if( rule->rhslen + 1 != RARRAY(ary)->len )
            rb_raise( rb_eNLPException, "Position list is short of long. (poslist=%d, rhslen=%d)", RARRAY(ary)->len, rule->rhslen );
        itemset->childbuf_size += rule->rhslen + 1;
        regobj = ruleobj;
    }
    else {
        /* Category s (Terminal s) ̃`FbN */
        cat = nlp_Category( catobj );
        if( cat->id == 0 )
            rb_raise( rb_eNLPException, "Category ID is zero." );
        if( RARRAY(ary)->len != 2 )
            rb_raise( rb_eNLPException, "Position list is short or long. (poslist=%d, not 2)", RARRAY(ary)->len );
        regobj = catobj;
    }
    
    /* ̓o^ (\z͌ŕʂɍŝł̓obt@ɓ邾) */
    rb_ary_push( itemset->rule_storage, regobj );
    rb_ary_push( itemset->pos_storage, ary );
    
    return 1;
}

static VALUE itemset_initialize_itr ( VALUE line, ItemSet* itemset )
{
    const char *p;
    
    if( TYPE(line) != T_STRING )
        line = rb_String( line );
    p = RSTRING(line)->ptr;
    
    /* wߍs (#...) ̏ꍇ */
    if( itemset_initialize_itr_command(itemset, p) )
        return Qnil;
    
    /* 錾 ([*,...] ...) ̏ꍇ */
    if( itemset_initialize_itr_item(itemset, p) )
        return Qnil;
    
    return Qnil;
}

static VALUE itemset_initialize ( int argc, VALUE * argv, VALUE self )
{
    VALUE volatile v;
    VALUE volatile gramobj;
    VALUE volatile stream;
    ItemSet* itemset;
    
    rb_scan_args( argc, argv, "11", &stream, &gramobj );
    if( NIL_P( gramobj ) )
        gramobj = rb_gv_get( "$grammar" );
    if( CLASS_OF(gramobj) != rb_cGrammar )
        rb_raise( rb_eNLPException, "Failed to initialize Ext::ItemSet. (grammar is not Ext::Grammar)" );
    
    /* IuWFNg̃ZbgAbv */
    Data_Get_Struct( self, ItemSet, itemset );
    Data_Get_Struct( gramobj, Grammar, itemset->grammar );
    
    itemset->rulemap = rb_ary_new();
    itemset->rule_storage = rb_ary_new();
    itemset->pos_storage = rb_ary_new();
    
    /* Xg[̓ǂݍ */
    rb_iterate( rb_each, stream, itemset_initialize_itr, (VALUE)itemset );
    
    /* ftHglݒ */
    v = itemset->start_name;
    if( NIL_P(v) )
        v = rb_String( rb_gv_get( "$Start" ) );
    itemset->start_name = rb_hash_aref( itemset->grammar->category_map, v );
    if( NIL_P( itemset->start_name ) )
        rb_raise( rb_eNLPException, "No such symbol -- ``%s.''", RSTRING(v)->ptr );
    
    if( CLASS_OF(itemset->start_name) != rb_cGrammar_Category )
        rb_raise( rb_eNLPException, "Start symbol expects Ext::Grammar::Category." );
    
    /* obt@Of[^gč\z */
    itemset_initialize_construct( itemset );
    
    /* obt@Of[^͕svȂ̂ GC ɉĂ炤 */
    itemset->rule_storage = Qnil;
    itemset->pos_storage = Qnil;
    
    return self;
}


/*   f[^Zbg̍\z
----------------------------------------------------------------------
 - nbV̍\͎̒ʂB
   1. ʒu (left, right) ł̃nbV
     2. gbvLʂł̍̃Xg (next_hash N)
       3. gbvL̃Xg (next_item N)
*/

static ParsedItemptr * create_item_hash ( ItemSet* itemset )
{
    /* hash ͍ŏIIɔ͈݂͂̂ɂnbVƂȂgbvL̂͂ЂƂ */
    ParsedItem_p  *hash;
    ParsedItem_p   ptr, item;
    int  sentlen = itemset->sent_len;
    int  i, addr;
    
    hash = rb_alloc_zero( ParsedItem_p, sentlen * sentlen );
    
    itemset->headbuf_size = 0;
    for( i = 0; i < itemset->itembuf_size; ++i ) {
        item = &itemset->itembuf[ i ];
        addr = item->left + item->right * sentlen;
        
        /* hash[addr]̃Xgɓgbv̍΂̉ɗƂ */
        ptr = hash[addr];
        while( ptr && ptr->top != item->top )
            ptr = ptr->next_hash;
        
        if( ptr ) { /* gbv (ʒu) ̃Xg */
            item->next_item = ptr->next_item;
            ptr->next_item = item;
        }
        else { /* gbṽXg */
            item->next_hash = hash[addr];
            hash[addr] = item;
            ++itemset->headbuf_size;
        }
    }
    
    return hash;
}

ParsedItem * itemset_find ( ItemSet * itemset, int left, int right, int top )
{
    ParsedItem_p ptr;
    VALUE volatile str;
    
    for( ptr = itemset->hashbuf[ left + right * itemset->sent_len ]; ptr; ptr = ptr->next_hash ) {
        if( ptr->top == top )
            return ptr;
    }
    
    str = rb_String( cfg_get_category(itemset->grammar, top) );
    rb_raise( rb_eNLPException, "No such item -- %s(%d,%d)", RSTRING(str)->ptr, left, right );
    
    return NULL; /* _~[ */
}

static void itemset_initialize_setup ( ItemSet * itemset )
{
    RArray_p poscache = RARRAY(itemset->pos_storage);
    int sentlen = itemset->sent_len - 1;
    
    ParsedItem_p item;
    RArray_p posary;
    Rule_p rule;
    int left, right;
    int cat, length;
    int i, n;
    
    /* nbV̍\z */
    itemset->hashbuf = create_item_hash( itemset );
    
    /* q̘A */
    for( i = 0; i < itemset->itembuf_size; ++i ) {
        item = &itemset->itembuf[i];
        if( !item->child )
            continue; /* [͎qȂ̂ňȍ~̑삪sv */
        
        rule = item->rule;
        posary = RARRAY( poscache->ptr[i] );
        
        left = item->left;
        for( n = 0; n < rule->rhslen; ++n ) {
            right = NUM2INT( posary->ptr[n+1] );
            if( right > sentlen )
                rb_raise( rb_eNLPException, "Position list is broken. (list[%d](%d) > sentlen(%d))", n, right, sentlen );
            
            cat = rule->rhs[n];
            if( cat >= 0 )
                item->child[n] = itemset_find( itemset, left, right, rule->rhs[n] );
            else
                item->child[n] = NULL;
            
            left = right;
        }
    }
    itemset->start = itemset_find( itemset, 0, sentlen, ru_id(itemset->start_name) );
    
    /* lރXg̍쐬 */
    itemset->headbuf = ALLOC_N( ParsedItem_p, itemset->headbuf_size );
    n = 0;
    length = itemset->sent_len * itemset->sent_len;
    for( i = 0; i < length; ++i ) {
        for( item = itemset->hashbuf[i]; item; item = item->next_hash )
            itemset->headbuf[ n++ ] = item;
    }
}

static void itemset_initialize_construct ( ItemSet * itemset )
{
    RArray_p rulecache = RARRAY( itemset->rule_storage );
    RArray_p poscache = RARRAY( itemset->pos_storage );
    
    ParsedItem_p  item;
    Category_p  cat;
    RArray_p  posary;
    Rule_p  rule;
    
    VALUE  itemobj;
    int  i, right, length, childptr;
    
    /* f[^obt@̏ */
    itemset->itembuf_size = rulecache->len;
    itemset->itembuf = rb_alloc_zero( ParsedItem, itemset->itembuf_size ); /* [NAĂ */
    itemset->childbuf = ALLOC_N( ParsedItem_p, itemset->childbuf_size );
    
    /* CXg̏ */
    right = 0;
    childptr = 0;
    for( i = 0; i < itemset->itembuf_size; ++i ) {
        item = &itemset->itembuf[i]; /*  item  */
        posary = RARRAY(poscache->ptr[i]);
        
        item->left = NUM2INT( posary->ptr[0] );
        item->right = NUM2INT( posary->ptr[ posary->len - 1 ] );
        if( item->left >= item->right )
            rb_raise( rb_eNLPException, "Position list is broken. (left(%d) > right(%d))", item->left, item->right );
        if( right < item->right )
            right = item->right; /* right ͕𒲂ׂ邽߂̕ϐ */
        
        itemobj = rulecache->ptr[i];
        if( CLASS_OF(itemobj) == rb_cGrammar_Rule ) {
            /* K (Rule) */
            Data_Get_Struct( itemobj, Rule, rule );
            
            item->rule = rule;
            item->top = rule->lhs;
            item->child = &itemset->childbuf[ childptr ];
            childptr += rule->rhslen + 1;
        }
        else {
            /* [ (terminal) */
            Data_Get_Struct( itemobj, Category, cat );
            
            item->rule = NULL;
            item->top = cat->id;
            item->child = NULL;
        }
        item->data = NULL;
        item->next_item = NULL;
        item->next_hash = NULL;
    }
    itemset->sent_len = right + 1;
    
    /* ݒ肳ꂽf[^A */
    itemset_initialize_setup( itemset );
}


/*  ParsedItem ̑ݒ肷
------------------------------------------------------------------- */
struct tagItemSet_SetLex_Data
{
    VALUE str_head;
};
typedef struct tagItemSet_SetLex_Data ItemSet_SetLex_Data;

static int itemset_set_lex_itr ( ItemSet_SetLex_Data* dat, ParsedItem_p item )
{
    VALUE volatile headobj;
    Rule_p rule;
    int head;
    
    /* ݒς݂̏ꍇ */
    if( item->attr.lex != 0 )
        return item->attr.lex;
    rule = item->rule;
    
    /* @Kɕtꂽ head 擾 */
    headobj = rule_get_property( rule, dat->str_head );
    if( !NIL_P(headobj) ) {
        /* head 񂪂 */
        head = NUM2INT(rb_Integer(headobj)) - 1;
        if( head < 0 || head >= rule->rhslen ) {
            headobj = rb_String(rule_typestring(0, NULL, rule->ruleobj));
            rb_raise( rb_eNLPException, "Rule #%d head index %d out of range. (%s)", rule->id, head, RSTRING(headobj)->ptr );
        }
    }
    else if( rule->rhslen == 1 ) {
        /* Eӂ̒ 1 ̏ꍇ͗B̗vf head ƍl */
        head = 0;
    }
    else {
        /* head f[^炸łȂꍇ lex tȂ */
        head = -1;
    }
    
    /* ǂ head KpΗǂ̂킩Ăꍇ lex 󂯎 */
    if( head >= 0 ) {
        if( item->child[head] ) {
            /* I[̏ꍇ͎q`Ă炤 */
            item->attr.lex = itemset_set_lex_itr( dat, item->child[head] );
        }
        else {
            /* I[L̏ꍇ͂ꂪ lex */
            item->attr.lex = rule->rhs[head];
        }
    }
    
    return item->attr.lex;
}

ItemSet* itemset_set_lex ( ItemSet* itemset )
{
    VALUE volatile str_head = rb_str_new2("head");
    
    ItemSet_SetLex_Data dat;
    ParsedItem* ptr;
    int i;
    
    dat.str_head = str_head;
    
    /* Ô߂ɏĂ */
    ptr = itemset->itembuf;
    for( i = itemset->itembuf_size; i > 0; --i )
        (ptr++)->attr.lex = 0;
    
    /* ݒ */
    ptr = itemset->itembuf;
    for( i = itemset->itembuf_size; i > 0; --i )
        itemset_set_lex_itr( &dat, ptr++ );
    
    return itemset;
}


/*  ParsedItem ̓e\
------------------------------------------------------------------- */
static void parseditem_display ( ParsedItem * item )
{
    ParsedItem  *c;
    Grammar     *gram;
    char   buf[ 1024 ];
    char  *p;
    int    i;
    
    if( !item ) {
        fprintf( stderr, "NULL(ParsedItem)\n" );
        return;
    }
    
    Data_Get_Struct( item->rule->gramobj, Grammar, gram );
    
    fprintf( stderr, "(%d,%d) ", item->left, item->right );
    rule_typestring_internal( item->rule, buf, -1 );
    fprintf( stderr, "%s ---", buf );
    for( i = 0; i < item->rule->rhslen; ++i ) {
        c = item->child[i];
        if( !c )
            continue;
        fprintf( stderr, " %d(%d,%d)", STR2CSTR(cfg_get_category(gram, item->top)), c->left, c->right );
    }
    fprintf( stderr, "\n" );
}


/* CYK ^Cv̎xOto͂
------------------------------------------------------------------- */
struct tagCYKData
{
    int  id; /* xOtp̃io[ */
};
typedef  struct tagCYKData  CYKData;

struct tagCYKDispData
{
    ItemSet  *itemset;
    CYKData  *databuf;
    ParsedItem_p  *itemlist;
    int  sgsize;
    
    /* ȉ͍Ɨp̏O[of[^ */
    int  id;
    VALUE  str_head;
};
typedef  struct tagCYKDispData  CYKDispData;

static void cyk_dispdat_free ( CYKDispData * d )
{
    if( d ) {
        ruby_xfree( d->databuf );
        ruby_xfree( d->itemlist );
        ruby_xfree( d );
    }
}

static VALUE getstr_rule ( Rule * rule )
{
    return rule_typestring(0, NULL, rule->ruleobj);
}

static VALUE getstr_dotted_rule ( Rule * rule, int dot )
{
    VALUE v = INT2FIX( dot );
    
    return rule_typestring(1, &v, rule->ruleobj);
}

static void visit_cyk ( CYKDispData * dispdat, ParsedItem * item )
{
    CYKData  *data;
    int  i;
    
    if( !item || !item->child )
        return;
    
    data = (CYKData *)item->data;
    if( !data )
        rb_raise( rb_eNLPException, "ItemSet is broken. (id=%d)", dispdat->id );
    if( data->id >= 0 )
        return; /* łɏ񂵂Ă */
    
    for( ; item; item = item->next_item ) {
        for( i = 0; i < item->rule->rhslen; ++i ) {
            visit_cyk( dispdat, item->child[i] );
        }
    }
    data->id = dispdat->id; /* data  headbuf ɓo^ꂽ̃f[^̈ */
    ++dispdat->id;
}

static int output_basenode_condition ( int count, char* buf, char* format, ... )
{
    va_list ap;
    int rcv;
    int ret;
    
    va_start( ap, format );
    if( count == 0 ) {
        *(buf++) = '|';
        ret = 1;
    }
    else {
        *(buf++) = ',';
        ret = 1;
    }
    
    rcv = vsprintf( buf, format, ap );
    if( rcv == EOF )
        rb_raise( rb_eIOError, "itemset.c: output_basenode_condition(): failed vsprintf()." );
    ret += rcv;
    
    va_end( ap );
    
    return ret;
}

static void itemset_display_cyk_output ( CYKDispData * dispdat, VALUE out )
{
    ParsedItem_p  ptr, child;
    Grammar  *gram;
    CYKData  *data;
    VALUE volatile v;
    VALUE volatile cat = Qnil;
    char  buf[2048];
    char *p;
    int   i, j, len, count;
    
    gram = dispdat->itemset->grammar;
    for( i = 0; i < dispdat->id; ++i ) {
        ptr = dispdat->itemlist[i];
        
        v = rb_String( cfg_get_category(gram, ptr->top) );
        len = sprintf( buf, "[ %d ]\t\t# %s(%d,%d)\n", i, RSTRING(v)->ptr, ptr->left, ptr->right );
        rb_io_write( out, rb_str_new(buf, len) );
        
        for( ; ptr; ptr = ptr->next_item ) {
            data = (CYKData*)ptr->data;
            
            p = buf;
            
            /* {m[h\z (̓At@xbgɏo͂悤C邱) */
            count = 0;
            v = getstr_rule( ptr->rule );
            p += sprintf( p, " +- (:%s", RSTRING(v)->ptr );
            if( ptr->attr.lex != 0 ) {
                cat = rb_String( cfg_get_category(gram, ptr->attr.lex) );
                p += output_basenode_condition( count++, p, "lex=%s", RSTRING(cat)->ptr );
            }
            p += sprintf( p, ":)" );
            
            for( j = 0; j < ptr->rule->rhslen; ++j ) {
                child = ptr->child[j];
                if( !child )
                    continue;
                p += sprintf( p, " @%d", ((CYKData *)child->data)->id );
            }
            *(p++) = '\n';
            *(p++) = '\0';
            if( p - buf > sizeof(buf)/sizeof(buf[0]) )
                rb_fatal( "FATAL: itemset_display_cyk_output() needs longer string buffer. Do rebuild." );
            rb_io_write( out, rb_str_new2( buf ) );
        }
    }
}

static int count_sgsize ( ItemSet * itemset )
{
    /* headbuf ̂ child ̂ȏ゠Gg̐xOtɂȂ */
    ParsedItem_p*  ptr = itemset->headbuf;
    ParsedItem_p  item;
    int  sgsize = 0;
    int  i;
    
    for( i = 0; i < itemset->headbuf_size; ++i ) {
        for( item = *ptr; item; item = item->next_item ) {
            if( item->child )
                break;
        }
        if( item )
            ++sgsize;
        ++ptr;
    }
    
    return sgsize; /* itemset->headbuf_size ([̐錾Ȃꍇ) */
}

static VALUE itemset_display_cyk_setup ( ItemSet* itemset )
{
    ParsedItem_p *headbuf = itemset->headbuf;
    ParsedItem *itembuf = itemset->itembuf;
    int sgsize = count_sgsize( itemset );
    
    VALUE volatile dispdat_obj;
    CYKDispData  *dispdat;
    CYKData  *data;
    int  i, addr;
    
    dispdat = ALLOC(CYKDispData);
    MEMZERO(dispdat, CYKDispData, 1); /* dispdat ̓[NAĂ */
    
    dispdat_obj = Data_Wrap_Struct( rb_cData, NULL, cyk_dispdat_free, dispdat );
    dispdat->itemset = itemset;
    dispdat->sgsize = sgsize;
    
    /* ܂͏ */
    for( i = 0; i < itemset->itembuf_size; ++i )
        itemset->itembuf[i].data = NULL;
    dispdat->databuf = ALLOC_N( CYKData, itemset->itembuf_size );
    MEMZERO(dispdat->databuf, CYKData, itemset->itembuf_size); /* databuf ̓[NAĂ */
    for( i = 0; i < itemset->itembuf_size; ++i ) {
        itembuf[i].data = &dispdat->databuf[i];
        dispdat->databuf[i].id = -1;
    }
    dispdat->id = 0;
    visit_cyk( dispdat, itemset->start );
    if( dispdat->id != sgsize )
        rb_raise( rb_eNLPException, "ItemSet is broken. (id=%d, sgsize=%d)", dispdat->id, sgsize );
    
    /* Otёւ */
    dispdat->itemlist = ALLOC_N( ParsedItem_p, sgsize );
    for( i = 0; i < sgsize; ++i ) {
        data = (CYKData *)headbuf[i]->data;
        addr = (sgsize-1) - data->id;
        if( addr < 0 || addr >= sgsize )
            rb_raise( rb_eNLPException, "Found unvisited item." );
        dispdat->itemlist[addr] = headbuf[i];
        data->id = addr;
    }
    
    return dispdat_obj;
}

static VALUE itemset_display_cyk ( VALUE self, VALUE out )
{
    VALUE volatile dispdat_obj;
    CYKDispData *dispdat;
    ItemSet *itemset;
    
    /* xOt̃f[^\z */
    Data_Get_Struct( self, ItemSet, itemset );
    dispdat_obj = itemset_display_cyk_setup( itemset );
    Data_Get_Struct( dispdat_obj, CYKDispData, dispdat );
    
    /* o */
    itemset_display_cyk_output( dispdat, out );
    
    return self;
}


/* Earley ^Cv̎xOto͂
----------------------------------------------------------------------
 - ParsedItem H
    * EarleyItem: ParsedItem Ɋ֘AÂꂽ left,right,top Ɠ̂ЂƂB
 - H
    * EarleyItem: ParsedItem  left,right,top Ɠ̂̊B
    * EarleyItem: hbgЂƂ߂B
    * EarleyItem: hbgʒu (LJEg) ̍B

 - next_right 邽߂:
    * (l,r)[<X> --> <A> * <B>] Ƃ (l,*)[<X> --> <A> * <B>] ׂĂHȂ΂ȂȂB
    * (l,*)[<X> --> <A> * <B>] ŃOAgɖ߂܂Ń[v΂悢B
    * (l,r)... Cӂ (l,r2)... TB
      (l,r2)  next_right ɑ}B(l,r2) ΎgɃNB

 - (l,r)[<X> --> <A> a * <B>] ɑΉpX (l,r-1)[<X> --> <A> * a <B>] ЂƂ݂ȂB
   (l,l)[<X> --> * ...] ɑΉpXK݂̂łЂƂ݂ȂB
   (l,r)[<X> --> <A> <B> * <C>] ɑΉpX͕݂B
    (l,i)[<X> --> <A> * <B> <C>] <-- right ݂̂ς
    (i,r)[<B> --> ... *] <-- Eӂ݂̂ς

 - EarleyItem ӂɒ߂vf  left, right, rule, dot (ׂĂvΓ)
*/

/* ---- Earley  ---- */
struct tagEarleyItem
{
    ParsedItem const * item; /* 쐬Ɏgpf[^ */
    
    int  id;   /* ΉxOt̃CfNX */
    int  flag; /* vpeBtO */
    
    Rule_p  rule;  /* ΉK */
    int  dot;      /* hbgʒu */
    int  left;     /* [ʒu */
    int  right;    /* E[ʒu */
    
    int  waiting; /* hbǧɂL */
    
    struct tagEarleyItem  *next_hash; /* nbVp̃f[^ */
    
    struct tagEarleyItem  *next_right; /* right ݂̂ςO */
    struct tagEarleyItem  *prev;       /* hbgЂƂ߂ right Oւ̃|C^ */
};
typedef  struct tagEarleyItem    EarleyItem;
typedef  struct tagEarleyItem *  EarleyItemptr;
typedef struct tagEarleyItem* EarleyItem_p;

#if USE_MACRO
#define is_visited(ITEM) ((ITEM)->flag & 0x01)
#define set_visited(ITEM) ((ITEM)->flag |= 0x01)
#define is_comp(ITEM) ((ITEM)->flag & 0x02)
#define set_comp(ITEM) ((ITEM)->flag |= 0x02)
#define is_scan(ITEM) ((ITEM)->flag & 0x04)
#define set_scan(ITEM) ((ITEM)->flag |= 0x04)
#else
static int is_visited ( EarleyItem const * eitem )
{
    return eitem->flag & 0x01;
}
static void set_visited ( EarleyItem const * eitem )
{
    eitem->flag |= 0x01;
}
static int is_comp ( EarleyItem const * eitem )
{
    return eitem->flag & 0x02;
}
static void set_comp ( EarleyItem const * eitem )
{
    eitem->flag |= 0x02;
}
static int is_scan ( EarleyItem const * eitem ) /* hbgO̍I[LH */
{
    return eitem->flag & 0x04;
}
static void set_scan ( EarleyItem const * eitem )
{
    eitem->flag |= 0x04;
}
#endif

static EarleyItem_p init_earley_item ( EarleyItem_p eitem, ParsedItem const * item, int dot )
{
    eitem->item = item;
    
    eitem->id   = -1;
    eitem->flag = 0;
    
    if( dot >= item->rule->rhslen ) {
        eitem->waiting = -1;
        set_comp( eitem );
    }
    else {
        if( item->child[dot] )
            eitem->waiting = item->child[dot]->top;
        else
            eitem->waiting = -1;
    }
    
    eitem->dot  = dot;
    eitem->rule = item->rule;
    eitem->left = item->left;
    if( dot > 0 && !item->child[dot-1] )
        set_scan( eitem );
    
    eitem->next_hash = NULL;
    eitem->next_right = NULL;
    eitem->prev = NULL;
    
    return eitem;
}

static VALUE getstr_earley_rule ( EarleyItem * eitem )
{
    return getstr_dotted_rule( eitem->item->rule, eitem->dot );
}


/* ---- nbVe[u ---- */
struct tagEarleyHash
{
    int  sent_len; /* +1 */
    int  cap;
    EarleyItemptr  *ptr;
};
typedef  struct tagEarleyHash  EarleyHash;

static EarleyHash * init_earley_hash ( EarleyHash * hash, int sentlen )
{
    hash->sent_len = sentlen + 1;
    hash->cap = hash->sent_len * hash->sent_len;
    hash->ptr = ALLOC_N( EarleyItem_p, hash->cap );
    
    return hash;
}

#if USE_MACRO
#define  earley_hash_at(HASH,LEFT,RIGHT)  ((HASH)->ptr[(LEFT) + (RIGHT)*((HASH)->sent_len)])
#else
static EarleyItemptr earley_hash_at ( EarleyHash * hash, int left, int right )
{
    return hash->ptr[ left + right * hash->sent_len ];
}
#endif

static EarleyHash * setitem_earley_hash ( EarleyHash * hash, EarleyItem * eitem )
{
    hash->ptr[ eitem->left + eitem->right * hash->sent_len ] = eitem;
    return hash;
}

static EarleyHash * clear_earley_hash ( EarleyHash * hash )
{
    MEMZERO( hash->ptr, EarleyItem_p, hash->cap );
    return hash;
}

static void delete_earley_hash ( EarleyHash * hash )
{
    if( hash ) {
        ruby_xfree( hash->ptr );
    }
}

static EarleyItem_p add_earley_hash ( EarleyHash* hash, EarleyItem_p eitem )
{
    EarleyItem_p  ptr;
    Rule_p  rule = eitem->rule;
    int  dot = eitem->dot;
    int  left = eitem->left;
    int  right = eitem->right;
    int  i;
    
    #if 0
    /* Ô߂̃`FbN (Ȃf[^邱Ƃ̂) */
    if( left < 0 || left > hash->sent_len || right < 0 || right > hash->sent_len )
        rb_raise( rb_eNLPException, "add_earley_hash(): Broken item." );
    #endif
    
    /* ܂͊Svf[^ */
    for( ptr = earley_hash_at( hash, left, right ); ptr; ptr = ptr->next_hash ) {
        if( ptr->rule == rule && ptr->dot == dot )
            return ptr;
    }
    
    /* right ݂̂ςOɒǉ */
    for( i = left; !eitem->next_right && i < hash->sent_len; ++i ) {
        for( ptr = earley_hash_at( hash, left, i ); ptr; ptr = ptr->next_hash ) {
            if( ptr->rule == rule && ptr->dot == dot ) {
                eitem->next_right = ptr->next_right;
                ptr->next_right = eitem;
                break;
            }
        }
    }
    if( !eitem->next_right )
        eitem->next_right = eitem;
    
    /* hbgЂƂ߂O */
    if( dot > 0 ) {
        --dot;
        if( dot == 0 ) {
            for( ptr = earley_hash_at( hash, left, left ); ptr; ptr = ptr->next_hash ) {
                if( ptr->rule == rule && ptr->dot == 0 ) {
                    eitem->prev = ptr;
                    break;
                }
            }
        }
        else {
            for( i = left; i < right; ++i ) {
                for( ptr = earley_hash_at( hash, left, i ); ptr; ptr = ptr->next_hash ) {
                    if( ptr->rule == rule && ptr->dot == dot ) {
                        eitem->prev = ptr;
                        break;
                    }
                }
            }
        }
        if( !eitem->prev )
            rb_raise( rb_eNLPException, "add_earley_hash(): (%d,*)[%s] is not found.", left, STR2CSTR(getstr_dotted_rule( rule, dot )) );
        ++dot;
    }
    
    /* nbVɒǉ */
    eitem->next_hash = earley_hash_at( hash, eitem->left, eitem->right );
    setitem_earley_hash( hash, eitem );
    
    return eitem;
}


/* ---- display \bhŎgp鋤ʃf[^ ---- */
struct tagEarleyDispData
{
    ItemSet       *itemset;
    EarleyItem    *itembuf;
    EarleyItem_p  *itemlist;
    EarleyHash     hash;
    int            sgnum;
    int            itembuf_len;
    char           strbuf[ 1024 ];
};
typedef  struct tagEarleyDispData  EarleyDispData;

static void earley_dispdat_free ( EarleyDispData * data )
{
    if( data ) {
        delete_earley_hash( &data->hash );
        ruby_xfree( data->itembuf );
        ruby_xfree( data->itemlist );
        ruby_xfree( data );
    }
}


/* ---- xOt̏o ---- */
static void construct_earley_hash ( EarleyDispData* dispdat )
{
    ItemSet* itemset = dispdat->itemset;
    EarleyHash* hash = &(dispdat->hash);
    
    ParsedItem const *item;
    EarleyItem_p eitem;
    int i, j, right;
    
    item  = dispdat->itemset->itembuf;
    eitem = dispdat->itembuf;
    
    clear_earley_hash( hash );
    for( i = 0; i < itemset->itembuf_size; ++i ) {
        right = item->left;
        for( j = 0; j < item->rule->rhslen; ++j ) {
            init_earley_item( eitem, item, j );
            eitem->right = right;
            add_earley_hash( hash, eitem ); /* Ă̂̂ꍇ̓LZ */
            if( item->child[j] ) {
                /* item->child Ă */
                right = item->child[j]->right;
            } else {
                ++right;
            }
            ++eitem;
        }
        
        init_earley_item( eitem, item, item->rule->rhslen );
        eitem->right = right;
        add_earley_hash( hash, eitem );
        ++eitem;
        
        ++item;
    }
}

static void earley_visit_item ( EarleyDispData * dispdat, EarleyItem * eitem )
{
    EarleyItem  *prev, *child;
    int  flag;
    
    if( is_visited(eitem) )
        return;
    set_visited( eitem );
    
    if( eitem->dot > 0 ) { /* hbgЂƂO̍To */
        if( !is_scan( eitem ) ) { /* eitem: ... <A> * ... ̏ꍇ */
            for( prev = eitem->prev; prev; prev = (prev->next_right == eitem->prev)? NULL: prev->next_right ) {
                if( prev->right >= eitem->right )
                    continue;
                flag = 0;
                for( child = earley_hash_at( &dispdat->hash, prev->right, eitem->right ); child; child = child->next_hash ) {
                    if( !is_comp( child ) || child->item->top != prev->waiting )
                        continue;
                    earley_visit_item( dispdat, child );
                    ++flag;
                }
                if( flag > 0 )
                    earley_visit_item( dispdat, prev );
            }
        }
        else { /* eitem: ... a * ... ̏ꍇ */
            flag = 0;
            for( prev = eitem->prev; prev; prev = (prev->next_right == eitem->prev)? NULL: prev->next_right ) {
                if( prev->right + 1 != eitem->right )
                    continue;
                if( flag > 0 )
                    rb_raise( rb_eNLPException, "Found too many (%d,%d)[%s].", prev->left, prev->right, STR2CSTR(getstr_earley_rule(prev)) );
                earley_visit_item( dispdat, prev );
            }
        }
    }
    
    eitem->id = dispdat->sgnum;
    ++(dispdat->sgnum);
}

static void earley_visit ( EarleyDispData * dispdat )
{
    ParsedItem const *item;
    EarleyItem *eitem;
    
    /* Jnɑ鍀ڂKJn */
    dispdat->sgnum = 0;
    for( item = dispdat->itemset->start; item; item = item->next_item ) {
        for( eitem = earley_hash_at( &dispdat->hash, item->left, item->right ); eitem; eitem = eitem->next_hash ) {
            if( is_comp(eitem) && eitem->rule == item->rule )
                earley_visit_item( dispdat, eitem );
        }
    }
}

static void construct_earley_list ( EarleyDispData * dispdat )
{
    EarleyHash * hash = &dispdat->hash;
    EarleyItem * eitem;
    int  i, addr;
    
    dispdat->itemlist = rb_alloc_zero( EarleyItemptr, dispdat->sgnum );
    for( i = 0; i < hash->cap; ++i ) {
        for( eitem = hash->ptr[i]; eitem; eitem = eitem->next_hash ) {
            if( eitem->id < 0 )
                rb_raise( rb_eNLPException, "(%d,%d)[%s] is not used.", eitem->left, eitem->right, STR2CSTR(getstr_earley_rule(eitem)) );
            addr = (dispdat->sgnum - 1) - eitem->id;
            eitem->id = addr;
            if( dispdat->itemlist[addr] )
                rb_raise( rb_eNLPException, "Found too many ID #%d.", addr );
            dispdat->itemlist[addr] = eitem;
        }
    }
}

static void earley_output_sgraph ( EarleyDispData * dispdat, VALUE out )
{
    EarleyItem  *eitem, *prev, *child;
    int  id, flag, right;
    
    for( id = 0; id < dispdat->sgnum; ++id ) {
        eitem = dispdat->itemlist[id];
        sprintf( dispdat->strbuf, "[ %d ]\t\t# (%d,%d) %s\n", id, eitem->left, eitem->right, STR2CSTR(getstr_earley_rule(eitem)) );
        rb_io_write( out, rb_str_new2( dispdat->strbuf ) );
        
        /* visit Ɠ */
        if( eitem->dot > 0 ) {
            if( !is_scan( eitem ) ) {
                for( prev = eitem->prev; prev; prev = (prev->next_right == eitem->prev)? NULL: prev->next_right ) {
                    if( prev->right >= eitem->right )
                        continue;
                    for( child = earley_hash_at( &dispdat->hash, prev->right, eitem->right ); child; child = child->next_hash ) {
                        if( !is_comp( child ) || child->item->top != prev->waiting )
                            continue;
                        sprintf( dispdat->strbuf, " +- @%d @%d\n", prev->id, child->id );
                        rb_io_write( out, rb_str_new2( dispdat->strbuf ) );
                    }
                }
            }
            else {
                for( prev = eitem->prev; prev; prev = (prev->next_right == eitem->prev)? NULL: prev->next_right ) {
                    if( prev->right + 1 != eitem->right )
                        continue;
                    sprintf( dispdat->strbuf, " +- @%d\n", prev->id );
                    rb_io_write( out, rb_str_new2( dispdat->strbuf ) );
                }
            }
        }
        else {
            sprintf( dispdat->strbuf, " +- (:%s:)\n", STR2CSTR(getstr_rule( eitem->item->rule )) );
            rb_io_write( out, rb_str_new2( dispdat->strbuf ) );
        }
    }
}

static int count_earley_item ( ItemSet* itemset )
{
    ParsedItem* itembuf = itemset->itembuf;
    int count = 0;
    int i;
    
    for( i = 0; i < itemset->itembuf_size; ++i ) {
        count += itembuf->rule->rhslen + 1; /* +1: hbgʒũp^[ɂ */
        ++itembuf;
    }
    
    return count;
}

static VALUE itemset_display_earley ( VALUE self, VALUE out )
{
    /*  Ruby \bh̓ */
    VALUE volatile dispdat_obj;
    EarleyDispData* dispdat;
    ItemSet* itemset;
    
    static int count = 0;
    
    /* ʃf[^̐ */
    Data_Get_Struct( self, ItemSet, itemset );
    dispdat = rb_alloc_zero( EarleyDispData, 1 );
    dispdat_obj = Data_Wrap_Struct( rb_cData, NULL, earley_dispdat_free, dispdat );
    dispdat->itemset = itemset;
    
    dispdat->itembuf_len = count_earley_item( itemset );
    dispdat->itembuf = ALLOC_N( EarleyItem, dispdat->itembuf_len );
    
    /* Earley ̃XgƃnbV\z */
    init_earley_hash( &dispdat->hash, itemset->sent_len - 1 );
    construct_earley_hash( dispdat );
    
    /*  */
    earley_visit( dispdat );
    
    /* o */
    construct_earley_list( dispdat );
    earley_output_sgraph( dispdat, out );
    
    return self;
}


/* ================================================================ */


/*   Viterbi p[U
------------------------------------------------------------------- */

/* ɑΉf[^ (ׂẴf[^Ɋ蓖Ă) */
struct tagViterbiData
{
    real_t  inside;       /* ؂̓m (Cɂݒ肳Ȃ) */
    real_t  inside_path;  /* ̃pXɑ΂m */
    int     visited;
    int     index;        /* ItemSet::headbuf ɂCfNX */
};
typedef  struct tagViterbiData  ViterbiData;

#define VDAT(ITEMPTR) ((ViterbiData *)((ITEMPTR)->data))

/* ԏ */
struct tagViterbiState;
typedef  struct tagViterbiState  ViterbiState;
typedef  ViterbiState *  ViterbiStateptr;
typedef  ViterbiState *  ViterbiState_p;
struct tagViterbiState
{
    real_t  sum;  /* ݌vޓx */
    
    ParsedItem_p * path;   /* I񂾃pX̃Xg ( itemset->headbuf_size) */
    ParsedItem_p * stack;  /* WJ҂̃X^bN */
    
    int  stack_len;
    
    /* Jڏ */
    ViterbiState_p  prev_state;  /* JڑȌ */
    ParsedItem_p  expander;  /* JڂɎgpp^[ */
};

/* ЂƂ̉͂t[ */
struct tagViterbiFrame
{
    ViterbiState_p * state;
};
typedef  struct tagViterbiFrame  ViterbiFrame;

/* Viterbi p[UCf[^ */
struct tagViterbiDispData
{
    ViterbiData*  databuf;
    ItemSet*  itemset;
    Grammar*  grammar;
    PCFG*  pcfg;
    VALUE  out;
    
    int  nbest;  /* ̉\o͂邩H */
    
    ViterbiFrame  framebuf[2];
    ViterbiFrame *currframe;
    ViterbiFrame *nextframe;
    
    ViterbiState_p * stateptrbuf;
    ViterbiState   * statebuf;
    
    /* \p̃obt@ */
    VALUE str_nl, str_open, str_close, str_spc;
};
typedef  struct tagViterbiDispData  ViterbiDispData;


static void errp_viterbi_state ( ViterbiDispData * data, ViterbiState * state )
{
    int  i, n;
    int  headsize = data->itemset->headbuf_size;
    
    for( i = 0; i < data->nbest; ++i ) {
        errp( "[%d]\n", i );
        errp( "sum: %g\n", state->sum );
        errp( "stack:" );
        for( n = 0; n < state->stack_len; ++n )
            errp( " %d", VDAT(state->stack[n])->index );
        errp( "\n" );
        errp( "path :" );
        for( n = 0; n < headsize; ++n ) {
            if( state->path[n] )
                errp( " %d", VDAT(state->path[n])->index );
        }
        errp( "\n----\n" );
    }
}

static void free_viterbi_dispdata ( ViterbiDispData * dat )
{
    ViterbiState * state;
    int  n;
    
    if( dat ) {
        for( n = dat->nbest*2 - 1; n >= 0; --n ) {
            state = &dat->statebuf[n];
            ruby_xfree( state->stack );
            ruby_xfree( state->path );
        }
        ruby_xfree( dat->databuf );
        ruby_xfree( dat->stateptrbuf );
        ruby_xfree( dat->statebuf );
        ruby_xfree( dat );
    }
}

static void mark_viterbi_dispdata ( ViterbiDispData * dat )
{
    if( dat ) {
        rb_gc_mark( dat->out );
    }
}

static void viterbitree_alloc_data ( ViterbiDispData * data )
{
    ItemSet     * itemset = data->itemset;
    ViterbiData * dataptr;
    ParsedItem  * itemptr;
    int  n;
    
    /* 蓖Ăāc ([NAĂ) */
    data->databuf = rb_alloc_zero( ViterbiData, itemset->itembuf_size );
    
    /* e犄蓖ĂɃN𒣂 */
    itemptr = itemset->itembuf;
    dataptr = data->databuf;
    for( n = 0; n < itemset->itembuf_size; ++n ) {
        dataptr->index = -1;
        (itemptr++)->data = dataptr++;
    }
    
    /* Ō͊e headbuf ̃CfNXUĂ */
    for( n = 0; n < itemset->headbuf_size; ++n ) {
        for( itemptr = itemset->headbuf[n]; itemptr; itemptr = itemptr->next_item ) {
            dataptr = (ViterbiData*)(itemptr->data);
            dataptr->index = n;
        }
    }
}


/*   Viterbi p[UEmvZ
------------------------------------------------------------------- */

static real_t viterbitree_inside_itr( ViterbiDispData * data, ParsedItemptr item )
{
    /* item (t(l,r)) ̓m߂ */
    ViterbiData*  vitem = (ViterbiData*)(item->data);
    ViterbiData*  dat;
    ParsedItem_p  ptr;
    PCFGAttr  attr;
    real_t  r, sum;
    int  n;
    
    #if DEBUGSW
    if( vitem->visited < 0 ) {
        /* ̃G[N̂ ParsedItem.child C (next_item Xg̐擪) ĂȂ */
        rb_raise( rb_eNLPException, "[BUG/nlputil.so] Visited non-head item." );
    }
    #endif
    if( vitem->visited )
        return vitem->inside;
    
    sum = 0;
    for( ptr = item; ptr; ptr = ptr->next_item ) {
        dat = (ViterbiData*)(ptr->data);
        #if DEBUGSW
        dat->visited = -1; /* G[`FbNp̕z */
        #endif
        
        /* pX̊m߂āc */
        attr.lex = ptr->attr.lex;
        r = pcfg_get_rule_probability( data->pcfg, ptr->rule->id );
        for( n = ptr->rule->rhslen - 1; n >= 0; --n ) {
            if( ptr->child[n] )
                r *= viterbitree_inside_itr( data, ptr->child[n] );
        }
        dat->inside_path = r;
        
        /* vlɉ */
        sum += r;
    }
    vitem->inside = sum; /* inside ̓C (headbuf ɑ݂鍀) ɂݒ肵Ȃ */
    vitem->visited = 1;
    
    return sum;
}

static void viterbitree_inside ( ViterbiDispData * data )
{
    /* Cɑ΂mvZ */
    viterbitree_inside_itr( data, data->itemset->start );
}



/*   Viterbi p[UE\
----------------------------------------------------------------------
Viterbi ̃ASY

- ݂̃t[ׂ̂ẴX^bNɂȂ܂Łc
  - (X^bNłȂ) ꂼ̌ɑ΂āc
    - X^bNgbv̍  (Ot) o
    -  ̊epX q ɑ΂Jڂl
    - ꂽp^[tꂽtXgɊi[B
      ([JڑȌ, JڃpXԍ, p[^] ̃Xg)
  - Jڌ̃Xg𐶐B(obt@̃Rs[Ȃ)
  - JڑOƑJڌ̃t[ꊷB(|C^̃Xbv)

- X^bN̍\
  - KWJAΉ鍀uĂB
      : Start
    --> Start-RHS[2], Start-RHS[1], Start-RHS[0]
    --> Start-RHS[2], Start-RHS[1], Start-RHS[0]-RHS[2], Start-RHS[0]-RHS[1]
    --> ...

- pX̍\
  - eOtŁAI񂾃pX̃AhXLqĂB

* t[
 ... uJڑOvƁuJڌvœpӂB
     t[ɂ͑JڏԂ̃XgĂB(N-bese  N )
*/

static void viterbitree_swap_frame ( ViterbiDispData * data )
{
    if( data->currframe == &data->framebuf[0] ) {
        data->nextframe = &data->framebuf[0];
        data->currframe = &data->framebuf[1];
    }
    else {
        data->nextframe = &data->framebuf[1];
        data->currframe = &data->framebuf[0];
    }
}

static ViterbiState * viterbi_state_init ( ViterbiDispData * data, ViterbiState * state )
{
    ItemSet * itemset = data->itemset;
    
    state->sum = 1.0;
    
    /* gĂȂԂƂď */
    state->stack_len = -1;
    state->expander  = NULL;
    
    /* X^bÑTCY͍őł headbuf ܂ */
    state->stack = ALLOC_N( ParsedItemptr, itemset->headbuf_size );
    
    /* gpXf[^͑I񂾃pX̋L^Ȃ̂ headbuf Ɠ (NULL Ŗ߂) */
    state->path = rb_alloc_zero( ParsedItemptr, itemset->headbuf_size );
    
    return state;
}

static void viterbitree_init_expansion ( ViterbiDispData * data )
{
    ViterbiFrame * currframe = data->currframe;
    ViterbiFrame * nextframe = data->nextframe;
    ItemSet *      itemset   = data->itemset;
    int            nbest     = data->nbest;
    int            headsize  = data->itemset->headbuf_size;
    
    ViterbiState  *curr, *next;
    int  n, len;
    
    /* ͏IĂ̂l߂Ă */
    len = 0;
    for( n = 0; n < nbest; ++n ) {
        curr = currframe->state[ n ];
        if( curr->stack_len == 0 ) {
            /* WJ~܂ curr[n]  next[len] ɃRs[ */
            next = nextframe->state[ len++ ];
            memcpy( next->path, curr->path, sizeof(ParsedItemptr) * headsize );
            next->stack_len = 0;
            next->sum = curr->sum;
            next->expander = NULL;
        }
    }
    
    /* expander  NULL Ȃ΋̏Ԃƍl */
    for( ; len < nbest; ++len ) {
        next = nextframe->state[ len ];
        next->expander  = NULL;
        next->stack_len = -1;
    }
}

static int viterbitree_add_pattern ( ViterbiDispData * data, ViterbiStateptr from )
{
    ViterbiFrame * frame = data->nextframe;
    int            nbest = data->nbest;
    
    ParsedItemptr  item;
    ViterbiState * state;
    real_t  param;
    int     n;
    
    if( from->stack_len <= 0 )
        return 0;
    
    /* X^bNgbvo */
    item = from->stack[ --(from->stack_len) ];
    
    /* from ɑ΂đJډ\ȂׂẴpX */
    for( ; item; item = item->next_item ) {
        /* \[g̊l ݐϖޓx~pX̓m */
        param = from->sum * VDAT(item)->inside_path;
        
        /* ǂɑ}邩`FbN */
        for( n = 0; n < nbest; ++n ) {
            state = frame->state[ n ];
            
            /* ̏Ԃ΂̏ꏊg */
            if( !state->expander && state->stack_len < 0 )
                break;
            
            /* ޓxԂ΂̒Oɑ} */
            if( state->sum < param ) {
                state = frame->state[ nbest-1 ]; /* ǂoobt@g */
                memmove( &frame->state[n+1], &frame->state[n], sizeof(ViterbiStateptr) * (nbest - n - 1) );
                frame->state[n] = state;
                break;
            }
            
            /* l̏ꍇ͍ŏɊi[ꂽ̗̂p (łA͂ǂɂ) */
            if( state->sum == param ) {
                ;
            }
        }
        if( n >= nbest )
            continue; /* ǂɂꏊȂ */
        
        /* Jڏݒ */
        state = frame->state[ n ]; /* ...Ô */
        state->sum        = from->sum;
        state->expander   = item;
        state->prev_state = from;
    }
    
    return 1;
}

static void viterbitree_make_state ( ViterbiDispData * data )
{
    ItemSet      * itemset = data->itemset;
    ViterbiFrame * frame   = data->nextframe; /* Ƃ nextframe ōs */
    int            nbest   = data->nbest;
    
    ParsedItem_p  item;
    ViterbiState  *state, *prev;
    PCFGAttr attr;
    Rule * rule;
    int    i, n, headsize;
    
    headsize = itemset->headbuf_size;
    for( n = 0; n < nbest; ++n ) {
        state = frame->state[n];
        item  = state->expander;
        if( !item )
            continue; /* ̏ԂȂ΃XLbv */
        
        prev = state->prev_state;
        
        /* pX̃Rs[ƒǉ */
        memcpy( state->path, prev->path, sizeof(ParsedItemptr) * headsize );
        state->path[ VDAT(item)->index ] = item;
        
        /* X^bNւ̒ǉ
         *  - prev ̃X^bNRs[Ă expander ̎qĂ
         */
        rule = item->rule;
        memcpy( state->stack, prev->stack, sizeof(ParsedItemptr) * prev->stack_len );
        state->stack_len = prev->stack_len;
        for( i = rule->rhslen - 1; i >= 0; --i ) {
            if( item->child[i] ) {
                state->stack[ (state->stack_len)++ ] = item->child[i];
            }
        }
        
        /* ݐϖޓx̒ǉ */
        /* data->pcfg  rule->id gĊm𓾂 */
        attr.lex = item->attr.lex;
        state->sum *= pcfg_get_rule_probability_ex( data->pcfg, rule->id, &attr );
    }
}

static void viterbitree_get_nbest ( ViterbiDispData * data )
{
    ItemSet      * itemset = data->itemset;
    ViterbiState * state;
    ViterbiFrame * frame;
    ParsedItem_p  p, head;
    int  i, n, count;
    int  nbest = data->nbest;
    
    /* ZbgAbv (Ԃ͊Jn) */
    data->stateptrbuf = ALLOC_N( ViterbiStateptr, nbest * 2 );
    data->statebuf    = ALLOC_N( ViterbiState   , nbest * 2 );
    for( n = 0; n < 2; ++n ) {
        frame = &data->framebuf[n];
        frame->state = &data->stateptrbuf[ nbest*n ];
        for( i = 0; i < nbest; ++i ) {
            frame->state[i] = &data->statebuf[ i + nbest*n ];
            viterbi_state_init( data, frame->state[i] );
        }
    }
    data->currframe = &data->framebuf[0];
    data->nextframe = &data->framebuf[1];
    
    state = data->currframe->state[0];
    state->stack[0] = itemset->start;
    state->stack_len = 1;
    
    /* JڂԂ͏ */
    do
    {
        /* Jڐi[obt@NA */
        viterbitree_init_expansion( data );
        
        frame = data->currframe;
        count = 0;
        for( n = 0; n < nbest; ++n ) {
            /* Jڃp^[lăXgɓo^ */
            count += viterbitree_add_pattern( data, frame->state[n] );
        }
        
        /* Jڌ̃Xg𐶐 */
        viterbitree_make_state( data );
        
        /* t[̃Xbv */
        viterbitree_swap_frame( data );
    }
    while( count > 0 );
}

static void viterbitree_display_itr ( ViterbiDispData * data, VALUE out, ParsedItemptr * pathmap, ParsedItemptr item )
{
    int  n;
    
    if( !item )
        rb_raise( rb_eNLPException, "[BUG/nlputil.so] ItemSet is broken. (index=%d)", index );
    
    rb_io_write( out, data->str_open );
    rb_io_write( out, rb_String(cfg_get_category( data->grammar, item->rule->lhs )) );
    for( n = 0; n < item->rule->rhslen; ++n ) {
        rb_io_write( out, data->str_spc );
        if( item->child[n] ) {
            viterbitree_display_itr( data, out, pathmap, pathmap[ VDAT(item->child[n])->index ] );
        }
        else {
            rb_io_write( out, rb_String(cfg_get_category( data->grammar, item->rule->rhs[n] )) );
        }
    }
    rb_io_write( out, data->str_close );
}

static void viterbitree_display ( ViterbiDispData * data )
{
    ParsedItemptr   startitem = data->itemset->start;
    ViterbiFrame  * frame     = data->currframe;
    VALUE  out = data->out;
    
    VALUE volatile str_nl    = rb_str_new( "\n", 1 );
    VALUE volatile str_open  = rb_str_new( "[", 1 );
    VALUE volatile str_close = rb_str_new( "]", 1 );
    VALUE volatile str_spc   = rb_str_new( " ", 1 );
    
    ParsedItemptr * pathmap;
    ViterbiState  * state;
    char  buf [256];
    int   startindex;
    int   n;
    
    data->str_nl    = str_nl;
    data->str_open  = str_open;
    data->str_close = str_close;
    data->str_spc   = str_spc;
    
    startindex = VDAT(startitem)->index;
    for( n = 0; n < data->nbest; ++n ) {
        state = frame->state[n];
        if( state->stack_len < 0 )
            continue;
        
        pathmap = state->path;
        sprintf( buf, "#%d (/%d) %g\n", n+1, data->nbest, state->sum );
        rb_io_write( out, rb_str_new2( buf ) );
        viterbitree_display_itr( data, out, pathmap, pathmap[ VDAT(startitem)->index ] );
        rb_io_write( out, data->str_nl );
    }
}

static void viterbitree_execute ( ViterbiDispData* dispdat )
{
    viterbitree_alloc_data( dispdat ); /* f[^\蓖Ăāc */
    
    viterbitree_inside( dispdat );  /* mZo */
    
    viterbitree_get_nbest( dispdat );  /* n-best ŃT[` */
    
    viterbitree_display( dispdat );  /* ʂ\ */
}

static VALUE itemset_display_setup ( int argc, VALUE* argv, VALUE self )
{
    ViterbiDispData* dispdata;
    ItemSet* itemset;
    VALUE volatile dispobj;
    VALUE nbestobj;
    VALUE pcfgobj;
    VALUE out;
    int nbest;
    
    /* ZbgAbv */
    rb_scan_args( argc, argv, "11", &out, &nbestobj );
    nbest = (NIL_P(nbestobj))? 1: NUM2INT(nbestobj); /* nbest ̃ftHg 1 */
    if( nbest < 1 )
        rb_raise( rb_eArgError, "Invalid argument. (nbest=%d)", nbest );
    Data_Get_Struct( self, ItemSet, itemset );
    
    /* IuWFNg̐ */
    dispdata = rb_alloc_zero(ViterbiDispData, 1);
    dispobj = Data_Wrap_Struct( rb_cData, NULL, free_viterbi_dispdata, dispdata );
    
    dispdata->out = out;
    dispdata->nbest = nbest;
    
    /* f[^\蓖Ă */
    dispdata->itemset = itemset;
    dispdata->grammar = itemset->grammar;
    pcfgobj = cfg_get_property( dispdata->grammar->ruby_object, rb_str_new2("pcfg") );
    if( CLASS_OF(pcfgobj) == rb_cPCFG ) {
        Data_Get_Struct( pcfgobj, PCFG, dispdata->pcfg );
    }
    else {
        rb_raise( rb_eNLPException, "PCFG not associated." ); /* mt̕@IuWFNg */
    }
    
    return dispobj;
}

static VALUE itemset_display_viterbi ( int argc, VALUE * argv, VALUE self )
{
    VALUE volatile dispobj;
    ViterbiDispData * dispdat;
    
    dispobj = itemset_display_setup( argc, argv, self );
    Data_Get_Struct( dispobj, ViterbiDispData, dispdat );
    
    viterbitree_execute( dispdat );
    
    return self;
}

static VALUE itemset_display_viterbi_lex ( int argc, VALUE * argv, VALUE self )
{
    VALUE volatile dispobj;
    ViterbiDispData * dispdat;
    
    dispobj = itemset_display_setup( argc, argv, self );
    Data_Get_Struct( dispobj, ViterbiDispData, dispdat );
    itemset_set_lex( dispdat->itemset );
    
    viterbitree_execute( dispdat );
    
    return self;
}


/* ================================================================ */


/*  CYK ^xOt̐ (P Lexicalized)
----------------------------------------------------------------------
 Ot head  (headbuf Ȃ)  ItemSet ƓɂȂB(CqtȂ)
 epXɏ񂪒ǉBA͏o͎ɒH΂̂ŏKv͂ȂB
*/
static VALUE itemset_display_cyk_lex ( VALUE self, VALUE out )
{
    VALUE volatile str_head = rb_str_new2("head");
    VALUE volatile dispdat_obj;
    CYKDispData* dispdat;
    ItemSet* itemset;
    
    /* xOt̃f[^\z */
    Data_Get_Struct( self, ItemSet, itemset );
    dispdat_obj = itemset_display_cyk_setup( itemset );
    Data_Get_Struct( dispdat_obj, CYKDispData, dispdat );
    dispdat->str_head = str_head;
    
    /* eɑ΂ lex f[^ݒ肷 */
    itemset_set_lex( itemset );
    
    /* o */
    itemset_display_cyk_output( dispdat, out );
    
    return self;
}


/* ================================================================ */


/* \؂̐𐔂
------------------------------------------------------------------- */

static VALUE itemset_count_tree_itr ( ParsedItemptr top )
{
    VALUE  *ptr, localsum;
    int i;
    
    if( !top )
        return INT2FIX(1);
    
    ptr = (VALUE *)(top->data);
    if( NIL_P(*ptr) ) {
        *ptr = INT2FIX(0);
        for( ; top; top = top->next_item ) {
            localsum = INT2FIX(1);
            for( i = top->rule->rhslen - 1; i >= 0; --i )
                localsum = rb_funcall( localsum, id_mult, 1, itemset_count_tree_itr( top->child[i] ) );
            *ptr = rb_funcall( *ptr, id_plus, 1, localsum );
        }
    }
    
    return *ptr;
}

static VALUE itemset_count_tree ( VALUE self )
{
    ItemSet * itemset;
    VALUE  ary = Qnil;
    int    i;
    
    /* f[^̃ZbgAbv */
    Data_Get_Struct( self, ItemSet, itemset );
    ary = rb_ary_new4( itemset->itembuf_size, NULL );
    for( i = 0; i < itemset->itembuf_size; ++i ) {
        RARRAY(ary)->ptr[i] = Qnil;
        itemset->itembuf[i].data = &RARRAY(ary)->ptr[i];
    }
    
    return itemset_count_tree_itr( itemset->start );
}


/* 𓾂
----------------------------------------------------------------------
*/
static VALUE itemset_length ( VALUE self )
{
    ItemSet * itemset;
    
    Data_Get_Struct( self, ItemSet, itemset );
    
    return INT2FIX( itemset->sent_len - 1 );
}


VALUE itemset_get_start_category ( ItemSet const * itemset )
{
    Grammar const * grammar;
    Rule const * rule;
    
    if( !itemset->start )
        return Qnil;
    
    rule = itemset->start->rule;
    if( !rule )
        rb_raise( rb_eRuntimeError, "found ruleless top-item in itemset" );
    
    grammar = itemset->grammar;
    
    return cfg_get_category( grammar, rule->lhs );
}


VALUE itemset_reconstruct_chart ( ItemSet const * itemset )
{
    VALUE volatile chartobj;
    ParsedItem const * item;
    Chart * chart;
    Rule * rule;
    State  state;
    int  i, j, right;
    
    chartobj = chart_new( itemset_get_start_category(itemset), itemset->grammar, 0, itemset->sent_len - 1 );
    Data_Get_Struct( chartobj, Chart, chart );
    
    for( i = itemset->itembuf_size - 1; i >= 0; --i ) {
        item = &itemset->itembuf[i];
        rule = item->rule;
        if( rule ) {
            /* K蓖ĂĂ̂Ń`[gɏԂo^ */
            right = item->left;
            
            /* Ԃ̐Ɠo^ */
            for( j = 0; j <= rule->rhslen; ++j ) {
                state_new( &state, rule, item->left );
                state.dot = j;
                state.right = right;
                
                /* dĂȂΓo^ */
                chart_add_state2( chart, &state );
                
                if( j < rule->rhslen ) {
                    if( item->child[j] )
                        right = item->child[j]->right;
                    else
                        right += 1;
                }
            }
        }
        else {
            ; /* K蓖ĂĂȂ̂͏I[L݂̂̏ */
        }
    }
    
    return chartobj;
}


/* ItemSet NX̍\z
----------------------------------------------------------------------
*/
void init_itemset ( void )
{
    /* lCeBuł̍\͌ʂ` */
    rb_cItemSet = rb_define_class_under( rb_mExt, "ItemSet", rb_cData );
    
    /* NX\bh */
    rb_define_singleton_method( rb_cItemSet, "new", itemset_new, -1 );
    
    /* \bh */
    rb_define_method( rb_cItemSet, "initialize"	, itemset_initialize, -1 );
    rb_define_method( rb_cItemSet, "length", itemset_length, 0 );
    rb_define_method( rb_cItemSet, "display", itemset_display, 1 );
    rb_define_method( rb_cItemSet, "display_cyk", itemset_display_cyk, 1 );
    rb_define_method( rb_cItemSet, "display_cyk_lex", itemset_display_cyk_lex, 1 );
    rb_define_method( rb_cItemSet, "display_earley", itemset_display_earley, 1 );
    rb_define_method( rb_cItemSet, "display_viterbi", itemset_display_viterbi, -1 );
    rb_define_method( rb_cItemSet, "display_viterbi_lex", itemset_display_viterbi_lex, -1 );
    rb_define_method( rb_cItemSet, "count_tree", itemset_count_tree, 0 );
}


