/*                                                     project Gemini 
----------------------------------------------------------------------
    s񑀍p[eBeB
      $Id: matrix.c,v 1.6 2002/11/21 13:48:03 abe Exp $

        Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#include "rubytool.h"
#include "matrix.h"
#include <time.h>


/* f[^̈̊mۂƉ
----------------------------------------------------------------------
*/

static Matrix * matrix_new ( Matrix * m )
{
    if( !m )
        m = ALLOC( Matrix );
    
    m->rows = 0;
    m->cols = 0;
    m->buf = NULL;
    m->m = NULL;
    
    return m;
}

static Matrix * matrix_set_anchor ( Matrix * m )
{
    int  i, j;
    
    j = 0;
    for( i = 0; i < m->rows; ++i ) {
        m->m[ i ] = &m->buf[ j ]; /* m->m[ i ] = &m->buf[ i * m->cols ]; */
        j += m->cols;
    }
    
    return m;
}

Matrix * matrix_init ( Matrix * m, int row, int col )
{
    if( !m )
        m = matrix_new( NULL );
    
    m->rows = row;
    m->cols = col;
    REALLOC_N( m->buf, double, row * col );
    REALLOC_N( m->m, double *, row );
    matrix_set_anchor( m );
    MEMZERO( m->buf, double, row * col );
    
    return m;
}

Matrix matrix_create ( int row, int col )
{
    Matrix  m;
    
    m.rows = row;
    m.cols = col;
    m.buf = ALLOC_N( double, row * col );
    m.m = ALLOC_N( double *, row );
    matrix_set_anchor( &m );
    MEMZERO( m.buf, double, row * col );
    
    return  m;
}

void matrix_free ( Matrix * m )
{
    if( m ) {
        ruby_xfree( m->m );
        ruby_xfree( m->buf );
        m->m   = NULL;
        m->buf = NULL;
    }
}

Matrix matrix_dup ( Matrix const * source )
{
    Matrix  m;
    
    matrix_new( &m );
    if( !source )
        return m;
    m = matrix_create( source->rows, source->cols );
    memcpy( m.buf, source->buf, sizeof(double) * m.rows * m.cols );
    
    return m;
}


/* Pʍs̍쐬
----------------------------------------------------------------------
*/
Matrix * matrix_init_element ( Matrix * m )
{
    int  i;
    
    if( !m || m->rows != m->cols )
        return NULL;
    
    rb_zero_memory( m->buf, double, m->rows * m->cols );
    for( i = 0; i < m->rows; ++i )
        m->m[i][i] = 1.0;
    
    return m;
}


/* tšvZ
----------------------------------------------------------------------
 - m g
*/
Matrix * matrix_inverse ( Matrix * m )
{
    int  n, *work;
    int  i, j, k, r, ti, u;
    double  api, pivot, td, w, w1, wmax;
    
    if( !m || m->rows != m->cols )
        return NULL;
    
    if( m->rows < 2 ) {
        m->m[0][0] = 1 / m->m[0][0];
        return m;
    }
    
    n = m->rows;
    work = ALLOCA_N( int, n * sizeof(int) );
    
    w1 = 1.0;
    for( i = 0; i < n; i++ )
        work[i] = i;
    
    for( k = 0, u = 0; k < n; ++k, ++u )
    {
        wmax = 0.0;
        for( i = k; i < n; i++ ) {
            w = fabs( m->m[i][k] );
            if( w > wmax ) {
                wmax = w;
                r = i;
            }
        }
        pivot = m->m[r][k];
        api = fabs( pivot );
        if( api < 0.0000000001 )
            return NULL;
        w1 *= pivot;
        
        if( r != k ) {
            w1 = -w1;
            ti = work[k];
            work[k] = work[r];
            work[r] = ti;
            for( j = 0; j < n; ++j ) {
                td = m->m[u][j];
                m->m[u][j] = m->m[r][j];
                m->m[r][j] = td;
            }
        }
        for( i = 0; i < n; ++i )
            m->m[u][i] /= pivot;
        for( i = 0; i < n; i++ ) {
            if(i != k) {
                w = m->m[i][k];
                if( w != 0.0 ) {
                    for( j = 0; j < n; ++j ) {
                        if( j != k )
                            m->m[i][j] -= w * m->m[k][j];
                    }
                    m->m[i][k] = -w / pivot;
                }
            }
        }
        m->m[k][k] = 1.0 / pivot;
    }
    
    for(i = 0; i < n; i++) {
        while( (k = work[i]) != i ) {
            ti = work[k];
            work[k] = work[i];
            work[i] = ti;
            for( j = 0; j < n; j++ ) {
                td = m->m[j][i];
                m->m[j][i] = m->m[j][k];
                m->m[j][k] = td;
            }
        }
    }
    
    return m;
}


/* s̏Z
----------------------------------------------------------------------
 - Ԃsɂ͕K matrix_free() Kp邱ƁB
*/
Matrix matrix_mul ( Matrix const * left, Matrix const * right )
{
    Matrix  m;
    double  sum;
    int  r, c, i;
    
    matrix_new( &m );
    if( !left || !right || left->cols != right->rows )
        return m;
    
    m = matrix_create( left->rows, right->cols );
    for( r = 0; r < left->rows; ++r ) {
        for( c = 0; c < right->cols; ++c ) {
            sum = 0;
            for( i = 0; i < left->cols; ++i )
                sum += left->m[r][i] * right->m[i][c];
            m.m[r][c] = sum;
        }
    }
    
    return m;
}


/* ts񉉎Z̃eXg
----------------------------------------------------------------------
*/
int matrix_test_inverse ( int size )
{
    Matrix  m, inv, result;
    double  delta;
    int  i, j, flag;
    long t;
    
    srand( (unsigned)time( &t ) );
    m = matrix_create( size, size );
    
    i = size * size;
    for( j = 0; j < i; ++j )
        m.buf[j] = (double)rand() / rand();
    inv = matrix_dup( &m );
    matrix_inverse( &inv );
    
    /* `FbN */
    flag = 1;
    result = matrix_mul( &m, &inv );
    for( i = 0; flag && i < result.rows; ++i ) {
        for( j = 0; flag && j < result.cols; ++j ) {
            if( i == j )
                delta = result.m[i][j] - 1.0;
            else
                delta = result.m[i][j];
            if( fabs(delta) > 0.00000001 )
                flag = 0;
        }
    }
    
    matrix_free( &result );
    matrix_free( &inv );
    matrix_free( &m );
    return flag;
}


/* s̕\
----------------------------------------------------------------------
*/
void matrix_display ( FILE * out, Matrix const * m )
{
    int  i, j;
    
    for( i = 0; i < m->rows; ++i ) {
        for( j = 0; j < m->cols; ++j ) {
            fprintf( out, "% 8.1f ", m->m[i][j] );
        }
        fprintf( out, "\n" );
    }
    fprintf( out, "----\n" );
}


