#!/usr/local/bin/ruby
# $Id: gutil.rb,v 1.28 2002/10/30 07:39:09 abe Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   Grammar Utility
#
#      Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>
######################################################################
require File::join( File.dirname(__FILE__), "nlputil.rb" )


#
#   Used Exception Objects
######################################################################

class NoInputException < RuntimeError
  # ͂f[^^ĂȂƂ̗O
end

class RuleException < RuntimeError
  # @KɋNO
end

class TooMuchException < RuntimeError
  # f[^Ƃ̗O
end

class UnknownModeException < RuntimeError
  # sȎs[hnꂽƂ̗O
end


#
#   Core Object
######################################################################

class GrammarUtility < NLPModule

  DefaultDepth = 3

  VERSION = <<"--------"
#{GEMINI::mklogo("Grammar Utility   ver. 2.0")}
        Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>

--------

  USAGE = <<"--------"
usage:
  gutil.rb [mode] (option)

mode:
  first [grammar file]        output rules and FIRST set.

  pcfg [grammar file]         create gEM parameter file. (PCFG)

  check [grammar file]        check the grammar.

  extract [supportgraph]      extract parameter from supportgraph.

options:
  --depth SIZE                <dSupportGraph> search depth
  -d SIZE                       (default=#{DefaultDepth})
--------

  # RXgN^
  #----------------------------------------
  def initialize ()
    super()
    @VERSION = VERSION
    @USAGE = USAGE
  end
  
  # R}hCIvVł̏
  #----------------------------------------
  ## nꂽ͂̈Ԗڂ𓮍샂[hƂĉ߂B
  def fetch_options ( argv, argtbl, optspec )
    inputs = super( argv, argtbl, optspec )
    if inputs then
      @mode = String(inputs.shift)
    end
    return inputs
  end
  
  # 샂[h̎w
  #----------------------------------------
  def set_action_mode ( mode )
    @mode = mode
    return self
  end
  
  # [eBeBIuWFNg̎s
  #----------------------------------------
  ## RXgN^Ŏw肳ꂽIvVgĎssB
  def call_body ( argtbl )
    @Depth = argtbl['depth'] || DefaultDepth
    if @mode.empty? then
      raise(UnknownModeException, "select program mode.")
    else
      methodname = "mode_#{@mode}"
      if self.public_methods.include?(methodname) then
        self.__send__(methodname)
      else
        raise(UnknownModeException, "unknown mode ``#{@mode}''.")
      end
    end
  end
  
  
  # FIRST W߂
  #----------------------------------------
  # ͂^ĂȂꍇ NoInputException B
  # @KɃG[ꍇ RuleException B
  def mode_first ()
    v_print(" + MODE = first\n")
    
    stream = filename_to_stream( @inputs.shift || "-" )
    raise(NoInputException, "no grammar.") unless stream
    
    grammar = nil
    root = Block.open(stream)
    root.out = @output
    root.each do |block|
      if GrammarBlock === block then
        block.display(@output) {
          block.each do |child|
            if GrammarBlock::RuleBlock === child then
              firstset = child.display {
                rule = GrammarBlock::RuleBlock.new(child).read()
                grammar = Ext::Grammar.new( Array(rule), nil )
                get_first( grammar )
              }
              Block.display("dFIRST", nil) {
                firstset.each_with_index do |list, n|
                  next unless list
                  list = list.sort.map {|i| String(grammar.getcat(i)) }
                  @output << "FIRST( #{grammar.getcat(n)} ) = #{list.join(' ')}\n"
                end
              }
            elsif GrammarBlock::FIRSTBlock === child then
              child.skip
            else
              child.dump
            end
          end
        } # block.display(@output)
      else
        block.dump(@output)
      end
    end
  end

  def get_first ( grammar )
    print_first_progress = lambda {|itr, n| v_print { sprintf("       - \#%d : %-6d\r", itr, n) } }
    
    rules = grammar.rules
    nonterm = grammar.size_nonterm
    preterm = grammar.size_preterm
    terminal = nonterm + preterm - 1
    v_print(
      "   - #{rules.size} rule(s)\n" \
      "   - #{nonterm-1} nonterm(s)\n" \
      "   - #{preterm-1} preterm(s)\n"
    )
    
    catrange = -(preterm-1) .. (nonterm-1)
    firsttbl = Array.new( nonterm )
    firstset = Array.new( nonterm )
    offset = preterm
    
    rules.each_with_index do |rule, n|
      next unless rule
      top, lc = Integer(rule.lhs), Integer(rule.rhs.first)
      raise(RuleException, "LHS out of range. (#{rule.typestring})") unless top != 0 and catrange === top
      raise(RuleException, "Left corner out of range. (#{rule.typestring})") unless lc != 0 and catrange === lc
      firsttbl[top] = Array.new(terminal) unless firstset[top]
      firsttbl[top][lc + offset] = true
      firstset[top] = [ ] unless firstset[top]
      firstset[top] << lc
    end
    
    itr  = 1
    t    = 0
    temp  = Array.new()
    
    changed = true
    while changed do
      changed = false
      for cat in 1 ... nonterm do
        t = update_progress_info( t ) { print_first_progress.call(itr, cat) }
        
        localset = firstset[cat]
        next unless localset
        
        localtbl = firsttbl[cat]
        temp.clear
        localset.each do |keycat|
          next if keycat < 0
          
          value = firstset[keycat]
          next unless value
          
          # FIRST[cat] ̊evfɑΉ FIRST WKp
          value.each do |i|
            addr = i + offset
            unless localtbl[addr] then
              localtbl[addr] = true
              temp << i
            end
          end
        end
        if temp.size > 0 then
          changed = true
          localset.concat( temp )
        end
      end
      itr += 1 if changed
    end
    print_first_progress.call(itr, nonterm-1) ; v_print(EOL)
    v_print(" + OK.\n")
    
    return firstset
  end
  
  
  # PCFG p[^ubNo͂
  #--------------------
  def mode_pcfg ()
    v_print(" + MODE = pcfg\n")
    
    grammarfile = @inputs.shift || "-"
    stream = filename_to_stream( grammarfile )
    raise(NoInputException, "no grammar.") unless stream
    
    grammar = nil
    root = Ext::Block.open( stream )
    root.out = @output
    root.each do |block|
      if GrammarBlock === block then
        raise(TooMuchException, "too many <pGrammar>.") if grammar
        block.display(@output) {
          grammar = GrammarBlock.new(block).read()
        }
      elsif ParamBlock === block then
        block.skip
      else
        block.dump(@output)
      end
    end
    raise(NoInputException, "No <pGrammar>.") unless grammar
    
    # p[^̏o
    grouptbl = { }
    grammar.rules.each do |rule|
      lhs = rule.lhs
      ary = grouptbl[lhs]
      ary = grouptbl[lhs] = [ ] unless ary
      ary << rule.typestring
    end
    output_param_block( grouptbl )
    v_print(" + OK.\n")
  end
  
  def output_param_block ( src )
    # src ̓nbV: ӋL  K
    Block.display("pParameter", nil, @output) do
      src.each_pair do |lhs, ary|
        sum = 0
        datary = Array.new(ary.size)
        ary.each_index do |i|
          seed = rand(255) + 1
          datary[i] = seed
          sum += seed
        end
        
        Block.display("dGroup", nil) do
          ary.each_index do |i|
            print "#{Float(datary[i])/sum} #{ary[i]}#{EOL}"
          end
        end
      end
    end
  end
  
  
  # @`FbN
  #--------------------
  def mode_check ()
    v_print(" + MODE = check\n")
    
    stream = filename_to_stream( @inputs.shift || "-" )
    raise(NoInputException, "no grammar.") unless stream
    
    # --quiet w肳ĂȂ΃ubNo͂
    root = Block.open(stream)
    root.out = @output
    root.each do |child|
      if GrammarBlock === child then
        child.display(@output) do
          check_gramblock( child, @output )
        end
      else
        child.dump(@output)
      end
    end
  end

  RuleLine = Struct.new( :string, :lineno )
  
  def check_gramblock ( block, out )
    id = block['id']
    if id then
      v_print(" + Grammar (ID=#{id}) found.\n")
    else
      v_print(" + Grammar found.\n")
    end
    
    block.each do |child|
      case child
      when GrammarBlock::RuleBlock then
        # ŏɕ@KLbV
        lineary = [ ]
        child.out = nil
        while _ = child.gets do
          lineary << RuleLine.new(_, child.lineno)
        end
        
        # Rg̍폜ƑŐ󔒂̍폜
        for line in lineary do
          line.string.sub!(/\#.*$/, '')
          line.string.strip!
          line.string.gsub!(/\s\s+/, ' ')
        end
        
        lineary = check_rule( lineary )
        child.display(out) do
          lineary.each {|line| out << line.string << "\n" }
        end
      else
        child.dump(out)
      end
    end
  end
  
  def check_rule ( lines )
    v_print("   - check repeated rules...\n")
    checker = { }
    result = [ ] # cK̊i[ꏊ
    lines.each do |line|
      if checker.include?(line.string) then
        v_print("     - line #{line.lineno}: #{line.string}\n")
      else
        checker[line.string] = true
        result << line
      end
    end
    checker = nil
    lines = result
    
    v_print("   - check bracket balance...\n")
    mkerrmes = lambda {|line, mes| "line #{line.lineno}: #{mes}" }
    $errors = [ ]
    lines.each_with_index do |line, n|
      $_ = line.string.sub( /-+>/, '' ) # --> Ƃ܂̂ō폜Ă
      
      if $_.count('<') != $_.count('>') then
        _ = mkerrmes.call(line, "unbalanced < and >, ``#{line.string}''")
        $errors << _
        v_print("     - #{_}\n")
      end
      $_.split.each do |cat|
        if /^<.*[<>].*>$/ === cat then
          _ = mkerrmes.call(line, "suspicious category string ``#{cat}''")
          $errors << _
          v_print("     - #{_}\n")
        end
        if cat.count('{') != cat.count('}') then
          _ = mkerrmes.call(line, "unbalanced { and }, ``#{line.string}''")
          $errors << _
          v_print("     - #{_}\n")
        end
      end
    end
    raise(RuntimeError, "found some errors.") if $errors.size > 0
    
    v_print("   - check loop...\n")
    result = [ ]
    lines.each do |line|
      if /^<(.+?)>\s*-+>\s*<\1>$/ === line.string then
        v_print("     - line #{line.lineno}: #{line.string}\n")
      else
        result << line
      end
    end
    lines = result
    
    return lines
  end
  
  
  # xOtp[^L[𒊏o
  #--------------------
  def mode_extract ()
    print_progress = lambda {|total| v_print { sprintf( "    - \#%d\r", total ) } }
    
    v_print(" + MODE = extract\n")
    require File::join( File.dirname(__FILE__), "gem.so" )
    
    t = 0
    total = 0
    
    $rules = { }
    @inputs.each do |filename|
      stream = open_stream( filename )
      v_print(" + #{(stream)? stream.path: filename}\n")
      unless stream then
        print_warn("No such file -- #{filename}")
      else
        print_progress.call(total)
        block = Ext::Block.open( stream )
        find_supportgraph_block(block, @Depth) do |sgblock|
          begin
            extract_read(sgblock)
            total += 1
            t = update_progress_info( t ) { print_progress.call(total) }
          rescue
            v_print(EOL); print_warn( parent, "\##{total}: #{$!}" )
            t = 0
          end
        end
      end
    end
    
    # oꂽm[hp[^f[^\z
    grammar = Ext::Grammar.new()
    $rules = $rules.keys
    params = { }
    $rules.each do |rulestr|
      rule = grammar.getrule( rulestr.split('|').first )
      params[rule.lhs] = [ ] unless params[rule.lhs]
      params[rule.lhs] << rulestr
    end
    output_param_block( params )
  end
  
  def extract_read ( block )
    bases = { }
    
    # xOtm[h𒊏o
    sg = Ext::SupportGraph.new(block, bases)
    block.close
    
    # O[oe[u bases i[
    bases.each_key do |key|
      $rules[key] = true unless $rules.include?(key)
    end
  end

end


######################################################################
if $0 == __FILE__ then
### ̃t@Ĉ܂܎sꂽꍇ̏
print_error_message = lambda {|mes| $stderr << "ERROR: #{mes}\n" }

begin
  optspec = [ ]
  optspec << ['--depth', '-d', ARG_RQ]
  
  core = GrammarUtility.new()
  argtbl = { }
  inputs = core.fetch_options( ARGV, argtbl, optspec )
  if inputs then
    begin
      core.call( argtbl, inputs )
    rescue UnknownModeException
      print_error_message.call( $! )
      core.print_usage( $stdout )
    end
  end
rescue
  print_error_message.call( $! )
end

###
end
######################################################################
