#!/usr/local/bin/ruby
# $Id: mksg.rb,v 1.22 2002/11/21 13:45:16 abe Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   ItemSet --> SupportGraph converter
#
#      Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>
######################################################################
require File::join( File.dirname(__FILE__), "nlputil.rb" )


#
#   Global Method
######################################################################

# ftHgl
$OPT_id  =  ""

$VERSION = <<"--------"
#{GEMINI::mklogo("ItemSet to SupportGraph converter   ver. 2.0")}
        Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>

--------

$USAGE = <<"--------"
usage:
  ruby mksg.rb (options) (corpus files)

options:
  --grammar FILE              using grammar
  -g FILE

  --cyk                       output CYK-type  (default)

  --cyk-lex                   output simple Lexicalized CYK-type

  --earley                    output Earley-type

  --id STRING                 specify id attribute of <pCorpus>
--------


#
#   Startup
######################################################################

fetch_options(
  ['--grammar', '-g', GetoptLong::REQUIRED_ARGUMENT],
  ['--cyk',     GetoptLong::NO_ARGUMENT],
  ['--cyk-lex', GetoptLong::NO_ARGUMENT],
  ['--earley',  GetoptLong::NO_ARGUMENT],
  ['--id',      GetoptLong::REQUIRED_ARGUMENT]
)
v_print $VERSION
$Verbose = false if $defout.tty?

if defined?($OPT_grammar) and !String($OPT_grammar).empty? then
  read_grammar_file( $OPT_grammar )
else
  $grammar = Ext::Grammar.new()
end


#
#   Application Core
######################################################################

class MksgApplication < NLPApplication
  
  MODE_TABLE = {
    :display_cyk => 'CYK',
    :display_cyk_lex => 'CYK (simple Lexicalized)',
    :display_earley => 'Earley',
  }
  
  def set_processor ( flag, symbol )
    return unless flag
    if @processor then
      fatal_exit "Output mode is exclusive."
    end
    @processor = symbol
  end
  
  def initialize (*arg)
    super(*arg)
    @total = 0
    @processor = nil
    set_processor( $OPT_cyk, :display_cyk )
    set_processor( $OPT_cyk_lex, :display_cyk_lex )
    set_processor( $OPT_earley, :display_earley )
    @processor ||= :display_cyk
    v_print " + output ``#{MODE_TABLE[@processor]}'' type supportgraph.\n"
  end
  attr_reader :total # ǂݍ񂾕̑
  
  def print_progress ()
    v_print "    - \##{@total}\r"
  end
  
  def read_sentence ( sentence )
    itemset = nil
    sentence.each( ItemSetBlock ) do |cls, block|
      if cls then
        raise NLPException, "Found too many <dItemSet>." if itemset
        itemset = block.display { cls.new(block).read }
      else
        if SupportGraphBlock === block then
          block.skip
        else
          block.dump
        end
      end
    end
    return itemset
  end
  
  def on_execute ( corpus )
    t = 0 ; print_progress()
    corpus.each do |sentence|
      t = update_progress_info( t ) { print_progress() }
      @total += 1
      begin
        sentence.out = $defout
        sentence.display do
          itemset = read_sentence( sentence )
          next unless itemset
          
          # ϊ
          display_block(SupportGraphBlock) do
            itemset.send( @processor, $defout )
          end
          
        end # sentence.display
      rescue
        print_warn( corpus.stream, "\##{@total}: #$!" )
        t = 0
      end
    end # corpus.each
    print_progress() ; v_print "\n"
  end

end


#
#   Conversion
######################################################################

app = MksgApplication.new( $files )

if $OPT_id.empty? then
  attr_override = nil
else
  attr_override = { 'id'=>$OPT_id }
end

app.execute( attr_override ) # \̖͂{

v_print " + OK.\n"
