#
# $Id: nlputil.rb,v 1.43 2002/11/21 13:52:38 abe Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   NLP utilities
#
#      Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>
######################################################################
require File::join( File.dirname(__FILE__), "bstream.rb" )


#
#   gemini Information
######################################################################

module GEMINI
  Version = "[gemini 2.0]"
  
  def self.mklogo ( title, col = 70 )
    title = String(title).ljust( col - GEMINI::Version.size )
    return title + GEMINI::Version
  end
end


#
#   Constant and utilities
######################################################################

EOL = "\n"

def empty_line? ( line )
  /^(\#.*)?$/ === String(line)
end


#
#   Stream expansion
######################################################################

class IO
  def path ()
    case self
    when STDIN
      "<STDIN>"
    when STDOUT
      "<STDOUT>"
    when STDERR
      "<STDERR>"
    else
      "<IO>"
    end
  end
end


#
#   String expansion
######################################################################

class String
  def dequote! ()
    if (self[0] == ?\" and self[-1] == ?\") or (self[0] == ?\' and self[-1] == ?\') then
      self.replace(self[1..-2])
    end
    return self
  end
  
  def dequote ()
    if (self[0] == ?\" and self[-1] == ?\") or (self[0] == ?\' and self[-1] == ?\') then
      self[1..-2]
    else
      self.dup
    end
  end
end


#
#   Struct::Tms expansion
######################################################################

class Struct
  class Tms
    def + ( tms )
      Struct::Tms.new( self.utime + tms.utime, self.stime + tms.stime, self.cutime + tms.cutime, self.cstime + tms.cstime )
    end
    def ptime ()
      self.utime + self.stime
    end
  end
end


#
#   Exception related NLP
######################################################################

class NLPException < RuntimeError
  def initialize ( *arg )
    super( *arg )
  end
end


#
#   Word Tree (PoSList alternative)
######################################################################

class WordTree
  # WordTree ͕؂\
  #  children  nil ̏ꍇ leaf  (label ɏI[LȂǂi[)
  def initialize ( *arg )
    @chart = nil
    case arg.size
    when 1,2 # new( initstring [, grammar] )
      tokens, gram = arg
      gram = $grammar if !gram and defined?($grammar)
      parse_initstr( tokens, gram )
    when 3 # new( label, range, children )
      @label, @range, @children = arg
    else
      raise(ArgumentError, "wrong # of arguments.")
    end
  end
  
  # ؂͈̒͂𓾂
  attr_reader :range
  alias to_range range
  
  def first ()
    @range.first
  end
  alias left first
  
  def last ()
    @range.last
  end
  alias right last
  
  # gbvɕtꂽx𓾂
  attr_reader :label
  
  # ӕ̃Xg𓾂
  attr_reader :children
  
  # ؂̒钷𓾂
  def size ()
    @range.length
  end
  alias length size
  
  # ؂ɑΉ`[gւ̃ANZT
  attr_accessor :chart
  
  # xXg𓾂
  def glob ( ary = nil )
    raise(TypeError, "empty tree label") unless @label
    if ary then
      ary << [ @label, @range ]
      Array(@children).each {|child| child.glob(ary) }
    else
      ary = [ ]
      self.glob( ary )
    end
    return ary
  end
  
  # [H
  def leaf? ()
    (@children)? false: true
  end
  alias terminal? leaf?
  
  # I[L̃Xg𓾂
  def terminals ()
    unless defined?(@terms) then
      @terms = [ ]
      for child in @children do
        if child.leaf? then
          @terms << child.label
        else
          @terms.concat( child.terminals() )
        end
      end
    end
    return @terms
  end
  
  def inspect ( out = "" )
    if @children then
      if @label then
        out << "[#{@label}"
        @children.each {|child| child.inspect(out) }
        out << "]"
      else
        out << "(" # ")"
        @children.each {|child| child.inspect(out) }
        out << ")"
      end
    else
      out << "{#{Array(@label).join(' ')}}"
    end
    return out
  end
  
  #---- ȉ̓vCx[g\bh ----#
  private
  
  # p[WOp
  def getcat ( gram, label )
    return label unless gram
    cat = gram.getcat(label)
    raise(RuntimeError, "unknown category ``#{label}''") unless cat
    return cat
  end
  def create_token ( src )
    token = String(src).scan(/\'.*?\'|\".*?\"|[()\[\]{}]|[^()\[\]{}\s]+/)
    token.each {|s| s.dequote! }
    return token
  end
  def parse_initstr ( tokens, gram ) # gbvx͗p
    tokens = create_token( tokens )
    tokens.reverse! # ̂߂ɂЂԂ (pop() Ŏo悤ɂ)
    
    if tokens.last == '[' then
      # c[
      ret = parse_tree( tokens, gram, 1 )
    else
      # ʏ̕
      ret = parse_bracket( tokens, gram, 0, true )
    end
    raise(ArgumentError, "tree syntax error") unless tokens.empty?
    @label, @range, @children = ret.label, ret.range, ret.children
  end
  def parse_wordlist ( tokens, gram ) # PꃊXg
    return [ getcat(gram, tokens.pop) ] unless tokens.last == '{'
    tokens.pop
    
    words = [ ]
    until tokens.last == '}' do
      raise(ArgumentError, "word-list not closed") if tokens.empty?
      words << getcat( gram, tokens.pop )
    end
    tokens.pop
    
    return words
  end
  def parse_bracket ( tokens, gram, left, toplevel ) # ʕt
    unless toplevel then
      return nil unless tokens.last == '(' # ')'
      tokens.pop
    end
    
    right = left
    children = [ ]
    until ((toplevel)? tokens.empty?: (tokens.last == ')')) do
      raise(ArgumentError, "bracket not closed") if !toplevel and tokens.empty?
      case tokens.last
      when '(' then # q bracket
        child = parse_bracket( tokens, gram, right, false )
        if child then
          children << child
          right = children.last.range.last
        end
      else
        children << WordTree.new( parse_wordlist(tokens, gram), right ... right+1, nil )
        right += 1
      end
    end
    tokens.pop unless toplevel
    return nil if children.empty?
    #if children.size == 1 and !children.first.label then # ( ( ... ) )  ( ... ) ɂ
    #  children = children.first.children
    #end
    return WordTree.new( nil, left ... right, children )
  end
  def parse_tree ( tokens, gram, left ) # Tree ͂ƈႤ\
    return nil unless tokens.last == '['
    tokens.pop
    
    label = getcat( gram, tokens.pop )
    right = left
    children = [ ]
    until tokens.last == ']' do
      raise(ArgumentError, "tree-bracket not closed") if tokens.empty?
      case tokens.last
      when '[' then # q Tree
        children << parse_tree( tokens, gram, right )
        right = children.last.range.last
      else # [h
        children << WordTree.new( parse_wordlist(tokens, gram), right ... right+1, nil )
        right += 1
      end
    end
    tokens.pop
    return WordTree.new( label, left ... right, children )
  end
end


#
#   Corpus Components
######################################################################

# <<  >>
#  NLPBlock ̔hNXɂ́AJ萔Ƃ BLOCK_NAME w肷B
#   display_block() O[o\bhpł悤ɂȂB
#
class NLPBlock
  def initialize ( block )
    raise( NLPException, "NLPBlock expects Block. (given #{block.class})" ) unless block.is_a?(Block)
    @block = block
  end
  
  def close ()
    @block.close()
    return self
  end
  def closed? ()
    @block.closed?
  end
  
  def display ( out = $defout )
    @block.display( out ) { yield(out) }
  end
  def dump ( out = $defout )
    @block.dump( out )
  end
  def skip ()
    @block.skip
  end
  
  def out= ( out )
    @block.out = out
  end
  alias monitor= out=
  
  def out ()
    @block.out
  end
  alias monitor out
  
  def stream ()
    @block.stream
  end
  def path ()
    @block.stream.path
  end
  def lineno ()
    @block.stream.lineno
  end
  
  def [] ( key )
    @block[ key ]
  end
  def []= ( key, value )
    @block[ key ] = value
  end
  def block_property ()
    @block.property
  end
  
end

class SentenceBlock < NLPBlock
  BLOCK_NAME = "pSentence"
  
  def self.=== ( block )
    PacketBlock === block and /^[pP][sS]entence$/ === block.name
  end
  
  def initialize ( block )
    raise( ArgumentError, "SentenceBlock expects <pSentence>." ) unless SentenceBlock === block
    super( block )
  end
  
  def read ()
    ret = @block.getb()
    return (block_given?)? yield(ret): ret
  end
  
  def each_element ( *list )
    list = list.flatten
    raise ArgumentError, "SentenceBlock::each_element() expects Class. (#{$_.class})" if $_ = list.find {|i| not Class === i }
    @block.each do |block|
      yield( list.find {|i| i === block }, block )
    end
    return self
  end
  alias each each_element
  
end


#
#   Support graph Block
######################################################################

class SupportGraphBlock < NLPBlock
  BLOCK_NAME = "dSupportGraph"
  
  def self.=== ( block )
    DataBlock === block and /^[dD](Support[gG]raph|supportgraph)$/ === block.name
  end
  
  def initialize ( block )
    raise ArgumentError, "SupportGraphBlock expects <dSupportGraph>." unless SupportGraphBlock === block
    super( block )
  end
  
  def read ()
    raise NLPException, "Already <dSupportGraph> is closed." if @block.closed?
    
    sg = Ext::SupportGraph.new( @block )
    @block.close
    return (block_given?)? yield(sg): sg
  end
  
end


#
#   Item set Block
######################################################################

class ItemSetBlock < NLPBlock
  BLOCK_NAME = "dItemSet"
  
  def self.=== ( block )
    DataBlock === block and /^[dD](Item[sS]et|itemset)$/ === block.name
  end
  
  def initialize ( block )
    raise ArgumentError, "ItemSetBlock expects <dItemSet>." unless ItemSetBlock === block
    super( block )
  end
  
  def read ( gram = $grammar )
    raise NLPException, "Already <dItemSet> is closed." if @block.closed?
    
    itemset = Ext::ItemSet.new( @block, gram )
    @block.close
    return (block_given?)? yield(itemset): itemset
  end
  
  def self.display_itemset ( startname, chart, property = nil, out = $defout )
    display_block(ItemSetBlock, property, out) do
      out << "\#@ROOT = #{startname}\n"
      chart.display_itemset(out)
    end
  end
  
end


#
#   PoS Block
######################################################################

class PoSBlock < NLPBlock
  BLOCK_NAME = "dPoS"
  
  def self.=== ( block )
    DataBlock === block and /^[dD](PoS|pos)$/ === block.name
  end
  
  def initialize ( block )
    raise ArgumentError, "PoSBlock expects <dPoS>." unless PoSBlock === block
    super( block )
  end
  
  def read ( gram = $grammar )
    raise NLPException, "Already <dPoS> is closed." if @block.closed?
    
    poslist = WordTree.new( @block.read(), gram )
    @block.close
    return (block_given?)? yield(poslist) : poslist
  end
end


#
#   Corpus
######################################################################

class CorpusBlock < NLPBlock
  BLOCK_NAME = "pCorpus"
  
  def self.=== ( block )
    PacketBlock === block and /^[pP][cC]orpus$/ === block.name
  end
  
  def initialize ( block )
    raise ArgumentError, "CorpusBlock requires <pCorpus>." unless CorpusBlock === block
    super( block )
  end
  
  def read ()
    raise NLPBlock, "Already <pCorpus> is closed." if @block.closed?
    
    ret = nil
    while block = @block.getb() do
      if SentenceBlock === block then
        ret = SentenceBlock.new(block)
        break
      end
      block.close
    end
    return (ret)? ((block_given?)? yield(ret): ret):nil
  end
  alias get_sentence read
  alias getb get_sentence
  
  def each_sentence ()
    while sent = get_sentence() do
      yield( sent )
      sent.close()
    end
  end
  alias each each_sentence
end


#
#   Corpus File Block
######################################################################

class CorpusStream
  
  def initialize ( stream )
    @root = Block.open( stream )
    if block_given? then
      yield( self )
      @root.close()
    end
  end
  
  def get_corpus ()
    while block = @root.getb() do
      return CorpusBlock.new( block ) if CorpusBlock === block
      block.close
    end
    return nil
  end
  alias getb get_corpus
  
  def stream ()
    @root.stream
  end
  def close ()
    @root.close()
    return self
  end
  def each_corpus ()
    while corpusblock = get_corpus() do
      yield( corpusblock )
      corpusblock.close()
    end
    return self
  end
  alias each each_corpus
  
  def path ()
    @root.stream.path
  end
  def lineno ()
    @root.stream.lineno
  end
  
end


#
#   Estimated tree
######################################################################

class EstimateBlock < NLPBlock
  BLOCK_NAME = "dEstimate"
  
  EstTree = Struct.new( :tree, :prob )
  
  def self.=== ( block )
    DataBlock === block and /^[dD][eE]stimates?$/ === block.name
  end
  
  def initialize ( block )
    raise( ArgumentError, "EstimateBlock requires <dEstimate>." ) unless EstimateBlock === block
    super( block )
  end
  
  RE_index = /^\#\s*(\d+)(\s*\(\/\s*\d+\s*\))?\s+/
  def read ( grammar = $grammar )
    raise "Already <dEstimate> is closed." if @block.closed?
    
    ary = [ ]
    
    $_ = @block.gets()
    while $_ do
      # n Ԗڂ̐\؂ł鎖s܂œǂݐi߂
      until RE_index === $_ do
        $_ = @block.gets()
        raise NLPException, "Syntax error. (no index line)" unless $_
      end
      treeindex = Integer($1)
      treeprob  = Float($')
      
      line = ""
      until RE_index === ($_ = @block.gets()) do
        break unless $_
        line << $_
      end
      tree = WordTree.new( line, grammar )
      
      ary[treeindex-1] = EstTree.new( tree, treeprob )
    end
    @block.close
    
    # CfNXȂ𒲂ׂĂ
    ary.each_with_index do |tree, n|
      raise NLPException, "Syntax error. (<dEstimate> \##{n} empty)" unless tree
    end
    
    return (block_given?)? yield(ary): ary
  end
end


#
#   Tree block
######################################################################

class TreeBlock < NLPBlock
  BLOCK_NAME = "dTree"
  
  def self.=== ( block )
    DataBlock === block and /^[dD][tT]ree$/ === block.name
  end
  
  def initialize ( block )
    raise( ArgumentError, "TreeBlock expects <dTree>." ) unless TreeBlock === block
    super( block )
    @tree = nil
  end
  
  def read ( grammar = $grammar )
    raise "Already <dTree> is closed." if @block.closed?
    tree = WordTree.new( @block.read(), grammar )
    @block.close
    return (block_given?)? yield(tree): tree
  end
end


#
#   Log block
######################################################################

LogBlock = HashBlock.new( "dLog", /^[dD][lL]og$/ )
class LogBlock
  BLOCK_NAME = "dLog"
end


#
#   Param block
######################################################################

class ParamBlock < NLPBlock
  BLOCK_NAME = "pParameter"
  
  def self.=== ( block )
    PacketBlock === block and /^[pP][pP]arameter$/ === block.name
  end
  
  def initialize ( block )
    raise(ArgumentError, "ParamBlock expects <dParameter>.") unless ParamBlock === block
    super( block )
  end
  
  def read ()
    raise("Already <dParameter> is closed.") if @block.closed?
    
    # [ [ [Prob, SWName], [Prob, SWName], ... ], ... ] ̍\
    param = [ ]
    @block.each do |child|
      next unless DataBlock === child and /^[dD][gG]roup$/ === child.name
      sum = 0.0
      group = [ ]
      child.each do |line|
        line = line.split('#', 2).first
        line.strip!
        next if empty_line?(line)
        begin
          line = line.split(' ', 2)
          case line.length
          when 1 then
            raise NLPException, "Parameter key is empty."
          when 2 then
            pair = [ Float(line.first), String(line.last) ]
            sum += pair.first
            group << pair
          end
        rescue
          raise NLPException, "<dGroup> syntax error. (#{$!})"
        end
      end
      if sum != 1.0 then
        # a 1 ł͂Ȃ̂ŐKsĂ
        group.each_index do |n|
          group[n][0] = group[n][0] / sum
        end
      end
      param << group
    end
    @block.close
    
    return (block_given?)? yield(param): param
  end
end


#
#   Grammar block
######################################################################

class GrammarBlock < NLPBlock
  BLOCK_NAME = "pGrammar"
  
  def self.=== ( block )
    PacketBlock === block and /^[pP][gG]rammar$/ === block.name
  end
  
  class RuleBlock < NLPBlock
    BLOCK_NAME = "dRule"
    
    def self.=== ( block )
      DataBlock === block and /^[dD][rR]ules?$/ === block.name and GrammarBlock === block.parent
    end
    def initialize ( block )
      raise( ArgumentError, "RuleBlock expects <dRule>." ) unless RuleBlock === block
      super( block )
    end
    def read ()
      raise "Already <dRule> is closed." if @block.closed?
      
      lines = [ ]
      @block.each do |line|
        line.strip!
        next if empty_line?(line)
        lines << line
      end
      @block.close
      return lines
    end
  end
  # RuleBlock
  
  class FIRSTBlock < NLPBlock
    BLOCK_NAME = "dFIRST"
    
    def self.=== ( block )
      DataBlock === block and /^[dD]([fF]irst|FIRST)$/ === block.name and GrammarBlock === block.parent
    end
    def initialize ( block )
      raise( ArgumentError, "FIRSTBlock expects <dFIRST>." ) unless FIRSTBlock === block
      super( block )
    end
    def read ()
      raise "Already <dRule> is closed." if @block.closed?
      
      table = { }
      prevline = ""
      @block.each do |line|
        line.strip!
        next if empty_line?(line)
        if /^(?:FIRST|[fF]irst)\s*\((.+?)\)\s*=\s*/ === line then
          key, prevline = $1.strip, $'
          table[key] = prevline
        else
          prevline << line # sŏIȂꍇlĂ̏u
        end
      end
      @block.close
      
      ret = { }
      table.each_pair {|key, value| ret[key] = String(value).split(' ') }
      return ret
    end
  end
  # FIRSTBlock
  
  def initialize ( block )
    raise( ArgumentError, "GrammarBlock expects <pGrammar>." ) unless GrammarBlock === block
    super( block )
  end
  
  def raw_read ()
    raise("Already <pGrammar> is closed.") if @block.closed?
    
    rule = [ ]
    firstset = { }
    @block.each do |child|
      child.display(self.out) do
        case child
        when RuleBlock then
          rule.concat( RuleBlock.new(child).read )
        when FIRSTBlock then
          firstset.update( FIRSTBlock.new(child).read )
        end
      end
    end
    @block.close
    
    return [rule, firstset]
  end
  
  def read ()
    rule, firstset = raw_read()
    raise(NLPException, "Grammar requires <dRule>.") if rule.empty?
    
    grammar = Ext::Grammar.new( Array(rule), Array(firstset) )
    grammar.block_property = @block.property
    
    return (block_given?)? yield(grammar): grammar
  end
end


#
#   Sentence text block
######################################################################

class TextBlock < NLPBlock
  BLOCK_NAME = "dText"
  
  def self.=== ( block )
    DataBlock === block and /^[dD][tT]ext$/ === block.name
  end
  
  def initialize ( block )
    raise( ArgumentError, "TextBlock expects <dText>." ) unless TextBlock === block
    super( block )
  end
  
  def read ()
    raise "Already <dText> is closed." if @block.closed?
    ret = @block.read.strip.split
    return (block_given?)? yield(ret): ret
  end
  
end


#
#   Utility
######################################################################


## ԑ
########################################

BEGIN {
  $get_process_time = 
    case RUBY_VERSION
    when /^1\.6\./
      lambda { Time::times }
    else #when /^1\.7\./
      lambda { Process::times }
    end
}
def get_process_time ()
  $get_process_time.call()
end

def time_block ( &proc )
  t1 = $get_process_time.call()
  $_ = proc.call()
  t2 = $get_process_time.call()
  
  $time = Struct::Tms.new( t2.utime - t1.utime, t2.stime - t1.stime, t2.cutime - t1.cutime, t2.cstime - t1.cstime )
  
  return $_
end

# ܂gpȂ (gp}sȌۂ)
def time_block_withoutGC ( &proc )
  GC.start
  GC.disable
  $_ = time_block( &proc )
  GC.enable
  
  return $_
end


## bZ[Wo͊֌W
########################################

def display_block ( cls, property = nil, out = $defout )
  Block.display( cls::BLOCK_NAME, property, out ) { yield(out) }
end

def v_print ( arg = nil )
  return unless $Verbose
  $stderr.print( block_given? ? yield : arg )
end

def print_backtrace ()
  $stderr << "    #{$@.first}\n" if $DEBUG and $@
end

module GEMINI
  def self.make_warn_string ( *arg )
    case arg.size
    when 2 then
      object, mes = arg
      
      # IuWFNg̕ϊ
      case object
      when Block
        object = object.stream
      end
      
      # IuWFNgɂ鏈
      case object
      when IO
        mes = "#{object.path}: #{object.lineno}: #{mes}"
      end
      
      String(mes)
    else
      String(arg.last)
    end
  end
end

def fatal_exit ( *arg )
  unless arg.empty? then
    $stderr << "ERROR: " << GEMINI::make_warn_string(*arg) << "\n"
    print_backtrace()
  end
  exit(-1)
end

def print_warn ( *arg )
  $stderr << GEMINI::make_warn_string(*arg) << "\n"
  print_backtrace()
end

def update_progress_info ( t )
  if t+1 <= Time.now.tv_sec then
    t = Time.now.tv_sec
    yield
  end
  return t
end


## Xg[֌W
########################################

def open_stream ( filename )
  if IO === filename then
    stream = filename
  else
    filename = String(filename)
    if filename == "-" then
      stream = $stdin
    else
      stream = (File::exist?(filename))? open(filename): nil
    end
  end
  return stream
end

def filename_to_stream ( filename )
  stream = open_stream( filename )
  if stream then
    if block_given? then
      yield( stream )
    else
      v_print " + #{stream.path}\n"
    end
  else
    print_warn "No such file -- #{filename}"
  end
  
  return stream
end

FoundCorpusInfo = Struct.new( :root, :corpus )
def find_first_corpus ( files )
  # Ԓl\̂Ȃ̂Œӂ邱
  raise ArgumentError, "find_first_corpus() expect Array for files." unless files.is_a?(Array)
  until files.empty? do
    stream = filename_to_stream( files.shift )
    next unless stream
    
    root = CorpusStream.new(stream)
    corpus = root.get_corpus()
    return FoundCorpusInfo.new(root, corpus) if corpus
  end
  return nil
end

def process_all_corpus ( files )
  Array(files).each do |filename|
    if FoundCorpusInfo === filename then
      root, corpus = Array(filename)
      yield( corpus )
    else
      stream = filename_to_stream( filename )
      next unless stream
      root = CorpusStream.new(stream)
    end
    
    root.each {|corpus| yield(corpus) }
    root.close
  end
end

def stream_read_grammar ( stream )
  Block.open(stream).each do |block|
    if GrammarBlock === block then
      begin
        return GrammarBlock.new(block).read()
      rescue NLPException
        fatal_exit( block, $! )
      end
    end
  end
  return nil
end

def stream_read_param ( stream )
  Block.open(stream).each do |block|
    if ParamBlock === block then
      begin
        return ParamBlock.new(block).read()
      rescue NLPException
        fatal_exit( stream, $! )
      end
    end
  end
  return nil
end

def read_grammar_file ( filename )
  fatal_exit "Select grammar. (option --grammar or -g)" if !filename or filename.empty?
  
  stream = open_stream( filename )
  fatal_exit "No such grammar file -- #{filename}." unless stream
  
  v_print " + #{stream.path}\n" \
          "   - now reading...\r"
  $grammar = stream_read_grammar( stream )
  unless $grammar then
    v_print "\n"
    fatal_exit "No grammar found."
  end
  
  gramid = $grammar.block_property['id']
  size_rule = $grammar.size_rule
  size_nonterm = $grammar.size_nonterm
  size_preterm = $grammar.size_preterm
  if gramid then
    v_print "   - ID=#{gramid}, Rule=#{size_rule}, NT=#{size_nonterm}, PT=#{size_preterm}\n"
  else
    v_print "   - Rule=#{size_rule}, NT=#{size_nonterm}, PT=#{size_preterm}\n"
  end
  
  return $grammar
end

def read_param_file ( filename )
  fatal_exit "Select parameter. (option --param or -p)" unless filename
  
  stream = open_stream( filename )
  fatal_exit "No such parameter file -- #{filename}." unless stream
  
  v_print " + parameter: #{stream.path}\n" \
          "   - now reading...\r"
  $param = stream_read_param( stream )
  unless $param then
    v_print "\n"
    fatal_exit("No parameter found in ``#{stream.path}''.")
  end
  v_print "   - OK.           \n"
  
  return $param
end

module GEMINI
  def self.print_usage ()
    $stderr << (($USAGE.is_a?(Proc))? $USAGE.call: $USAGE)
  end
  
  def self.print_version ()
    $stderr << $VERSION
  end
  
  def self.print_help ()
    GEMINI::print_version
    GEMINI::print_usage
  end
end

require "getoptlong.rb"
ARG_NO = GetoptLong::NO_ARGUMENT
ARG_RQ = GetoptLong::REQUIRED_ARGUMENT
def fetch_options ( *optionlist )
  $OPT_quiet = $OPT_help = $OPT_version = nil
  
  parser = GetoptLong.new
  parser.set_options(
    ['--quiet', '-q', ARG_NO],
    ['--help', '-h', ARG_NO],
    ['--version', ARG_NO],
    *optionlist
  )
  
  begin
    parser.each do |optname, optarg|
      optname = optname.dup
      optname.sub!(/^--?/, '') # 擪 -- (or -) ͕sv
      optname.gsub!(/-/, '_' ) # ׂĂ -  _ ɕϊ
      eval("$OPT_#{optname} = '#{optarg}'")
    end
  rescue # GetoptLong::InvalidOption
    GEMINI::print_usage()
    exit
  end
  
  $OPT_v = ($OPT_quiet)? false: true
  $Verbose = $OPT_v
  
  $files = (ARGV.empty?)? [ STDIN ]: ARGV.dup
  
  if $OPT_version then
    GEMINI::print_version()
    exit
  end
  
  if $OPT_help then
    GEMINI::print_help()
    exit
  end
end

def find_supportgraph_block ( block, depth = $Depth )
  if SupportGraphBlock === block then
    yield( block )
  else
    if Ext::PacketBlock === block then
      if depth < 0 or block.level+1 <= depth then
        block.each do |child|
          find_supportgraph_block(child, depth) {|sgblock| yield(sgblock) }
        end
      end
    end
  end
  block.close()
end


#
#   Core class for NLP applications
######################################################################
# NLPApplication
#  ̃R[pXǂݍŏA̃R[pXƂČʂԂ
#  AvP[ṼfB
#  Ԗڂ̃R[pXǂ݂ȂRs[B
#  (attr_override ^ꂽꍇ͂̑Rs[)
#
# NLPModule
#  IuWFNgꂽAvP[Vp̃x[XT[rX񋟂B
#

class NLPModule
  
  def initialize ( stdin = $stdin, stdout = $stdout, stderr = $stderr )
    @Verbose = false
    @OPT_v = nil
    @stdin = stdin
    @stdout = stdout
    @stderr = stderr
  end
  
  # ͗Ԃ
  #  argv ... ͂ꂽR}hC
  #  table ... w肳ꂽIvVރnbVe[u
  #  optspec ... IvV̎dl
  def fetch_options ( argv, table, optspec )
    inputs = nil
    argv_buf = ARGV.dup
    begin
      ARGV.replace(Array(argv))
      parser = GetoptLong.new
      parser.set_options(
        ['--quiet', '-q', ARG_NO],
        ['--help', '-h', ARG_NO],
        ['--version', ARG_NO],
        *Array(optspec)
      )
      parser.each do |optname, optarg|
        optname = optname.dup
        optname.sub!(/^--?/, '') # 擪 -- (or -) ͕sv
        table[optname] = optarg
      end
      @OPT_v = (table['quiet'])? false: true
      @Verbose = @OPT_v
      @Verbose = false if @stdout.is_a?(IO) and @stdout.tty?
      inputs = (ARGV.empty?)? [ @stdin ]: ARGV.dup
      
      if table['version'] then
        print_version( @stdout )
        return nil
      end
      if table['help'] then
        print_version( @stdout )
        print_usage( @stdout )
        return nil
      end
      
      print_version(@stderr) if @Verbose
    ensure
      ARGV.replace(argv_buf)
    end
    return inputs
  end
  
  # call Ă΂܂ŎgpłȂ
  def v_print ( arg = nil )
    return nil unless @Verbose
    string = String((block_given?)? yield: arg)
    @stderr << string
    return string
  end
  
  # \bh call_body(argtbl) Ē`Kv
  def call ( argtbl, inputs, output = nil, errout = nil )
    @inputs = Array(inputs)
    @output = output || @stdout
    @errout = errout || @stderr
    return call_body(argtbl)
  end
  
  # (Proc Ȃ) ̏o
  def print_information ( out, info )
    out << ((info.is_a?(Proc))? info.call(): String(info))
  end
  
  # o[W̏o
  def print_version ( out = @stdout )
    print_information( out, @VERSION )
  end
  
  # g̏o
  def print_usage ( out = @stdout )
    print_information( out, @USAGE )
  end
  
  # x̏o
  def print_warn ( *arg )
    @errout << GEMINI::make_warn_string(*arg) << "\n"
  end
  
  # quiet [hœ쒆𒲂ׂ
  def quiet? ()
    !@Verbose
  end
  
  # R[pX̏ (o͕t)
  def process_multi_corpus ( files, attr_override = nil )
    files = Array(files).dup
    
    # t@C݂邩ǂmF
    check_all_files( files )
    
    # ŏ̃R[pX
    prefetch = find_first_corpus( files )
    raise(RuntimeError, "no corpus found") unless prefetch
    corpusprop = prefetch.last.block_property
    if attr_override then
      raise(ArgumentError, "NLPModule::process_multi_corpus() requires Hash") unless attr_override.is_a?(Hash)
      corpusprop.update(attr_override)
    end
    
    display_block(CorpusBlock, corpusprop, @output) do
      files.each do |filename|
        if String === filename then
          stream = filename_to_stream( filename )
          next unless stream
          root = CorpusStream.new(stream)
        else
          # filename  FindResult ̃nY
          root, corpus = Array(filename)
          yield(corpus)
        end
        
        # 擪ȊO <pCorpus> 
        root.each {|corpus| yield(corpus) }
        root.close
      end
    end
    
    return self
  end
  def check_all_files ( files )
    not_exist = 0
    Array(files).each do |file|
      unless File::exist?(file) then
        print_warn("no such file -- #{file}.")
        not_exist += 1
      end
    end
    if not_exist > 0 then
      print_warn("#{not_exist} file#{(not_exist > 1)? 's': ''} not found.")
      return false
    end
    return true
  end
  def find_first_corpus ( files )
    until files.empty? do
      stream = filename_to_stream( files.shift )
      next unless stream
      
      root = CorpusStream.new(stream)
      corpus = root.get_corpus()
      if corpus then
        files.unshift([root, corpus])
        return files.first
      end
    end
    return nil
  end
  def filename_to_stream ( filename )
    if filename.is_a?(String) then
      stream = (filename == "-")? @stdin: open(filename)
    else
      stream = filename
    end
    
    if stream then
      on_open_stream( stream )
    else
      print_warn("No such file -- #{filename}")
    end
    
    return stream
  end
  private :check_all_files, :filename_to_stream, :find_first_corpus
  
  # R[pX̏ (\)
  def read_multi_corpus ( files )
    check_all_files( files ) # t@C݂邩ǂmF
    
    # et@C̃R[pXf[^񂵂Ă
    files.each do |filename|
      stream = filename_to_stream( filename )
      next unless stream
      root = CorpusStream.new(stream)
      root.each {|corpus| yield(corpus) }
      root.close
    end
    
    return self
  end
  
  # Xg[JƂɌĂяonh
  def on_open_stream ( stream )
    v_print(" + #{stream.path}\n")
  end
  
  # @̓ǂݍ
  def read_grammar ( gramfile )
    stream = filename_to_stream(gramfile)
    return nil unless stream
    
    g = nil
    Block.open(stream).each do |block|
      if GrammarBlock === block then
        g = GrammarBlock.new(block).read()
        break
      end
    end
    unless g then
      print_warn("no grammar found")
      return nil
    end
    
    if @Verbose then
      ary = [ ]
      id = String(g.block_property['id'])
      ary << "ID=#{id}" unless id.empty?
      ary << "rule=#{g.size_rule}"
      ary << "nonterm=#{g.size_nonterm}"
      ary << "preterm=#{g.size_preterm}"
      v_print("   - #{ary.join(', ')}\n")
    end
    
    return g
  end
  
  # p[^t@C̓ǂݍ
  def read_param ( paramfile )
    stream = filename_to_stream( paramfile )
    return nil unless stream
    
    ret = nil
    Block.open(stream).each do |block|
      if ParamBlock === block then
        ret = ParamBlock.new(block).read()
        break
      end
    end
    unless ret then
      print_warn("no parameter found")
      return nil
    end
    
    v_print("   - OK.\n")
    
    return ret
  end
  
  # ԂōƏ󋵂\邽߂̃R[h
  def update_progress_info ( t )
    if t+1 <= Time.now.tv_sec then
      t = Time.now.tv_sec
      yield()
    end
    return t
  end
end

# obsolete Ȃ̂ NLPModule Ɉڍs邱
class NLPApplication
  
  def initialize ( files )
    @files = files.dup
    not_exist = 0
    @files.each do |file|
      unless File::exist?(file) then
        print_warn("No such file -- #{file}.")
        not_exist += 1
      end
    end
    print_warn("#{not_exist} files not found.") if not_exist > 0
  end
  
  def on_open_stream ( stream )
    v_print(" + #{stream.path}\n")
  end
  
  def on_execute ( corpus )
    fatal_exit("NLPApplication::on_execute must be override.")
  end
  
  def filename_to_stream ( filename )
    if IO === filename then
      stream = filename
    else
      filename = String(filename)
      if filename == "-" then
        stream = $stdin
      else
        stream = open(filename)
      end
    end
    
    if stream then
      on_open_stream( stream )
    else
      print_warn "No such file -- #{filename}"
    end
    
    return stream
  end
  
  FindResult = Struct.new(:root, :corpus)
  def find_first_corpus ()
    # Ԓl\̂Ȃ̂Œӂ邱
    until @files.empty? do
      stream = filename_to_stream( @files.shift )
      next unless stream
      
      root = CorpusStream.new(stream)
      corpus = root.get_corpus()
      return FindResult.new(root, corpus) if corpus
    end
    return nil
  end
  
  def execute ( attr_override = nil )
    # <pCorpus> ̑𓾂邽߂ɐǂ݂Ă
    prefetch = find_first_corpus()
    fatal_exit "No corpus found." unless prefetch
    
    @files.unshift(prefetch) # find_first_corpus() Ő擪vfD̂ŕ[
    corpusprop = prefetch.corpus.block_property
    if attr_override then
      raise ArgumentError, "NLPApplication::execute() expects Hash." unless attr_override.is_a?(Hash)
      corpusprop.update( attr_override )
    end
    
    display_block(CorpusBlock, corpusprop) do
      @files.each do |filename|
        if String === filename then
          stream = filename_to_stream( filename )
          next unless stream
          root = CorpusStream.new(stream)
        else
          # filename  FindResult ̃nY
          root, corpus = Array(filename)
          on_execute( corpus )
        end
        
        # 擪ȊO <pCorpus> 
        root.each {|corpus| on_execute(corpus) }
        root.close
      end
    end
  end
  
end


#
#   Extention
######################################################################

require File::join( File.dirname(__FILE__), "nlputil.so" ) unless defined?(Ext::Grammar)
